// polynomial arithmetic
// last change: 2004-07-07

#ifndef __POLYNOMIAL_HEADER__
#define __POLYNOMIAL_HEADER__

/*! @file
 * @brief
 * declaration of polynomial arithmetic interface for multiple precision numbers
 */

#include <gmp.h>

//! contains polynomial arithmetic concepts for mpz_t
namespace polynomial
{

//#define DEBUG    /* for explicite DEBUGGING-sessions */
//#define VERBOSE  /* be more verbose */

// this define should be triggered by Makefile
// #define USE_DFT /* enable use of early-alpha discrete fourier transform */

typedef mpz_t* TPolynom;
typedef const mpz_t* TconstPolynom;
/* Ein Polynom ist ein Feld aus mpz_t-Koeffizienten:
   P(x)=a0*x^0+a1*x^1+...+a[k-1]*x^[k-1],
   gespeichert als (a0,a1,a2,...,a[k-1]).
*/

/* Hinweise zur Beachtung bei Funktionsaufrufen:

 TPolynom bezeichnet einen Zeiger auf ein Koeffizientenfeld
 (welches ber seine Koeffizienten das Polynom definiert).

 Wie in C++ blich, sind daher folgende Aufrufkonventionen
 zu beachten bzw. zu unterscheiden:

 Parameter		Restriktionen/Eigenschaften
 -------------------------------------------------------------
 mpz_t* p		 Zeiger vernderbar, Daten vernderbar
 const mpz_t* p		 Zeiger vernderbar, Daten konstant
 mpz_t* const p		 Zeiger konstant,    Daten vernderbar
 const mpz_t* const p    Zeiger konstant,    Daten konstant

 Nun sind aber "const TPolynom P" und "TPolynom const P"
 (leider) quivalent und bezeichnen beide
 einen konstanten Zeiger auf ein vernderbares Polynom!

 Deshalb habe ich zustzlich einen Datentyp fr
 konstante Polynome definiert: TconstPolynom
 Damit gilt:

 Parameter              Restriktionen/Eigenschaften
 ----------------------------------------------------------------
 TPolynom P              P vernderbar, Koeffizienten vernderbar
 const TPolynom P        P konstant,    Koeffizienten vernderbar
 TPolynom const P	 P konstant,    Koeffizienten vernderbar
 TconstPolynom P         P vernderbar, Koeffizienten konstant
 const TconstPolynom P	 P konstant,    Koeffizienten konstant
 TconstPolynom const P   P konstant,    Koeffizienten konstant

*/


// a tiny helper class for temporary arrays mpz_t[] in C++
// self initializing and self destructing
// (Einziger Sinn und Zweck: Ein komfortabler Ersatz fr die nach ISO-Standard
//  nicht untersttzten variablen Arrays)
class TTempPolynom
{
private:
  mpz_t* const data;
  const int _capacity;
public:
  explicit TTempPolynom(const int n) : data(new mpz_t[n]), _capacity(n)
   {
     //cout << "allocate polynomial of size " << n << endl;
     for (int i=0; i<n; ++i) mpz_init(data[i]);
   }
  TTempPolynom(const int n, const int estimated_operand_size) : data(new mpz_t[n]), _capacity(n)
   {
     //cout << "allocate polynomial of size " << n << endl;
     for (int i=0; i<n; ++i) mpz_init2(data[i],estimated_operand_size);
   }
  ~TTempPolynom()
   {
     //cout << "dispose polynomial of size " << _capacity << endl;
     for (int i=_capacity-1; i>=0; --i) mpz_clear(data[i]);
     delete [] data;
   }
  const int capacity() const { return _capacity; }

  //const mpz_t& operator[] (const unsigned int i) const { return data[i]; }
  //mpz_t& operator[] (const unsigned int i) { return data[i]; }

  // dangerous, but helpful:
  // remember: TPolynom will not live longer than TempPolynom!
  operator const TPolynom() const { return data; }
};


/*!
 * @param P polynomial
 * @param k size of polynomial (numer of coefficients)
 * @remark
 *	   - This function prints the given polynomial to stdout.
 */
void print (const TconstPolynom P, int k);


/*!
 * @param res result value
 * @param P polynomial to evaluate
 * @param k size of @p P
 * @param x point at which @p P is to evaluate
 * @param m result will be reduced modulo @p m
 * @remark
 *	   - This function implements the horner scheme
 *         - use it only for single point evaluations of @p P
 *         - @p res = P(x) mod m
 */
void eval(mpz_t res, const TconstPolynom P, const int k, const mpz_t x, const mpz_t m);
// berechnet mit dem Hornerschema res = P(x) mod m


#ifdef USE_DFT
 // fast discrete fourier transformation
 // (will allocate internal temporary memory, which you may want to release)

/*!
 * @remark
 *	   - When the discrete fast fourier transform is activated, then it
 *           will use some internal temporary memory, which can be safely
 *           released using this function.
 */
void clear_dft_tempmemory() ;

#endif


/*!
 * @param Pr result polynomial
 * @param kr size of result polynomial
 * @param P1 first multiplicator polynomial
 * @param k1 size of @p P1
 * @param P2 second multiplicator polynomial
 * @param k2 size of @p P2
 * @result size of the product polynomial @p Pr
 * @remark
 *	   - This function implements the classical multiplication
 *         - complexity O(k1*k2)
 *         - use it only for small sized polynomials
 *         - @p &Pr must be disjoint from @p &P1 and @p &P2
 */
int classic_mul(const TPolynom Pr, const int kr,
                const TconstPolynom P1, const int k1,
                const TconstPolynom P2, const int k2);
// Pres = P1*P2, &Pres mu von &P1,&P2 verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(k1*k2)
// (Klassische Methode, alternativ zu Karatsuba implementiert fr kleine Polynome)


/*!
 * @param Pr result polynomial
 * @param kr size of result polynomial
 * @param P1 first multiplicator polynomial
 * @param k1 size of @p P1
 * @param P2 second multiplicator polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @result size of the product polynomial @p Pr
 * @remark
 *	   - This function implements the classical multiplication
 *         - complexity O(k1*k2)
 *         - use it only for small sized polynomials
 *         - @p &Pr must be disjoint from @p &P1 and @p &P2
 */
int classic_mul(const TPolynom Pr, const int kr,
                const TconstPolynom P1, const int k1,
                const TconstPolynom P2, const int k2, const mpz_t m);
// Pres = P1*P2, &Pres mu von &P1,&P2 verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(k1*k2)
// (Klassische Methode, alternativ zu Karatsuba implementiert fr kleine Polynome)



/*!
 * @param R result polynomial
 * @param kR size of result polynomial
 * @param P multiplicator polynomial
 * @param k size of @p P
 * @param m the result will be reduced modulo @p m
 * @result size of the product polynomial @p R
 * @remark
 *	   - This function implements the Karatsuba multiplication
 *         - complexity O(k^1.59)
 *         - @p &R must be disjoint from @p &P
 */
int square(const TPolynom R, const int kR, const TconstPolynom P, const int k, const mpz_t m);
// R = P^2, &R mu von &P verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(k^1.59) -> Karatsuba


/*!
 * @param R result polynomial
 * @param kR size of result polynomial
 * @param P first multiplicator polynomial
 * @param k size of @p P1
 * @result size of the product polynomial @p R
 * @remark
 *	   - This function implements the Karatsuba multiplication
 *         - complexity O(k^1.59)
 *         - @p &R must be disjoint from @p &P
 */
int square(const TPolynom R, const int kR, const TconstPolynom P, const int k);
// R = P^2, &R mu von &P verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(k^1.59) -> Karatsuba


/*!
 * @param R result polynomial
 * @param kR size of result polynomial
 * @param P1 first multiplicator polynomial
 * @param k1 size of @p P1
 * @param P2 second multiplicator polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @result size of the product polynomial @p R
 * @remark
 *	   - This function implements the various multiplication algorithms
 *         - complexity O(max(k1,k2)^1.59), Karatsuba
 *         - complexity O((k1+k2)*ld(k1+k2)), dft & chinese remaindering
 *         - use it for medium and big sized polynomials
 *         - @p &R must be disjoint from @p &P1 and @p &P2
 */
int mul(const TPolynom R, const int kR,
        TconstPolynom P1, int k1,
        TconstPolynom P2, int k2, const mpz_t m);
// Pres = P1*P2, &R mu von &P1,&P2 verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(max(k1,k2)^1.59) -> Karatsuba
// bzw. Komplexitt: O((k1+k2)*ld(k1+k2)) -> bei (optimaler) dft


/*!
 * @param R result polynomial
 * @param kR size of result polynomial
 * @param P1 first multiplicator polynomial
 * @param k1 size of @p P1
 * @param P2 second multiplicator polynomial
 * @param k2 size of @p P2
 * @result size of the product polynomial @p R
 * @remark
 *	   - This function implements the Karatsuba algorithms
 *         - complexity O(max(k1,k2)^1.59), Karatsuba
 *         - use it for medium and big sized polynomials
 *         - @p &R must be disjoint from @p &P1 and @p &P2
 */
int mul(const TPolynom R, const int kR,
        TconstPolynom P1, int k1,
        TconstPolynom P2, int k2);
// Pres = P1*P2, &R mu von &P1,&P2 verschieden sein
// liefert Grad des resultierenden Polynoms zurck
// Komplexitt: O(max(k1,k2)^1.59) -> Karatsuba


/*!
 * @param R result polynomial
 * @param kR size of result polynomial
 * @param P argument polynomial
 * @param k size of @p P
 * @param m the result will be reduced modulo @p m
 * @param scale (optional, with default 0, no scale)
 * @remark
 *	   - This function computes the reciprocal polynomial of R
 *         - The argument polynomial @p P will be multiplied (shifted) by x^(scale)
 */
void reciprocal(TPolynom R, int &kR, const TconstPolynom P, const int k, const mpz_t m, const unsigned int scale=0);
// berechnet das reziproke Polynom von P,
// R mu gengend Speicher bereitstellen!
// das bergebene Polynom wird zuvor mit x^scale multipliziert,
// also die Koeffizienten um scale Stellen verschoben


/*!
 * @param Q result polynomial (quotient)
 * @param kQ size of result quotient polynomial
 * @param R result polynomial (remainder)
 * @param kR size of result remainder polynomial
 * @param P1 dividend polynomial
 * @param k1 size of @p P1
 * @param P2 divisor polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @remark
 *	   - This function implements the classical division
 *         - complexity O(k1*k2)
 *         - use it only for small sized polynomials
 *         - @p &Q , @p &R , @p &P1 and @p &P2 should be disjoint from each other
 *         - this function is deprecated
 */
void classic_div(TPolynom Q, int &kQ, TPolynom R, int &kR,
         const TconstPolynom P1, int k1,
         const TconstPolynom P2, int k2, const mpz_t m) __attribute__ ((deprecated));


/*!
 * @param R result polynomial (remainder)
 * @param kR size of result polynomial
 * @param P1 dividend polynomial
 * @param k1 size of @p P1
 * @param P2 divisor polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @remark
 *	   - This function implements the classical division
 *         - complexity O(k1*k2)
 *         - use it only for small sized polynomials
 *         - @p &Pr should be disjoint from @p &P1 and @p &P2
 */
void classic_mod(TPolynom R, int &kR,
            const TconstPolynom P1, int k1,
            const TconstPolynom P2, int k2, const mpz_t m);
// Polynomrest der Polynomdivision, O(n^2)
// liefert P1 modulo P2 in R zurck


/*!
 * @param Q result polynomial (quotient)
 * @param kQ size of result polynomial
 * @param P1 dividend polynomial
 * @param k1 size of @p P1
 * @param P2 divisor polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @remark
 *	   - This function implements division using newton method (using reciprocal polynomial)
 *         - complexity depends on fastest multiplication, should be O(max(k1,k2)*ld(max(k1,k2))
 *         - @p &Pr should be disjoint from @p &P1 and @p &P2
 */
void div(TPolynom Q, int &kQ,
         const TconstPolynom P1, const int k1,
         const TconstPolynom P2, const int k2, const mpz_t m);
// (schnelle) Polynomdivision durch Multiplikation mit dem inversen Polynom


/*!
 * @param R result polynomial (remainder)
 * @param kR size of result polynomial
 * @param P1 dividend polynomial
 * @param k1 size of @p P1
 * @param P2 divisor polynomial
 * @param k2 size of @p P2
 * @param m the result will be reduced modulo @p m
 * @remark
 *	   - This function implements division using newton method (using reciprocal polynomial)
 *         - complexity depends on fastest multiplication, should be O(max(k1,k2)*ld(max(k1,k2))
 *         - use it for medium and large sized polynomials
 *         - @p &Pr should be disjoint from @p &P1 and @p &P2
 */
void mod(TPolynom R, int &kR,
         const TconstPolynom P1, int k1,
         const TconstPolynom P2, int k2, const mpz_t m);
// Polynomrest der Polynomdivision, schnelle(re) Methode
// liefert P1 modulo P2 in R zurck
/* asymptotisch schneller als "naives modulo", weil schnelle Multiplikation
   verwendet werden kann. */


/*!
 * @param res result coefficients
 * @param P polynomial to evaluate
 * @param k size of @p P
 * @param array_of_arguments points at which @p P will be evaluated
 * @param size number of points to evaluate
 * @param m the results will be reduced modulo @p m
 * @remark
 *	   - This function implements fast evaluation schemes
 *         - for medium and big size faster than horner evaluation
 */
void multipoint_eval(mpz_t* res,
                     const TconstPolynom P, const int k, 
                     const mpz_t* const array_of_arguments, const int size,
                     const mpz_t m);


/*!
 * @param res result polynomial (coefficients)
 * @param roots_array roots of polynomial
 * @param size number of roots
 * @param m the results will be reduced modulo @p m
 * @result size of result polynomial
 * @remark
 *	   - This function implements fast evaluation schemes
 *         - @p res must be set to NULL when function is called
 */
int construct_polynomial_from_roots
      (TPolynom &res,
       const mpz_t* const roots_array, const int size,
       const mpz_t m);
  // Aufgabe:
  // Erzeugt ein neues Polynom mit den in roots_array gegebenen Nullstellen,
  // legt dieses Polynom in "res" ab und liefert die Gre als Funktionswert
  // zurck. Um sicherzustellen, dass durch versehentliche bergabe eines
  // vorhandenen Polynoms in "res" kein Memory-Leak entsteht, wird zustzlich
  // verlangt, dass "res" beim Aufruf ein Pointer auf NULL ist.
  // additional remark:
  // Sure, it would be possible to return this pointer instead of using a reference
  // parameter, but what if you forget to delete it later? -- Using this technique,
  // the programmer is at least urged to provide a valid container fr our data!

} // namespace polynomial

#endif  // __POLYNOMIAL_HEADER__
