/* Faktorisierungsprogramm
   written 1997-2000, 2002-2004 by Thorsten Reinecke
   version 2.97: 2004-07 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization", 2nd ed.,
   Boston; Basel; Stuttgart: Birkhuser, 1994
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch eine
   Variation des multipolynomialen Quadratischen Siebs versucht.
   
   (vgl. Robert D. Silverman: The Multiple Polynomial Quadratic Sieve,
    Mathematics Of Computation, Vol.48, January 1987, p. 329-339)

   Die dort beschriebene Large Prime Variante verwende ich ebenfalls.
   Zustzlich benutze ich aber Large Primes als dynamische Faktoren zur
   dynamischen Erweiterung der (usrprnglich statischen) Faktorbasis, d.h.
   ich siebe zustzlich mit ihnen. Damit das trotz ihrer Gre effizient
   geschehen kann, benutze ich Priority Queues (z.B. Fibonacci-Heaps).
   Diese Methode ist auf eigenem Mist gewachsen; in der von mir studierten
   Literatur konnte ich keinen Hinweis auf hnliche Methoden finden, die
   die Faktorbasis dynamisch erweitern. -- Zustzlich zu diesem Verfahren
   habe ich mich durch ein als "Double Large Prime"-Variante bekanntes
   Verfahren inspirieren lassen. Dieses ist teilweise implementiert und an
   meine dynamische Erweiterung angepat.
   Eine Zyklensuche zur Konstruktion von vollstndig in der Faktorbasis
   aufgehenden Relationen aus Double Large Primes (LP1,LP2 - LP2,LP3 - ...
   ... - LPk,LP1) ist nunmehr ebenfalls implementiert.
   Weitergehende Verbesserungen sind allerdings noch nicht implementiert:
   Eine Absplittung dynamischer Faktoren aus "triple large primes"
   unter Zuhilfenahme von "double large primes"-Zyklen gerade bei der
   Faktorisierung sehr groer Zahlen drfte z.B. ebenfalls sehr ntzlich
   sein, da sie die dynamische Faktorbasis vergrert...

   Das durch die gesiebten Relationen entstehende lineare Gleichungssystem
   wird dabei simultan mit den eingefgten Relationen gelst.
   Um Speicherplatz zu sparen, die Korrektheit bertragener Relationen beim
   verteilten Sieben einfacher berprfen zu knnen sowie triviale Lsungen
   schneller zu entdecken, habe ich die Relationen normiert:
   Jede Relation hat bei mir nun die Form 
    "enthaltene Faktorbasiselemente mit ungeradem Exponenten" * x^2 = 1 (mod n),
   wobei x den bereits ermittelbaren Quadratwurzelanteil darstellt.
   Mehrfachpotenzen der enthaltenen Faktorbasiselemente werden bei
   (ganzzahliger) Halbierung des Exponenten zu x hinzumultipliziert. -- Wenn
   schlielich eine Relation der Form x^2=1 entstanden ist, so kann ein
   (hoffentlich nichttrivialer) Faktor ber gcd(x-1,n) gefunden werden.

   Es ist auch eine Implementierung elliptischer Kurven zur Ermittlung von
   Faktoren "gemigter Gre" enthalten. Dies ist das einzige Verfahren,
   mit dem sich momentan auch bei sehr groen Zahlen nichttriviale Faktoren
   entdecken lassen (da seine Laufzeit von der Gre des entdeckten Faktors
   abhngig ist).

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 100 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
  
   [ 
     Nachtrag Juni 2003:

     Bei greren Zahlen lassen sich mit den elliptischen Kurven immerhin
     nichttriviale Faktoren absplitten, z.B. als Dauerproze im Hintergrund:
     P190169 (480 Dezimalstellen); entdeckter Faktor:
     9820062895352727932801465820946503529 (37) [ecm,sigma=272319096]

     Fr die Version 2.90 habe ich einige Testlufe mit folgender
     Konfiguration gestartet: 1x Athlon 1.2 Ghz + 1x Athlon 1.75 Ghz
      + 1x Pentium III 1.0 Ghz + Pentium MMX 166 Mhz.
     Einige dieser Rechner waren zeitweise ber WAN (DSL-Anschlu) angebunden;
     elliptische Kurven wurden fr den Testlauf deaktiviert.
     10^93+3 wurde in knapp 20 Stunden faktorisiert.
     10^95+3 bentigte rund 3 Tage.

     Nachtrag September 2003:
     Im kleinen Netzwerk bentigte die Version 2.92 fr die Faktorisierung einer
     107-stelligen Lucaszahl mit dem Siebverfahren ca. 5 Wochen Rechenzeit.
     (Die Faktorisierung wurde dabei als reiner Hintergrundprozess betrieben.)
   ]
   
   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   Fr ein effizienteres Sieben wre vielleicht noch das "Number-Field-Sieve"
   sinnvoll. (Siebverfahren faktorisieren grundstzlich in Abhngigkeit von
   der Gre der zu faktorisierenden Zahl; sie sind besonders effizient,
   wenn die Zahl lediglich zwei Primteiler hat und diese von der Stelligkeit
   her nicht allzuweit voneinander entfernt liegen.)

   Ob das simultane Lsen des linearen Gleichungssystems sinnvoll ist, oder
   ob man vielleicht effizientere Methoden (z.B. Wiedemann-Verfahren) nach
   Abschlu der Siebphase anwendet, hngt im wesentlichen vom Speicherbedarf
   des Gleichungssystems ab. Beim verteilten Sieben hat das simultane Lsen
   des Gleichungssystems zumindest den Vorteil, parallel zur Siebphase abzu-
   laufen. 
*/

/*! @file
 * @brief
 * main implementation file of "Faktorisierungsprogramm"
 */


#ifdef AUTODETECTION
 // autodetection for gcc-compiler
 #warning "autodetection enabled..."
 #ifdef __GNUG__
  #if __GNUC__ > 2
  #else
    #error "Please compile again with a newer version of GCC (version >= 3.2.x, better >=3.4.x)"
  #endif

  #if #cpu (i386)
   #ifndef ASM386
    #warning "activating i386-inline assembler code"
    #define ASM386
    #ifdef ASMATHLON
     #warning "using athlon-inline assembler code"
     #ifndef ASM386_CMOV
     #warning "activating ASM386-CMOV usage, because ATHLON supports it :)"
     #define ASM386_CMOV
     #endif
    #endif
   #endif
  #else
   #ifdef ASM386
    #warning "deactivating i386-inline assembler code"
    #undef ASM386
    #undef ASM386_CMOV
    #undef ASMATHLON
   #endif
  #endif
  #if !defined (unix) && defined(USE_NETWORK)
    #error "network support only for Unix-systems" 
  #endif
 #endif
#endif /* autodetection */

using namespace std;


#ifndef USE_NETWORK
 #ifdef IS_SERVER
  #error "Illegal configuration: SERVER and no USE_NETWORK"
 #endif
#else
 #ifndef IS_STANDALONE
  const int server_port = 12364;
  #ifdef IS_SERVER
   #include <map>
  #endif
  #include "unix_buffer.H"
 #endif
#endif

#ifndef IS_SERVER
 #ifndef IS_CLIENT
  #undef IS_STANDALONE
  #define IS_STANDALONE
 #endif
#endif

#ifdef IS_SERVER
 #ifdef IS_CLIENT
  #error "Illegal Configuration SERVER-CLIENT"
 #else
  #ifdef IS_STANDALONE
   #error "Illegal Configuration SERVER-STANDALONE"
  #endif
 #endif
#endif

#ifdef IS_CLIENT
 #ifdef IS_STANDALONE
  #error "Illegal Configuration CLIENT-STANDALONE"
 #endif
#endif

#if defined(USE_FIBHEAP) && defined(USE_FAKEHEAP)
 #error "defines USE_FIBHEAP and USE_FAKEHEAP exclude each other!"
#endif

#ifdef REACT_ON_SIGUSR
 // a little class that catches signal SIGUSR1 and SIGUSR2
 // and stores these flags...
 // can be useful to abort some long-running algorithms
 // without aborting the program or altering configuration parameters.
 #include "usr_signals.cc"
 Cusr_signal_proxy USRSignalHandler;
#endif

extern "C"
{
 #include <unistd.h>
}

#include <cstdio>
#include <cstdlib>
#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <gmp.h>
#include "myBitString.H"
#include <list>
#include <set>
#include "utils.H"
#include <ctime>


inline unsigned long int mpz_remainder_ui_(const mpz_t n, const unsigned long int d)
{
  // should speedup some easy cases...
  if (d==1 || d==2 || d==4 || d==8 || d==16 || d==32 || d==64 || d==128)
   {
     unsigned long int h = (n->_mp_d[0]);
     return (n->_mp_size > 0) ? h&(d-1)
                              : (n->_mp_size < 0) ? d-(h&(d-1)) : 0;
   }
  else
   return mpz_fdiv_ui(n,d);
}


inline unsigned long int mpz_remainder_ui(const mpz_t n, const unsigned long int d)
{
  // refer gmp manual:
  // "unsigned long int mpz_fdiv_ui(mpz_t n, unsigned long int d)" returns
  // the remainder of the division; except for not needing any mpz_t result
  // value it is identical to
  // "unsigned long int mpz_mod_ui(mpz_t t, mpz_t n, unsigned long int d)".
  
  // I find the name "mpz_fdiv_ui" too much ambiguous.
  // Defining "unsigned long int mpz_mod_ui(mpz_t n, unsigned long int d)"
  // would be somewhat more straight forward, but it is also risky, because
  // one might assume a real gmp-function and not a wrapper function.
  // mpz_remainder_ui can easily be detected as a wrapper function just
  // by grepping the sources...

  // As this is a wrapper, we can add some optimizations, too
 
  return (__builtin_constant_p(d)) ? mpz_remainder_ui_(n,d) : mpz_fdiv_ui(n,d);
}


#if defined(mpz_odd_p)
 // replace mpz_odd_p / mpz_even_p macros with inline-C++
 // (better type checking + more explicite casts) 

#undef mpz_odd_p
#undef mpz_even_p

//! equivalent to mpz_odd_p macro, but more strict
inline int mpz_odd_p(const mpz_t z)
{
  return static_cast<int>(z->_mp_size != 0) & static_cast<int>(z->_mp_d[0]);
}

//! equivalent to mpz_even_p macro, but more strict
inline int mpz_even_p(const mpz_t z)
{
  return !mpz_odd_p(z);
}

#endif


struct myintcompare {
  inline const bool operator() (const int i1, const int i2) const
   {
     return i1 < i2;
   }
};
typedef set<int, myintcompare> intset;


#include <vector>
#include <stack>
#include <algorithm>

#ifdef USE_NETWORK
 extern "C"
 {
  #include <sys/utsname.h>
  #include <pthread.h>
 }
 #include <csignal>
 #include "mutex.H"
 #ifdef IS_CLIENT
  extern "C"
  {
   #include <sys/wait.h>
   #include <unistd.h>
  }
 #endif
#endif

#ifdef NOTIFY_PARENT
#if defined(IS_SERVER) && defined(USE_NETWORK) 
 #include <csignal>
#else
 #warning "undefining NOTIFY_PARENT, since this is not a networked server"
 #undef NOTIFY_PARENT
#endif
#endif


// output to screen/terminal can optionally be handled by a ncurses based
// wrapper, but on default all windows will be merged into cout.
#ifdef USE_NCURSES
 #include <my_ncurses.H>
#endif

ostream cout_titel(cout.rdbuf());
ostream cout_network(cout.rdbuf());
ostream cout_status(cout.rdbuf()); 


const string RecoveryFile         = "recovery.dat";
const string FactorizationsFile   = "factorizations.txt";
const string StaticRelationsFile  = "static_relations.dat";
const string SpecialRelationsFile = "special_relations.dat";
      string DynamicRelationsFile = "dynamic_relations.dat"; // default, may be overwritten by entry in configfile
      string ConfigFile           = "qsieve.cfg"; // default, may be overwritten by getenv("QSIEVE_CFG")


// Streams fr die Speicherung der Relationen und sonstiger Daten
// zur Verfgung stellen
// *************************************************************************
// **** wichtig:
// **** Wenn ostream und istream durch denselben filebuf fest miteinander
// **** gekoppelt sind, darf/kann man sich *nicht* darauf verlassen, dass
// **** tellp() und tellg() voneinander unabhngige Funktionen sind!
// **** *Jede* Schreiboperation kann dann *auch* den Wert von tellg()
// **** verndern und *jede* Leseoperation den von tellp()!
// ************************************************************************* 

// Fr die Relationen der dynamischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf DynamicRelations_buffer;
ostream DynamicRelations_to_file (&DynamicRelations_buffer);
istream DynamicRelations_from_file (&DynamicRelations_buffer);

// Fr Recovery
filebuf Recovery_buffer;
ostream Recovery_to_file (&Recovery_buffer);
istream Recovery_from_file (&Recovery_buffer);

#ifdef IS_CLIENT
string communication_name;
ofstream communication_stream;
#endif

#ifndef IS_CLIENT
// Fr das Speichern der gemachten Faktorisierungen
ofstream Factorization_to_file;
/* Bemerkung:
   Auf diesem Stream werden alle mit diesem Programm gemachten Faktorisierungen
   protokolliert. Die zu faktorisierenden Zahlen sowie alle gefundenen Faktoren
   werden angehngt.
   Das zugehrige File sollte im Bedarfsfall gelscht werden, wenn die Daten
   nicht mehr bentigt werden.
*/
#endif


typedef unsigned short int FBsizetype; // Typ der Faktorbasisgre
  // Wenn die Faktorbasis maximal 65535 Elemente haben soll,
  // sollte man "unsigned short int" whlen, um Speicherplatz zu sparen.
  // Wenn Speicherplatz keine Rolle spielt, ist "int" eine gute Wahl (denn
  // dann gibt es u.U. weniger implizite Typkonvertierungen).
  // Hinweis:
  // "FBsizetype" hat *nur* mit der Anzahl der Primzahlen in der statischen
  // Faktorbasis (also den Indices auf die darin enthaltenen Primzahlen),
  // aber *nichts* mit dem Typ (und der Gre) der Primzahlen aus der
  // statischen Faktorbasis selbst zu tun!

const FBsizetype MaxSize_StaticFactorbase = 30000; /* Maximale Gre der Primzahlbasis */ 
/* Hinweis: Je grer die zu faktorisierende Zahl ist, desto grer
   sollte prinzipiell auch die Primzahlbasis sein, da dann die Relationen
   einfacher gefunden werden knnen.
   Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
   den Speicher!
   Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
   Primzahlbasis verlangsamend aus, da fr ein greres Gleichungssystem
   natrlich auch mehr Relationen gesiebt werden mssen.
   Je nach Geschwindigkeit des Rechners, des verfgbaren Haupt- und
   Hintergrundspeichers (und auch dessen Zugriffszeiten) knnen die
   ttschlich zu verwendenden Werte stark variieren. Sie sind deshalb
   experimentell zu ermitteln.
   Die Maximalgre wird durch MaxSize_StaticFactorbase festgelegt und steht
   damit bereits zur Kompilierzeit fest.
   Bei einer Speicherkapazitt von 32 MB (realem) Hauptspeicher
   hat sich 20000 als brauchbarer Wert erwiesen.
*/

#include "Tfactor.H"
#include "qsieve.H"
#include "mpqsPolynom.H"


const int SingleLargePrime_Threshold = 1000000000;
// bis hierhin werden gefundene Faktoren als SingleLarge-Primes akzeptiert
// und in die Dynamic-Factor-Menge aufgenommen

const int DynamicFactor_SievingThreshold = MIN(SingleLargePrime_Threshold,60000000);
// bis hierhin wird mit den SingleLarge-Primes auch (aktiv) gesiebt.
// (Der Wert mu nicht notwendigerweise niedriger als SingleLargePrime_Threshold sein,
// aber wenn er kleiner ist, dann git es in den Siebroutinen fr Dynamische Faktoren
// einen CutOff an dieser Stelle.)


struct TSpecialFactorRelation {
  CmpqsFactor factor;
  streampos fpos;

  inline bool operator() (const TSpecialFactorRelation &t1, const TSpecialFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
};


class TDynamicFactorRelation
{
  /* Bemerkung:
     Falls factor nicht allzugro ist (im Augenblick ist dies als Obergrenze der
     int-Bereich), dann kann auch mit den Dynamic-Factors gesiebt werden.
     Nachfolgend die dazu erforderlichen Daten...
  */
public:
  int factor; // factor in int-Darstellung
  int SQRT_kN;
  streampos fpos;
  inline bool operator() (const TDynamicFactorRelation &t1, const TDynamicFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
  inline const bool sieveable() const
   {
     // are we allowed to sieve with this Large Prime?
     return factor<DynamicFactor_SievingThreshold;
   }
};


struct TSieve_Delta
{
  int delta;
  int factor;
  inline bool operator< (const TSieve_Delta &v) const
    {
#if defined (USE_FIBHEAP)
      return (delta < v.delta);
#else
      return (delta > v.delta);
#endif
      // delta > v.delta , if queue is used
      // delta < v.delta , if fibonacci-heap is used
    }
};

#ifndef IS_SERVER
#ifdef USE_FIBHEAP
 #include "fibheap.H"
 FibHeap <TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
 #include <queue>
#elif defined(USE_FAKEHEAP)
 #include "fakeheap.H"
 FakeHeap <TSieve_Delta> Sieve_Delta_Heap;
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_HeapOfSquares;    
#else
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
#endif
#endif

mpz_t n, /* zu faktorisierende Zahl */
      kN, /* das mit einem gnstigen Vorfaktor multiplizierte n
	     (nachdem n um einfach zu findende Faktoren reduziert wurde) */
      r; /* Quadratwurzel von n (oder bei optimierter Fassung k*n) */

int Vorfaktor_von_kN; // der noch zu ermittelnde Vorfaktor fr kN (kN=Vorfaktor_von_kN*n)


FBsizetype Size_StaticFactorbase = 5000; // default fr zu verwendende Gre der Primzahlbasis
/*  Constraint for AMD-3DNow! specific code: Size_StaticFactorbase%2 == 0  (because of SIMD)
    Sinnvolle Werte fr Faktorbasen liegen zwischen 1000 und 50000
    (je nach verfgbarem Speicher und der zu faktorisierenden Zahl).
    Der Wert wird normalerweise spter durch die Prozedur "tune_parameters"
    angepat.
*/

int PrimeNumbers[MaxSize_StaticFactorbase]; /* fr eine effiziente Probedivision
				       alle Primzahlen der Faktorbasis hier
				       akkumulieren */


int SQRT_kN_of_PrimeNumbers[MaxSize_StaticFactorbase];
int Delta_of_PrimeNumbers[MaxSize_StaticFactorbase][2]; /* fr den effizienten Zugriff
						    werden die pro Siebintervall berechneten
						    Werte hier abgelegt */
 
const int FB_max_additional_Powers = 1000; // maximale Anzahl weiterer Primzahlpotenzen, mit denen gesiebt wird
int NumberOf_more_PrimePowers = 0; // tatschliche Anzahl (wird im Programm ermittelt)
int FB_maxQuadrate = 0; /* tatschliche Anzahl der Quadrate,
			   mit denen normal gesiebt wird (wird im Programm ermittelt) */
int PrimePowers[FB_max_additional_Powers];
int SQRT_kN_of_PrimePowers[FB_max_additional_Powers];
int Delta_of_PrimePowers[FB_max_additional_Powers][2];
int SQRT_kN_of_PrimeSquares[MaxSize_StaticFactorbase];


class StaticRelations
{
 // this class is just here to encapsulate access to the functions
 // which are specific for handling static relations

 private:
  static CRelation* GLS[MaxSize_StaticFactorbase];
   // Gleichungssystem besteht aus Zeigern auf Relationen (Gleichungszeilen),
   // das mit Null-Pointern initialisiert ist, da anfangs noch keine Relationen
   // vorliegen
  static int Fuellung_GLS; // Fllung des Gleichungssystems mit Relationen

  // Fr die Relationen der statischen Faktorbasis
  // streams fr das Datenfile zur Verfgung stellen:
  static filebuf FileBuffer;
  static ostream StaticRelations_to_file;
  static istream StaticRelations_from_file;

 public:
  static void insert(CRelation *GL, const bool do_multi_combine_init=true);
  static void Load();

  static const int Count() { return Fuellung_GLS; }
  static void cleanup_files()
   {
     FileBuffer.close();
     remove(StaticRelationsFile.c_str());
   }
  static void cleanup_memory()
   {
     int count = 0;
     for (int i=0; i<MaxSize_StaticFactorbase; ++i)
      if (GLS[i]) { delete GLS[i]; ++count; }
     Fuellung_GLS=0;
     cout << count << " relations deallocated." << endl;
   }
  friend int main(const int argc, const char* const argv[]);
};
CRelation* StaticRelations::GLS[MaxSize_StaticFactorbase] = {NULL};
int StaticRelations::Fuellung_GLS = 0;
filebuf StaticRelations::FileBuffer;
ostream StaticRelations::StaticRelations_to_file(&StaticRelations::FileBuffer);
istream StaticRelations::StaticRelations_from_file(&StaticRelations::FileBuffer);


/* Zur Erluterung der Begrifflichkeit von Dynamic-Factors/Special-Factors:
   Ich verwende momentan eine modifizierte Multi-Large-Prime-Variante:
   Dynamic-Factors entsprechen der Large-Prime-Variante. Aber ich siebe mit jedem neu entdeckten
   dynamischen Faktor zustzlich...
   Special-Factors sind zusammengesetzte Restfaktoren (wie sie in der Double-Large-Prime-Variante
   auftreten). Es wird einerseits versucht, Special-Factors durch neu aufgenommene
   Dynamic-Factors zu splitten, um daraus weitere Dynamic-Factors zu gewinnen.
   Andererseits wird eine Zyklensuche auf diesen durchgefhrt, um weitere
   Relationen zu gewinnen.
   In einer spteren Implementierung knnte man die Zyklensuche noch
   auf Hyperkantenzyklen hherer Large-Primes erweitern.
*/

typedef set<TDynamicFactorRelation,TDynamicFactorRelation> TDynamicFactorRelations;

#ifdef IS_SERVER
class CServerDynamicFactorRelations : public TDynamicFactorRelations
{
 private:
  mutable CMutex monitor_mutex;
  vector<int> DynamicFactorList;
 public:
  inline void insert(const TDynamicFactorRelation& x)
   {
     if (TDynamicFactorRelations::insert(x).second)
      {
        if (!x.sieveable()) return; // x is a Double Large Prime, but a passive one: not sieveable
        monitor_mutex.lock();
        DynamicFactorList.push_back(x.factor);
        monitor_mutex.unlock();
      }
     else
      {
        cerr << "CServerDynamicFactorRelations::insert duplicate (rejected)!" << endl;
        exit(1);
      }
   }
  inline const int monitoredSize() // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int s=DynamicFactorList.size();
     monitor_mutex.unlock();
     return s;
   }
  inline const int operator[] (const int i) // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int x=DynamicFactorList[i];
     monitor_mutex.unlock();
     return x;
   }
};
CServerDynamicFactorRelations DynamicFactorRelations;
#else
TDynamicFactorRelations DynamicFactorRelations;
#endif

double Factor_Threshold = 3.0; /* Exponent, der grere Faktoren ber die
				  Faktorbasis hinaus erlaubt.
				  Sinnvolle Werte: 1.0 bis 3.0
				  1.0: keine zustzlichen Faktoren auerhalb der Faktorbasis 
				  bis 2.0: Primfaktoren auerhalb der Faktorbasis bis
				  zum Quadrat des grten Faktors der FB
				  ber 2.0: Faktoren bis zu grter Faktor der Faktorbasis
				  hoch Factor_Threshold.
			       */


class SpecialRelations
{
 // this class is just here to encapsulate access to the functions
 // which are specific for handling special relations
 // (and special factors (=DLP, double large prime factors outside the static factorbase)

 private:
  typedef set<TSpecialFactorRelation,TSpecialFactorRelation> TSpecialFactorRelations;
  static TSpecialFactorRelations SpecialFactorRelations;

  // Fr die Relationen zu den Specialfactors
  // streams fr das Datenfile zur Verfgung stellen:
  static filebuf FileBuffer;
  static ostream SpecialRelations_to_file;
  static istream SpecialRelations_from_file;


 public:
  static const int Count() { return SpecialFactorRelations.size()/2; }
  static void cleanup_files()
   {
     FileBuffer.close();
     remove(SpecialRelationsFile.c_str());
   }
  static void CycleSearch();
  static bool insert(const CmpqsFactor &DLP, CRelation *GL, const short int HitCount=0);
  static bool insert(const CmpqsFactor &DLP, const string &GL_String);
  static bool SpecialFactor_splitable(const CmpqsFactor &DLP);
  static void split_by_primefactor(const int Primfaktor);
  static void Load();
  friend int main(const int argc, const char* const argv[]);
};
SpecialRelations::TSpecialFactorRelations SpecialRelations::SpecialFactorRelations;
filebuf SpecialRelations::FileBuffer;
ostream SpecialRelations::SpecialRelations_to_file(&SpecialRelations::FileBuffer);
istream SpecialRelations::SpecialRelations_from_file(&SpecialRelations::FileBuffer);


const int PhysicalSieveSize = 64*1024; // =65536
/* In (physikalischen) Teilintervallen von "PhysicalSieveSize" werden die Relationen
   im (logischen) Siebintervall [-M,M] (siehe auch Deklaration von M) gesiebt.
   Es handelt sich deshalb um eine bereits zur Compilierungszeit feststehende Konstante.

   Hinweis: Je grer "PhysicalSieveSize", desto effizienter geht das Aussieben
   der Relationen. Gleichzeitig erhht sich aber auch der Speicherbedarf. 
   Der Einflu der Siebgre ist allerdings geringer als der der gewhlten
   Primzahlbasis. Ein Wert um 1 Million drfte daher vollkommen ausreichen. 
   Andererseits kann mglicherweise im Cachespeicher des Rechners gesiebt
   werden, wenn die Siebgre klein gewhlt wird. Kleine Siebgren erhhen
   zwar den rechnerischen Overhead, aber dafr ist der Zugriff auf das Sieb
   mglicherweise deutlich schneller. Cachemisses knnen je nach Level
   einzelne Zugriffe um den Faktor 10 bis 200 verteuern! Deshalb sollte man
   auch eine Siebgre um 100000 (bei 512K Cache) ausprobieren.

   Fr eventuellen optimierten Zugriff ber Assemblerbefehle sollte
   PhysicalSieveSize*TSIEBELEMENTSIZE ein Vielfaches der Maschinenwortgre (bei
   32-Bit-Rechnern also 4, bei 64-Bit-Rechnern 8; beim AMD-Athlon wegen der
   MMX-Untersttzung und wegen der Cacheline: 64) sein.  Idealerweise whle
   man ein Vielfaches von 1024, dann ist man auf der sicheren Seite.  (Das
   stellt keine wirkliche Einschrnkung dar, ermglicht es aber, das Sieb
   effizient zu initialisieren ohne Sonderflle beachten zu mssen. -- Wenn
   Assembler-Code aktiviert wird, verlasse ich mich auf diesen Constraint!)
*/

  
int LogicalSieveSize = 1000000; // das zu jedem Polynom gehrige Siebintervall ist [-LogicalSieveSize,LogicalSieveSize]
/*
   Constraint for AMD-3DNow! specific code: abs(LogicalSieveSize)<=2^23 (because of float precision)

   LogicalSieveSize ist ein im Unterschied zu "PhysicalSieveSize" ein
   logischer Wert, der weniger stark von der verwendeten Hardware abhngig
   ist. Die logische Intervallgre ist vor allem abhngig von der Gre der
   zu faktorisierenden Zahl zu whlen.  Der hier festgelegte Wert ist also
   lediglich ein Defaultwert, der von der Konfiguration vor Beginn eines
   konkreten Faktorisierungslaufs (zur Laufzeit) modifiziert werden kann
   (und soll).
   Bemerkung: In multiple-polynomial-quadratic-sieve-Varianten sollte man
   mglichst kleine Siebintervalle whlen, da die relative Trefferquote von
   Relationen mit der Intervallgre abnimmt. (In 10 kleinen Intervallen ist
   die Chance, eine Relation zu finden, hher als in einem 10mal so groen
   Intervall).
   Gegen kleine Intervalle spricht jedoch, dass Kosten fr das Wechseln der
   Polynome, Neuinitialisierungen und Wurzelberechnungen anfallen.
*/


int SieveOffset = -LogicalSieveSize;
/* Im allgemeinen wird das fr ein Polynom zu siebende Siebintervall 
   grer sein als die (physikalisch optimale) Siebgre.
   Das Siebintervall ist daher in Teilstcken zu sieben. Hierfr wird der
   Sieboffset bentigt, der den Startpunkt des jeweiligen Teilsiebs angibt.
 */

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// !!! WICHTIG !!! #define und typedef mssen einander entsprechen !!!
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#define TSIEBELEMENTSIZE 1 /* 1 = signed char, 2 = short signed int, 4 = signed int, 0 = no assembler inline code (=generic code) */
// obiges define sorgt bei Verwendung der ASM386-Direktive fr inline-assembler-code in der Siebroutine

#if TSIEBELEMENTSIZE == 0
 typedef signed char TSieveElement; // must be signed!
 #define log_Multiplikator 1 /* Siebwert wird nach damit nach TSiebelement konvertiert */
 #ifdef ASM386
  #warning "assembly code is disabled for sieving!"
 #endif
#elif TSIEBELEMENTSIZE == 1
 typedef signed char TSieveElement; // must be signed!
 #define log_Multiplikator 1 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#elif TSIEBELEMENTSIZE == 2
 typedef signed short int TSieveElement; // must be signed!
 #define log_Multiplikator 16 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#elif TSIEBELEMENTSIZE == 4
 typedef signed int TSieveElement; // must be signed!
 #define log_Multiplikator 256 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#else
 #error "invalid value for TSIEBELEMENTSIZE"
#endif

/* Je kleiner der Typ von TSieveElement ist, desto 
    a) kleiner ist der fr das Sieb bentigte Speicher,
    b) schneller kann deshalb gesiebt werden.
   Allerdings ist die Granularitt zum Erkennen von Relationen bei
   kleineren Datentypen grber und deshalb schlechter.
   Auerdem sind prozessorspezifische Effekte (z.B. Alignment-Penalties)
   zu bercksichtigen!
   In Frage kommen:
    - signed char (1 Byte)
    - short signed int (2 Byte)
    - signed int (4 Byte)
    - float, double (Fliekommatypen, mehrere (>3) Bytes)
   Die Bytegren knnen zwischen Systemen variieren. Ich habe aber nicht
   geprft, ob das Programm berhaupt auf mit einer Systemeinstellung von
   int < 4 Byte, short int < 2 Byte lauffhig ist.

   Wichtiger Hinweis: Wenn man Fliekommatypen als TSieveElement verwenden will (was ich nicht mehr empfehle),
   dann sind im Programm die "ceil"-Aufrufe fr TSieveElement kontraproduktiv und sollten entfernt werden.
   Aber es gibt keinen vernnftigen Grund mehr, mit Fliekommatypen zu sieben.
*/

// #define log_Multiplikator <VALUE> /* Siebwert wird nach damit nach TSiebelement konvertiert */
/*
  -> refer TSIEBELEMENTSIZE
  Da die Vorkommastellen der natrliche Logarithmen von Primzahlen
  zu hohe Ungenauigkeiten beim Sieben verursachen, werden sie noch
  mit log_Multiplikator multipliziert, um eine hhere Genauigkeit
  zu erreichen.
  Der Multiplikator kann auf 1 gesetzt sein, wenn TSieveElement Fliekommatyp
  ist. Ansonsten sollte man ihn mglichst gro whlen, und zwar so, dass
  der Threshold der blicherweise zu faktorisierenden Zahlen noch in den
  Datentyp TSieveElement pat.
  Damit durch Abrundungen keine Relationen durch die Lappen gehen, werden
  aber beim Sieben grundstzlich aufgerundete Werte abgezogen, so dass ein
  kleiner Multiplikator hier allenfalls ein zu engmaschiges Sieb erzeugt.
  Die exakten Auswirkungen verschiedener Einstellungen wren allerdings
  noch genauer zu untersuchen.
  Wenn man nicht genau wei, was man tut, sollte man eher TSieveElement
  "hhertypen". Der Multiplikator sollte dann deutlich ber 1 liegen (10 - 1000).
*/


inline TSieveElement log_DynamicLargePrime (const int p)
{
#if defined(ASM386) && TSIEBELEMENTSIZE == 1 && log_Multiplikator == 1
#warning "assembler i386 TSIEBELEMENT log lookup enabled"
  // we assume 8103<p<=1318815734 (=e^9)
  register TSieveElement logval;
  asm ( \
       "movb $21,%[L] \n\t" \
       "cmpl $485165195,%[P] # exp(20) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $178482300,%[P] # exp(19) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $65659969,%[P] # exp(18) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $24154952,%[P] # exp(17) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $8886110,%[P] # exp(16) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $3269017,%[P] # exp(15) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $1202604,%[P] # exp(14) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $442413,%[P] # exp(13) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $162754,%[P] # exp(12) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $59874,%[P] # exp(11) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $22026,%[P] # exp(10) \n\t" \
       "sbbb $0,%[L]" \
       : [L] "=&q" (logval) : [P] "r" (p) : "cc");
#ifdef DEBUG
  // remark: do not use std::log(), as it might be ambiguous for some
  // versions of g++ (-> see utils.H)
  if (logval!=ceil(log_Multiplikator*log(p)))
   {
     MARK;
     cout << "P: " << p << " log:" << static_cast<int>(logval) << "  " << ceil(log_Multiplikator*log(p)) << endl;
     exit(1);
   }
#endif
  return logval;
#else
  return static_cast<TSieveElement>(ceil(log_Multiplikator*log(p)));
#endif
}


TSieveElement SieveArray_[PhysicalSieveSize+64+8] __attribute__ ((aligned (64))); // Speicher fr das Sieb
TSieveElement* const SieveArray = &SieveArray_[64];
// SieveArray[-64..-1] are -1 (for fast value-based end-of-loop range detections)
// SieveArray[0..PhysicalSieveSize-1] contain actually sieve values
// SieveArray[PhysicalSieveSize..PhysicalSieveSize+8] contain dummy values


TSieveElement log_Primzahl_of_PrimePowers[FB_max_additional_Powers];
/* fr weitere Primzahlpotenzen stehen hier die konvertierten Logarithmen der Primzahlen */

bool collecting_phase_finished = false; /* switch, der die "aktive" und "passive" Phase des Siebens
					  bestimmt:
					  Wenn gengend Special-Relations und Dynamic-Relations gesammelt
					  sind, um ausreichend viele Static-Relations zu erhalten, wird
					  das Aufsammeln weiterer Special- und Dynamic-Relations durch
					  das Setzen von collecting_phase_finished=true unterbunden. */

int biggest_Prime_in_Factorbase;

CmpqsPolynom Polynom; // Objekt zur Verwaltung der Polynomberechnung fr multipolynomiales Sieb


// use wrapped C++ stream I/O for mpz_t (multiple precision numbers)
#include "mpz_wrapper.H"
using namespace my_mpz_wrapper;

#include "mpqsStatistics.cc" // statistical stuff about found relations, client connections, etc.


class SieveControl
{
 private:
  static const int PrimzahlStartIndex = 5;
  /* ab diesem Primzahlindex wird gesiebt.
     Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
     der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
     beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
     so einfach) gesiebt werden.
     Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
     grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
     viele Relationen durchs Sieb fallen...
   */

  inline static const double DirtyFactor(const int i)
   {
     if (i<PrimzahlStartIndex+4) return 1.0;
     return 1.0/sqrt(i); // irgendeine langsam gegen 0 strebende Funktion
   }
  /* Der DirtyFactor gibt die Wichtung an, mit der die i-te Primzahl der Faktorbasis,
     die "dirty" sind, als Siebtreffer auf das Sieb aufgeschlagen werden.
     Beispiel:
      Wenn p normal siebt, gibt es im Sieb an jeder p-ten Stelle einen Treffer, der
      voll gewichtet wird, z.B. mit log(p).
      Wenn p "dirty" ist, gibt es an *allen* Siebstellen einen Treffer fr p, der
      mit DirtyFactor skaliert ist, also z.B. DirtyFactor*log(p). Erst wenn
      spter eine potentielle Relation erkannt wird, dann wird der hierdurch
      verursachte Fehler korrigiert.
     Ein DirtyFactor von 1.0 verhlt sich gegenber einem normalen Siebvorgang
     bis auf Rundungseffekte neutral, beschleunigt aber den Siebvorgang.
     Je kleiner der DirtyFactor ist, desto mehr Primzahlen knnen im Laufe des
     Regelungsvorgangs "dirty" werden. (Die Erkennungsrate pro Siebintervall
     sinkt dann allerdings zugunsten der Siebgeschwindigkeit.)
     Der DirtyFactor 1/n verhlt sich grob gesprochen etwa so, als wrde jede
     n-te der "dirty" Primzahlen "automatisch" in jeder Relation vorhanden sein.
   */  

  static TSieveElement log_PrimeNumbers[MaxSize_StaticFactorbase];
  static TSieveElement log_PrimeNumbers_dirty[MaxSize_StaticFactorbase];
    // Hier stehen die konvertierten Logarithmen der Primzahlen aus der Faktorbasis

  static TSieveElement log_Threshold; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert
  static int _sieben_ab_Primzahlindex;

  static signed int DirtynessPegel;

public:
  inline static const TSieveElement PrimzahlSiebWichtung(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck; blicherweise ist
     // dies der (skalierte) Logarithmus, der in der Tabelle abgelegt wurde. 
     // Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers[i];
   }
  inline static const TSieveElement PrimzahlSiebWichtung_dirty(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck wenn diese als dirty
     // gekennzeichnet ist, d.h. auf ein Sieben mit dieser Zahl verzichtet
     // wird; dieser Wert wird dann fr alle Elemente des Siebs einheitlich
     // als Treffer gewichtet (statt tatschlich zu sieben); blicherweise
     // ist dies ein (skalierter) Logarithmus, der in der Tabelle abgelegt
     // wurde. Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers_dirty[i];
   }
private:
  static void IncreaseDirtyness(void)
   {
     // Threshold absenken und Index fr erste zu siebende Zahl inkrementieren
     if (_sieben_ab_Primzahlindex>(Size_StaticFactorbase>>2)) return; // das geht leider nicht...
     log_Threshold-=PrimzahlSiebWichtung_dirty(_sieben_ab_Primzahlindex);
     ++_sieben_ab_Primzahlindex;
     cout << "Dirtyness now " << _sieben_ab_Primzahlindex << endl;
   }

  static void DecreaseDirtyness(void)
   {
     // Threshold hochsetzen und Index fr erste zu siebende Zahl dekrementieren
     if (_sieben_ab_Primzahlindex<=PrimzahlStartIndex) return; // das geht leider/wollen wir nicht...
     log_Threshold+=PrimzahlSiebWichtung_dirty(--_sieben_ab_Primzahlindex);
     cout << "Dirtyness now " << _sieben_ab_Primzahlindex << endl;
   }

  static void RequestDirtyness(const signed int Level)
   {
     DirtynessPegel+=Level;
     if (DirtynessPegel>1000000)
      {
        DirtynessPegel=0;
        IncreaseDirtyness();
      } else
     if (DirtynessPegel<-1000000)
      {
        DirtynessPegel=0;
        DecreaseDirtyness();
      }
   }

 public:
  static void compute_SieveThreshold(void);
  static const TSieveElement GetLogThreshold(void) { return log_Threshold; }
  static const int sieben_ab_Primzahlindex(void) { return _sieben_ab_Primzahlindex; }
  static const bool Hit_after_DirtyCorrection(const signed int Siebpos, TSieveElement Value);

  static inline void RequestMoreDirtyness(const signed int Level = 1000)
   {
     RequestDirtyness(Level);
   }
  static inline void RequestLessDirtyness(const signed int Level = 50)
   {
     RequestDirtyness(-Level);
   }

  friend void compute_StaticFactorbase();
  friend void initialize_Sieve(void);
};
TSieveElement SieveControl::log_PrimeNumbers[MaxSize_StaticFactorbase];
TSieveElement SieveControl::log_PrimeNumbers_dirty[MaxSize_StaticFactorbase];
int SieveControl::_sieben_ab_Primzahlindex = SieveControl::PrimzahlStartIndex;
signed int SieveControl::DirtynessPegel = 0;


inline bool is_dynamic_factor(const int number)
{
  // returns, whether this number is already registered as a dynamic factor;
  // (dynamic factor -> prime number outside the static factorbase, also known as Single Large Prime)
  TDynamicFactorRelation relation;
  relation.factor = number;
  const TDynamicFactorRelations::iterator pos = DynamicFactorRelations.find(relation); // dynamischen Faktor suchen
  return (pos != DynamicFactorRelations.end());
}


#include "modulo.cc" // Modulo-Operationen fr unsigned int
                     // als .cc included fr optimierende Compiler
using namespace numtheory;


// einige Parameter fr easy-Faktorisierung (defaults)
bool SkipFermat    = false; // Fermatmethode versuchen
bool SkipPhi       = false; // Pollard-Phi-Methode versuchen
bool SkipFibonacci = false; // Fibonacci-Methode versuchen
bool SkipEasyECM   = false; // elliptic curve method im easy-mode versuchen
bool SkipECM       = false; // elliptic curve method versuchen
int rho_Phase      =   150000; // Grundinitialisierung fr Rho-Phase
int phi_Phase1     =   500000; // fr phi-Methode Phase 1
double phi_Phase2  = 10000000.0; // fr phi-Methode Phase 2 (=isolierter Faktor)
int elcu_Phase1    =    20000; // fr elliptische Kurven Phase 1
double elcu_Phase2 =  2000000.0; // fr elliptische Kurven Phase 2
int elcu_Kurven    =      100; // Anzahl der zu probierenden elliptischen Kurven 

const int probab_prime_checks = 25; // #rounds for mpz_probab_prime


#include "Tfactor.cc"


#ifndef IS_CLIENT

void SpecialRelations::CycleSearch()
{
  // Wenn man einen Zyklus von Large-Primes durch Double-Large-Primes
  // erzeugen kann, so neutralisieren sich die beteiligten Large-Primes und
  // man erhlt eine vollstndig in der statischen Faktorbasis aufgehende
  // Relation.

  if (SpecialFactorRelations.empty()) return; // es gibt nichts zu finden
  cout << "Starting cycle find for Special-Factor-Set" << endl;

  TSpecialFactorRelations arbeitsmenge;
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;

  arbeitsmenge.swap(SpecialFactorRelations);
  // Hinweis: um nicht doppelten Speicherbedarf zu haben, werden die
  // Special-Factor-Relations in die Arbeitsmenge geschoben (und nicht kopiert)
  // es ist dann darauf zu achten, dass sie von der Arbeitsmenge auch wieder
  // zurckgeschoben (und nicht nur gelscht) werden...
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;

  intset knotenmenge;
  typedef list<TSpecialFactorRelation> Tkantenliste;
  Tkantenliste kantenliste;

  TSpecialFactorRelations::iterator pos;
  TSpecialFactorRelation SearchRelation;
  int rechts,links;

  while (!arbeitsmenge.empty())
   {
     knotenmenge.clear(); // wichtig! (wegen Schleifenwiederholungen)
     pos=arbeitsmenge.begin();
     TSpecialFactorRelation relation = *pos;
     kantenliste.push_back(relation);
     links=relation.factor.LP1();
     rechts=relation.factor.LP2();
     knotenmenge.insert(links);
     knotenmenge.insert(rechts);
     arbeitsmenge.erase(pos); SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();

     //cout << "Starte neue Runde der Zyklensuche." << endl;
     while (!kantenliste.empty())
      {
        //for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
        // cout << (*p).factor << "|";
        //cout << endl;
        //cout << "Suche nach " << links << endl;
	//cout << "Arbeitsmenge enthlt: " << endl;
        //for (pos=arbeitsmenge.begin(); pos!=arbeitsmenge.end(); ++pos)
        // cout << (*pos).factor << ",";
	//cout << endl;

        SearchRelation.factor.set_for_search(links);
        pos=arbeitsmenge.lower_bound(SearchRelation);
        if (pos!=arbeitsmenge.end() && (*pos).factor.DLP_divisible_by(links) )
         {
           // links kann eine Kante eingefgt werden
           relation = *pos;
           arbeitsmenge.erase(pos); // alten Specialfactor entfernen
           SpecialFactorRelations.insert(relation);
           relation.factor.swap();
           arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
           SpecialFactorRelations.insert(relation);
           relation.factor.swap();
           links=relation.factor/links;
           if (knotenmenge.find(links)!=knotenmenge.end())
            { 
              // Zyklus gefunden!
	      statistical_data::gefundene_DLP_Zyklen++;
	      cout << "DLP-cycle found... " << endl;
              // nun noch: 
              // Zyklus auswerten und erhaltene Relation ins Gleichungssystem
              // einfgen.
              CRelation *GL = new CRelation;
              GL->combine(SpecialRelations_from_file,relation.fpos);
              mpz_mul_ui(GL->Delta,GL->Delta,links);
	      mpz_mod(GL->Delta,GL->Delta,n);
              rechts=relation.factor/links;
	      //cout << relation.factor;

              // Kantenliste bis zu der Position, wo der Wert "links"
              // erstmalig enthalten ist, durchgehen -> dort endet der Zyklus.
	      // dabei noch die Wurzel des Produkts der Specialfaktoren
              // hinzumultiplizieren (beim Zyklus sind dies alle beteiligten
              // Large Primes in einfacher Potenz), also die Werte "rechts"
	      unsigned int Zyklenlaenge=1;
              for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
	       {
                 Zyklenlaenge++;
		 //cout << " - " << p->factor;
                 mpz_mul_ui(GL->Delta,GL->Delta,rechts);
	         mpz_mod(GL->Delta,GL->Delta,n);
                 GL->combine(SpecialRelations_from_file,(*p).fpos);
                 if ((*p).factor.DLP_divisible_by(links)) break;
		 rechts=(*p).factor/rechts;
               }
              //cout << endl;
              cout << "DLP-cycle size: " << Zyklenlaenge << endl;
	      GL->SanityCheck(); // aus Sicherheitsgrnden...
     	      StaticRelations::insert(GL);
	    
  	      // Anschlieend eine verantwortliche (wir whlen die letzte Kante)
              // aus der Special-Relation-Menge entfernen:
              SpecialFactorRelations.erase(relation); // alten Specialfactor entfernen
              relation.factor.swap();
              SpecialFactorRelations.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
              relation.factor.swap();
              {
	        // nicht vergessen, die Relation auf File zu markieren!
                const streampos pos = SpecialRelations_to_file.tellp(); // Schreibposition merken
                SpecialRelations_to_file.seekp(relation.fpos); SpecialRelations_to_file << "U " << flush; // ungltig markieren
                SpecialRelations_to_file.seekp(pos); // wieder an die alte position
              }
 
              // Schlielich: Zyklensuche fortsetzen.
	      links=relation.factor/links;
              continue;
            }
           else
            {
              // kein Zyklus gefunden -> Kante und Knoten einfgen
              knotenmenge.insert(links);
              kantenliste.push_front(relation);
	      //cout << "Fge Kante " << relation.factor << " hinzu " << endl;
              continue;
            } 
         }
        // keine Kante zum linksseitigen Einfgen gefunden
        // -> Alternative 1: rechts einfgen (zum Zyklusfinden nicht notwendig)
        // -> Alternative 2: linksseitig eine Kante entfernen
	//cout << "Entferne Kante " << kantenliste.begin()->factor << endl;
        links=kantenliste.begin()->factor/links;
        kantenliste.pop_front(); // also linksseitig eine Kante entfernen
      }
   }
  cout << "Size: " << SpecialFactorRelations.size() << ", " << arbeitsmenge.size() << endl;
  cout << "cycle find finished." << endl;
}

bool SpecialRelations::insert(const CmpqsFactor &DLP, CRelation *GL, const short int HitCount /* =0 */)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren,
  // so dass man GL anschlieend ins Gleichungssystem eintragen kann.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat!" << endl;
     statistical_data::Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  const TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  if (pos==SpecialFactorRelations.end())
   {
     relation.fpos=GL->save(SpecialRelations_to_file,relation.factor,HitCount);
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
        SpecialRelations::CycleSearch();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit by already known Special Factor!" << endl;
     statistical_data::Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.in_mpz_ablegen(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->combine(SpecialRelations_from_file,(*pos).fpos);
     //StaticRelations::insert(GL); // sollte an anderer Stelle gemacht werden!
     return false;
   }
}


bool SpecialRelations::SpecialFactor_splitable(const CmpqsFactor &DLP)
{
  // Testen, ob Special-Factor erfolgreich gesplittet werden kann
  // true -> ja, false -> nein
  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     // cout << "Special case: DLP is a square!" << endl;
     return true;
   }
  TSpecialFactorRelation relation;
  relation.factor=DLP;
  TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  return (pos!=SpecialFactorRelations.end());
}


void Read_ConfigFile(void)
{
  ifstream in(ConfigFile.c_str());
  char eingabezeile[200];
  string zeile;
  if (in)
    {
      cout << "Reading configuration file" << endl; 
      while (!in.eof())
	{
	  in.getline(eingabezeile,sizeof(eingabezeile),'\n');
	  if (eingabezeile[0]=='#') continue; // Kommentarzeilen berlesen
	  zeile=eingabezeile;
          while (zeile[0]==' ') zeile.erase(0,1);
          if (zeile.empty() || zeile[0]=='#') continue; // leere Zeile/Kommentarzeile berlesen

          // Keywords verarbeiten
          istringstream is(zeile);
          string Keyword;
          is >> Keyword;
          if (Keyword=="DynamicRelationsFile")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> DynamicRelationsFile;
             continue;
           }
          if (Keyword=="WorkDir")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> s;
             cout << "Setting current working directory to \"" << s << "\"." << endl;
	     if (chdir(s.c_str())!=0)
              {
                cerr << "Changing working directory failed!" << endl;
                exit(2);
              };
             continue;
           }
          if (Keyword=="SkipFermat")
           {
	     SkipFermat=true;
             continue;
           }
          if (Keyword=="SkipPhi")
           {
	     SkipPhi=true;
             continue;
           }
          if (Keyword=="SkipFibonacci")
           {
	     SkipFibonacci=true;
             continue;
           }
          if (Keyword=="SkipEasyECM")
           {
	     SkipEasyECM=true;
             continue;
           }
          if (Keyword=="SkipECM")
           {
	     SkipECM=true;
             continue;
           }

        }
    }
  else
    {
       static char ch = 'N';
       cout << ConfigFile << " not found! Using default values!" << endl;

       if (ch!='Y')
         {
           cout << "Bitte mit [Y]es besttigen!" << endl;
           cout << "Please confirm with [Y]es." << endl;
           cin >> ch;
           if (ch!='Y') exit(1);
         }
    }

  char cwd[4096];
  cout << "Current working directory: " << getcwd(cwd,4096) << endl;
  cout << "Configuration file: " << ConfigFile << endl;
  cout << "DynamicRelationsFile: " << DynamicRelationsFile << endl;
  if (!Factorization_to_file.is_open()) Factorization_to_file.open(FactorizationsFile.c_str(),ios::out|ios::app);
}

void tune_parameters(const unsigned int Dezimalstellen)
{
  ifstream in(ConfigFile.c_str());
  char eingabezeile[200];
  string zeile;
  unsigned int Stellen, gewaehlte_Stellenzahl=0;
  if (in)
    {
      cout << "Reading configuration file" << endl; 
      while (!in.eof())
	{
	  in.getline(eingabezeile,sizeof(eingabezeile),'\n');
	  if (eingabezeile[0]=='#') continue; // Kommentarzeilen berlesen
	  zeile=eingabezeile;
          while (zeile[0]==' ') zeile.erase(0,1);
          if (zeile.empty() || zeile[0]=='#') continue; // leere Zeile/Kommentarzeile berlesen

          // Keywords verarbeiten
          istringstream is(zeile);
          string Keyword;
          is >> Keyword;
	  if (Keyword=="SkipFermat") continue;
	  if (Keyword=="SkipPhi") continue;
	  if (Keyword=="SkipFibonacci") continue;
          if (Keyword=="SkipEasyECM") continue;
	  if (Keyword=="SkipECM") continue;
          if (Keyword=="DynamicRelationsFile" || Keyword=="WorkDir")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             // is >> DynamicRelationsFile; // hier nicht (mehr) einlesen!
             continue;
           }

          // ansonsten: Zeile enthlt Tuning-Parameter
	  Stellen = atoi(zeile.substr(0, zeile.find(":")).c_str());
	  if (Stellen > Dezimalstellen || Stellen==0) break;
          gewaehlte_Stellenzahl=Stellen;
	  zeile.erase(0, zeile.find(":")+1);
	  rho_Phase=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase2=strtod(zeile.substr(0, zeile.find(",")).c_str(),NULL);
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase2=strtod(zeile.substr(0, zeile.find(",")).c_str(),NULL);
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Kurven=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  Size_StaticFactorbase=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.erase(0, zeile.find(":")+1);
          Factor_Threshold = atof(zeile.substr(0, zeile.find(":")).c_str());
          if (Factor_Threshold<=0.01 || Factor_Threshold>10.0)
           {
             MARK;
             cerr << "invalid value of Factor_Threshold: " << Factor_Threshold << endl;
             exit(1);
           }
	  zeile.erase(0, zeile.find(":")+1);

	  LogicalSieveSize = atoi(zeile.c_str());
	  LogicalSieveSize = MAX(LogicalSieveSize,PhysicalSieveSize);
	}
    }
  else
    {
       static char ch = 'N';
       cout << ConfigFile << " not found! Using default values!" << endl;
       if (ch!='Y')
         {
           cout << "Bitte mit [Y]es besttigen!" << endl;
           cout << "Please confirm with [Y]es." << endl;
           cin >> ch;
           if (ch!='Y') exit(1);
         }
    }
  
  cout << "Configuration:" << endl;
  cout << "#decimal digits: " << Dezimalstellen << endl;
  cout << "actual parameters for " << gewaehlte_Stellenzahl << " digits" << endl;
  cout << "Pollard-Rho: " << rho_Phase << endl;
  cout << "Pollard-Phi: phase 1: " << phi_Phase1 << " phase 2: " << setprecision(1) << phi_Phase2 << endl;
  cout << "elliptic curves: phase 1: " << elcu_Phase1
       << " phase 2: " << elcu_Phase2
       << " #curves: " << elcu_Kurven << endl; 
  cout << "Size of static factorbase: " << Size_StaticFactorbase << endl;
  cout << "factor-threshold (exponent): " << Factor_Threshold << endl;
  cout << "sieve interval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;

  if (Size_StaticFactorbase>MaxSize_StaticFactorbase)
    {
      cerr << "MaxSize_StaticFactorbase is too small (or Size_StaticFactorbase is to big)!" << endl;
      exit(1);
    }
}
#endif /* von #ifndef IS_CLIENT */


const char* const MAL(void)
{
  // liefert Leerstring bei erstem Aufruf, " * " sonst.
  // Aufgabe: Verhinderung aufwendiger Tests, um den Faktor 1 als
  // ersten bzw. letzten Faktor unterdrcken zu knnen.

#ifdef NOTIFY_PARENT
  // send SIGUSR2 to the parent to notify him that a new
  // factor has found
  kill(getppid(),SIGUSR2); // send SIGUSR2 to the parent process
#endif

  static bool erster_Aufruf = true;
  if (erster_Aufruf)
   {
     erster_Aufruf=false;
     return "";
   }
  else return " * ";
}

string MAL(const mpz_t factor, const unsigned int exponent=1)
{
  ostringstream helper;
  helper << MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str();
}


#include "polynomial.H" /* for fast polynomial arithmetic & fft-continuation */
#include "fft_param.cc"  /* discrete fast fourier transform */ 
#include "elliptic_curve.cc" /* elliptic curve method (ecm) */

#include "ExitManager.cc" /* controlled termination of program */

#ifndef IS_CLIENT
 // this stuff may only be needed by server or stand-alone version
 #include "easy_factor.cc"
#endif /* of #ifndef IS_CLIENT */



#include "sqrt_modulo.cc"
#include "mpz_sqrtmod.cc"

inline int SQRT_kN_mod_PrimeNumber(const unsigned int Primzahl)
{
  return sqrtmod(mpz_remainder_ui(kN,Primzahl),Primzahl);
}


int check_SQRT_kN_mod_PrimeNumber(const int Primzahl)
{
  // kann zu Debugging-Zwecken aufgerufen werden, um zu testen,
  // inwieweit die "SQRT_kN_mod_PrimeNumber"-Funktion korrekt arbeitet
  int w1 = SQRT_kN_mod_PrimeNumber(Primzahl);
  if (w1>Primzahl/2) w1=Primzahl-w1;

  mpz_t x;
  mpz_init_set_ui(x,w1); mpz_mul(x,x,x); mpz_sub(x,x,kN);
  if (mpz_mod_ui(x,x,Primzahl)!=0)
    {
      cerr << "WURZELALARM! " << w1 << " mod " << Primzahl << endl;
      exit(1);
    }
  mpz_clear(x);
  return w1;
}


void compute_StaticFactorbase()
{
#if defined(ASMATHLON) || defined(ASM_P3_SSE)
  // ATHLON specific sanity checks
  if ( LogicalSieveSize>(1<<23) || (Size_StaticFactorbase&3) )
   {
     MARK;
     cerr << "Sorry! Processor specific constraints not met." << endl;
     cerr << "For using AMD-3DNow! or SSE optimized version we need:" << endl;
     cerr << "Constraint 1 (SIMD): (Size_StaticFactorbase&3)==0" << endl;
     cerr << "Constraint 2 (float precision): LogicalSieveSize<=(1<<23) [=8388608]" << endl;
     cerr << "Please edit ConfigFile according to these constraints" << endl;
     cerr << "OR use a version without these optimizations." << endl;
     exit(1);
   }
#endif

  cout << "Calculating static factorbase." << endl;
  PrimeNumbers[0]=-1; /* nullte "Primzahl" ist -1 (Bercksichtigung des
		       Vorzeichens bei den Relationen! */
  SieveControl::log_PrimeNumbers_dirty[0]=SieveControl::log_PrimeNumbers[0]=0; // Vorzeichen hat keinen Logarithmuswert

  int Primzahl=2;
  // die Primzahl 2 ist ein Sonderfall. Ich gehe davon aus, dass sie
  // auf jeden Fall in der Faktorbasis vorhanden sein soll.
  PrimeNumbers[1]=2;
  SieveControl::log_PrimeNumbers_dirty[1]=SieveControl::log_PrimeNumbers[1]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)));

  Primzahl=1; // Initialwert 1 nur, wenn in der Schleife +=2 gezhlt wird!
  mpz_t x,y,wuqu;
  mpz_init(x); mpz_init(y); mpz_init(wuqu); // Fr Berechnung der Quadratwurzeln von Primzahlpotenzen
  for (int Primzahl_nr=2; Primzahl_nr<Size_StaticFactorbase; Primzahl_nr++)
    {
      do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; while(!is_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(x,Primzahl);
	}
      while (mpz_legendre(kN,x)<0); /* Primzahl gltig fr Primzahlbasis? */
      
      //printf("%i. Primzahl der Faktorbasis ist %i\r",Primzahl_nr,Primzahl);
      PrimeNumbers[Primzahl_nr]=Primzahl;
      
      // fr effizientes Sieben erforderlich (erspart dauernde Neuberechnung)
      SieveControl::log_PrimeNumbers[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)));
      SieveControl::log_PrimeNumbers_dirty[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)*SieveControl::DirtyFactor(Primzahl_nr)));
      
      // fr sptere Ermittlung der Deltawerte zum Sieben erforderlich
      SQRT_kN_of_PrimeNumbers[Primzahl_nr]=SQRT_kN_mod_PrimeNumber(Primzahl);
      
      // ggf. Prfung, ob Wurzel korrekt ist:
      if ( squaremod(SQRT_kN_of_PrimeNumbers[Primzahl_nr],Primzahl) != mpz_remainder_ui(kN,Primzahl) )
       {
         cerr << "Wurzel nicht korrekt!" << endl;
         cerr << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << "^2 <> "
              << mpz_remainder_ui(kN,Primzahl) << " (mod " << Primzahl << ")" << endl;
         exit(1);
       }


      if ( Primzahl < sqrt(std::numeric_limits<int>::max()) && Vorfaktor_von_kN%Primzahl!=0 )
       {
          // *** Wurzeln von kN modulo Primzahl^2 ermitteln ***
          const unsigned int P = Primzahl*Primzahl;
	  const unsigned int Wu = SQRT_kN_of_PrimeNumbers[Primzahl_nr];
          unsigned int iy = invmod(2*Wu,P);
          unsigned int ix = mpz_remainder_ui(kN,P);
          
	  unsigned int wuq=Wu*Wu;
          if (ix>=wuq) ix-=wuq; else ix=ix-wuq+P; // ix-=wuq; berlauf verhindern; wir rechnen modulo P
          ix=mulmod(ix,iy,P)+Wu;
          wuq = (ix<P) ? ix : ix-P; // schnellere Version fr: wuq=ix%P

#if 1 || defined(DEBUG)
          if (mulmod(wuq,wuq,P)!=mpz_remainder_ui(kN,P))
           {
             cerr << "PQ-Wurzel nicht korrekt!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << endl;
             cerr << "kN= " << kN << endl;
	     cerr << mulmod(Wu,Wu,P) << " != " << mpz_remainder_ui(kN,P) << endl;
             exit(1); 
           }
#endif
          SQRT_kN_of_PrimeSquares[Primzahl_nr]=wuq;
       } else SQRT_kN_of_PrimeSquares[Primzahl_nr]=-1;


      const int Threshold = LogicalSieveSize < std::numeric_limits<int>::max()/8 ? 8*LogicalSieveSize/Primzahl : LogicalSieveSize/Primzahl;
      if (Primzahl<=Threshold && Primzahl>2 && mpz_remainder_ui(kN,Primzahl)!=0)
	// Primzahl darf nicht zu gro sein, ungleich 2 und kein Teiler von kN, um mit Potenzen gesiebt zu werden
	{
	  FB_maxQuadrate=Primzahl_nr; /* nach Abschlu der Schleife ist dies der Wert, bis zu dem
					 Quadrate auf normale Weise gesiebt werden. */

	  // Wurzeln von kN modulo Primzahl^pot ermitteln
	  // Initialwerte:
	  int P_Potenz = Primzahl*Primzahl;
	  int Wu       = SQRT_kN_of_PrimeNumbers[Primzahl_nr];

          double Wichtung = Primzahl_nr<SieveControl::sieben_ab_Primzahlindex() ? 2.0 : 1.0;
           // mit den ersten Primzahlen der Faktorbasis wird (weil sie zu klein sind) gar nicht
           // gesiebt; dafr erhlt die kleinste Potenz dieser Zahl, die hier aufgenommen wird,
           // den vollen Siebwert der Potenz: log(P_Potenz)=Potenz*log(Primzahl).
           // Alle brigen Potenzen entsprechen dem nochmaligen Auftreten eines schon bekannten
           // Faktors und knnen deshalb nur in einfacher Zhlung (=log(Primzahl) bercksichtigt werden.

	  for (int pot=2; ;pot++)
	    {
	      //cout << "berechne Wurzel fr " << Primzahl << "^" << pot << endl;

	      mpz_set_ui(y,2*Wu); mpz_set_ui(x,P_Potenz);
	      if (mpz_invert(y,y,x)==0)
		{ 
		  cerr << "PPot-Wurzel fr " << Primzahl  << ": Inverses existiert nicht!" << endl;
		  break;
		}

              if (NumberOf_more_PrimePowers>=FB_max_additional_Powers)
               { 
                 static char weitermachen = 'n';
                 if (weitermachen!='y')
                  {
	           cerr << "berlauf ab " << Primzahl << "^" << pot << endl;
                   cerr << "Es kann nicht optimal mit Primzahlpotenzen gesiebt werden," << endl
                        << "da der hierfr reservierte Speicherplatz nicht ausreicht." << endl;
                   cerr << "Sie sollten \"FB_max_additional_Powers\" vergroessern!" << endl;
                   cerr << "Reserved memory for storing prime powers is too small..." << endl;
                   cerr << "sieving will be less efficient..." << endl;
                   cerr << "continue? (y/n) " << flush;
                   cin >> weitermachen;
                   if (weitermachen=='n') exit(1);
                  }
                 break; // berlauf durch Schleifenabbruch verhindern
               }

	      mpz_mod_ui(x,kN,P_Potenz); mpz_set_ui(wuqu,Wu); mpz_mul_ui(wuqu,wuqu,Wu); mpz_sub(x,x,wuqu);
	      mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod_ui(x,x,P_Potenz);
	      Wu = mpz_get_ui(x); if (Wu>P_Potenz-Wu) Wu=P_Potenz-Wu; // auf die "kleinste" Wurzel normieren
              // Wu ist nun die Quadratwurzel von kN (mod P_Potenz)

	      // ggf. Prfung, ob Wurzel korrekt ist:
	      mpz_set_ui(x,Wu); mpz_mul(x,x,x); mpz_sub(x,kN,x); 
	      if (mpz_mod_ui(x,x,P_Potenz)!=0)
		{
		  cerr << "PPot-Wurzel nicht korrekt!" << endl;
		  cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << " Nr. " << Primzahl_nr << endl;
		  cerr << "kN= " << kN << endl;
		  cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << endl;
		  cerr << "SQRT(kN) mod p^i=" << SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers] << endl;
		  cerr << "aber falsch: " << x << endl;
		  exit(1);
		}

              // Heuristik: Fr die letzte Potenz die Wichtung etwas erhhen,
              // da hier eine gewisse Chance besteht, dass der Treffer auch Treffer noch hherer Potenzen ist
              if (P_Potenz>Threshold) Wichtung+=4.0/Primzahl;

#if 1
              // Heuristik (im Versuch gemessen/ermittelt):
              // Bei "kleinen" Zahlen geht das Sieben deutlich schneller, wenn man die kleineren
              // Zahlen wegfallen lsst; das Sieb wird dadurch allerdings durchlssiger, aber das strt nicht,
              // da die Relationen nur so reinpurzeln...
              // Bei greren Zahlen werden aber bei genauerem Sieben in der gleichen Zeit mehr DLP erkannt, und
              // das ist dort der wichtige Effekt.
              if ( P_Potenz<250 && mpz_sizeinbase(kN,10)<70 ) Wichtung+=1.0;
              else
#endif
               {
                 // weil das Sieben mit kleinen Zahlen viel Zeit bentigt und ein Siebtreffer wenig
                 // bringt, verlangen wir eine Mindestgre, damit die Potenz beim Sieben
                 // bercksichtigt wird.

	         PrimePowers[NumberOf_more_PrimePowers]=P_Potenz;
	         SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers]=Wu;

		 log_Primzahl_of_PrimePowers[NumberOf_more_PrimePowers]
                   =static_cast<TSieveElement>(ceil(log_Multiplikator*Wichtung*log(Primzahl)));
                 Wichtung=1.0;
                 ++NumberOf_more_PrimePowers; // dies war eine weitere aufgenommene Primzahlpotenz
               }
	      
	      if (P_Potenz>Threshold) break; // nchste Primzahlpotenz wird zu gro...
	      P_Potenz*=Primzahl;
	    }
	}
    }
  mpz_clear(x); mpz_clear(y); mpz_clear(wuqu);

  biggest_Prime_in_Factorbase=Primzahl;
  cout << "prime numbers in static factorbase have been computed." << endl;
  
  if (Primzahl>=PhysicalSieveSize)
    cout << "Remark: Some members of the static factorbase are bigger than the sieve size!" << endl;
}


#ifndef IS_SERVER
void compute_Deltas_for(const TDynamicFactorRelation &DynRel, const int offset)
{
  TSieve_Delta sieb_delta;
  int d1,d2;
  
  sieb_delta.factor=DynRel.factor;

  const unsigned int inv_A2 = invmod(mpz_remainder_ui(Polynom.get_A2(),DynRel.factor),DynRel.factor);
  const unsigned int PolyB = mpz_remainder_ui(Polynom.get_B(),DynRel.factor);

  d1=DynRel.SQRT_kN-PolyB; if (d1<0) d1+=DynRel.factor;
  d1=mulmod(d1,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d1-offset,DynRel.factor);
  sieb_delta.delta+=offset;
#ifdef USE_FAKEHEAP
  while (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
   {
     Sieve_Delta_Heap.push(sieb_delta);
     sieb_delta.delta+=sieb_delta.factor;
   }
#else
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
    Sieve_Delta_Heap.push(sieb_delta);
#endif
  
  d2=DynRel.factor-DynRel.SQRT_kN-PolyB; if (d2<0) d2+=DynRel.factor;
  d2=mulmod(d2,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d2-offset,DynRel.factor);
  sieb_delta.delta+=offset;
  if ( (sieb_delta.delta<=LogicalSieveSize)  && (d1!=d2) ) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
   {
     Sieve_Delta_Heap.push(sieb_delta);
#ifdef USE_FAKEHEAP
     sieb_delta.delta+=sieb_delta.factor;
     while (sieb_delta.delta<=LogicalSieveSize)
      {
        Sieve_Delta_Heap.push(sieb_delta);
        sieb_delta.delta+=sieb_delta.factor;
      }
#endif
   }
}
#endif


// Forward Declarations
int GetValue(istream &in);
void PutValue(ostream &out, const int w);


#ifndef IS_CLIENT
void SpecialRelations::split_by_primefactor(const int Primfaktor)
{
  // Specialfaktoren liegen auerhalb der Primzahlbasis.
  // Mit ihnen wird nicht gesiebt. Sie sind auerdem zusammengesetzt.
  // Wenn es gelingt, von einem Specialfaktor einen Primfaktor aus der dynamischen FB
  // abzusplitten, dann bestehen gute Chancen, dadurch einen neuen dynamischen Faktor
  // zu erhalten oder gar einen Treffer im Gleichungssystem, wenn der entstandene
  // dynamische Faktor bereits in der dynamischen Faktorbasis bekannt ist.

  // Wenn ein neuer dynamischer Primfaktor entdeckt wird, so empfiehlt es sich,
  // in der Menge der Specialfaktoren (welches zusammengesetzte Faktoren sind) einen
  // Teilbarkeitstest auf diesen Primfaktor durchzufhren. Auf diese Weise knnen
  // weitere Primfaktoren isoliert werden, die als dynamische Faktoren zur Verfgung stehen.
  // Im Gegensatz zur bisherigen BUSY-Methode verwandelt die LAZY-Methode
  // den abgesplitteten Primfaktor in einen Korrekturfaktor. Die eigentliche
  // Auswertung der Relation wird damit auf einen spteren Zeitpunkt verschoben
  // und mglicherweise ganz verhindert, falls die betroffene Relation nicht
  // bentigt wird.

  static int Trashcounter = 0; /* Die (nach Elimination eines Faktors) nutzlos auf dem File
				  befindlichen Relationen */
  
  TSpecialFactorRelation SearchRelation;
  SearchRelation.factor.set_for_search(Primfaktor);

  while (true)
   {
    //cout << "splitten " << Primfaktor << endl;
    /* Wegen Rekursion etc. immer wieder korrekt neu aufsetzen, wenn
       ein Specialfaktor gekegelt wurde -> nchsthherer Special-Factor */
    const TSpecialFactorRelations::iterator pos=SpecialFactorRelations.lower_bound(SearchRelation);
    if (pos==SpecialFactorRelations.end()) break;
    if (!(*pos).factor.DLP_divisible_by(Primfaktor)) break;

    TSpecialFactorRelation gemerkte_Relation = *pos;
    SpecialFactorRelations.erase(pos); // remove old specialfactor
    gemerkte_Relation.factor.swap();
    SpecialFactorRelations.erase(gemerkte_Relation); // and remove also his swapped version
    gemerkte_Relation.factor.swap(); // swap back to original
    
    TDynamicFactorRelation relation;
    relation.factor=Primfaktor;
    const TDynamicFactorRelations::iterator p_el = DynamicFactorRelations.find(relation);
    if (p_el==DynamicFactorRelations.end()) 
      {
        MARK; 
        cerr << "Inconsistency in dynamic factorbase!" << endl;
        exit(1);
      }

    relation.factor=gemerkte_Relation.factor/Primfaktor; // dies ist der neue Restfaktor
    // now we have a correctly computed dynamic factor at hand

    // but is it really a new one, or did it exist already the dynamic factorbase?
    const TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(relation);
    if (p==DynamicFactorRelations.end())
     { 
       // er ist neu!
       //cout << "Special Factor: Dynamischer Faktor erzeugt!" << endl;
       statistical_data::Special_to_dynamic_Factor_hit++;
       relation.fpos = DynamicRelations_to_file.tellp();
       // nun die Relation modifiziert umkopieren:
       {
         const streampos rpos=SpecialRelations_from_file.tellg(); // Position merken
         SpecialRelations_from_file.seekg(gemerkte_Relation.fpos); // Position merken
         CmpqsFactor ret; string s;
         // Dateiformatflags retten
         const ios::fmtflags oldFlagsR = SpecialRelations_from_file.flags(); // alte Formatflags retten
         SpecialRelations_from_file.unsetf(ios::hex); SpecialRelations_from_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         const ios::fmtflags oldFlagsW = DynamicRelations_to_file.flags(); // alte Formatflags retten
         DynamicRelations_to_file.unsetf(ios::hex); DynamicRelations_to_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
         SpecialRelations_from_file >> s; // Startzeichen
         if (s != "G")
          {
	   MARK;
           cerr << "error: wrong position while reading relation" << endl;
           exit(1);
          }
         SpecialRelations_from_file >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
         SpecialRelations_from_file >> s; // delta
         DynamicRelations_to_file << "G " << setprecision(20) << relation.factor << " " << s << " ";

         // ab jetzt wird hexadezimal gearbeitet:
         SpecialRelations_from_file.unsetf(ios::dec); SpecialRelations_from_file.setf(ios::hex); // Hexadezimalausgabe
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         DynamicRelations_to_file.unsetf(ios::dec); DynamicRelations_to_file.setf(ios::hex); // Hexadezimalausgabe
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung

         int distance=GetValue(SpecialRelations_from_file);
         while (distance>=0)
          {
            PutValue(DynamicRelations_to_file,distance);
            distance=GetValue(SpecialRelations_from_file);
          }
         PutValue(DynamicRelations_to_file,-2);
         PutValue(DynamicRelations_to_file,Primfaktor);
         if (distance==-2)
          {
            distance=GetValue(SpecialRelations_from_file);
            while (distance>=0)
             {
               PutValue(DynamicRelations_to_file,distance);
               distance=GetValue(SpecialRelations_from_file);
             }            
          }
         PutValue(DynamicRelations_to_file,-1); DynamicRelations_to_file << endl;
         SpecialRelations_from_file.flags(oldFlagsR); // alte Formatflags wiederherstellen
         DynamicRelations_to_file.flags(oldFlagsW); // alte Formatflags wiederherstellen
         SpecialRelations_from_file.seekg(rpos); // position restaurieren
       }
       relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
#if !defined(IS_SERVER) && !defined(USE_FAKEHEAP)
       compute_Deltas_for(relation, SieveOffset+PhysicalSieveSize);
#endif
       DynamicFactorRelations.insert(relation); // Dynamicfactor einfgen
       SpecialRelations::split_by_primefactor(relation.factor);
     }
    else 
     { 
       // er existierte schon in der dyn. FB -> er fhrt zu einem Treffer im GLS!
       //cout << "Special Factor: Special Hit!" << endl;
       CRelation *GL = new CRelation;
       GL->combine(SpecialRelations_from_file,gemerkte_Relation.fpos);
       // gefundene dynamische Faktoren kombinieren:
       mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
       GL->combine(DynamicRelations_from_file,(*p).fpos);
       // nicht vergessen: auch Primfaktor eliminieren!
       mpz_mul_ui(GL->Delta,GL->Delta,(*p_el).factor); mpz_mod(GL->Delta,GL->Delta,n);
       GL->combine(DynamicRelations_from_file,(*p_el).fpos);
       statistical_data::Special_hit++; StaticRelations::insert(GL);
     } 

    { 
      // remove old (and superfluous) special-relation
      const streampos pos = SpecialRelations_to_file.tellp(); // save position
      SpecialRelations_to_file.seekp(gemerkte_Relation.fpos); SpecialRelations_to_file << "U "; // mark as invalid
      SpecialRelations_to_file.seekp(pos); // restore position
    }

    ++Trashcounter; // Die auf dem File verbleibenden sinnlosen Relationen zhlen
   }

 // Wenn sich sehr viel Mll angesammelt hat, knnte man nun eine Meldung
 // erstatten. Aber das drfte sich bei den heutigen Festplattengren
 // kaum lohnen; da nervt die Meldung mehr als sie ntzt...
}
#endif



#ifdef ASM_P3_SSE /* special code using SSE instructions (P3) */

void GetStaticFactorHits(vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )

  // (Pentium3) SSE specific code

  #warning "new Pentium3/SSE specific code"
  typedef int v4sf __attribute__ ((mode(V4SF)));
  register v4sf XMM6 asm("xmm6"); // XMM6 is now bound to %xmm6 (intention: signedness correction)
  register v4sf XMM7 asm("xmm7"); // XMM7 is now bound to %xmm7

   // refer GCC manual, "Specifying registers for local variables";
   // but remember: that does not necessarily mean, that %xmm7 cannot be
   // used for other variables as well! (It is only guaranteed, that if we access
   // XMM7, this access is synonym to accessing %xmm7; but accessing %xmm7 is not necessarily
   // synonym to accessing XMM7 (because if the compiler assumes, that XMM7 is dead, %xmm7 could
   // be used for other variables; and the compiler does *not* know, that we want
   // XMM7 to be alive by accessing %xmm7!).

  asm ( \
       "emms                       # enter mmx\n\t" \
       "movd %[Sieboffs],%%mm0   # -      mm0=(0,Sieboffs)\n\t" \
       "xorps %[_xmm6],%[_xmm6]    # L2AM   clear mm6 \n\t" \
       "punpckldq %%mm0,%%mm0  # L2AM   (0,Sieboffs) -> (Sieboffs,Sieboffs) \n\t" \
       "cvtpi2ps %%mm0,%[_xmm7]  # packed int to float \n\t" \
       "shufps $0x44,%[_xmm7],%[_xmm7] \n\t" \
       "cmpnleps %[_xmm7],%[_xmm6] # L2AM   set bits if signed, clear bits if unsigned" \
       : [_xmm7] "=x" (XMM7), [_xmm6] "=x" (XMM6) : [Sieboffs] "g" (Siebpos) : "mm0");

  // PrimeNumbers[0] und PrimeNumbers[1] werden ausgelassen (da sie gesondert behandelt werden mssen)
  // PrimeNumbers[2] und PrimeNumbers[3] werden hier gesondert behandelt (damit der restliche Anzahl durch 4 teilbar ist)
  {
     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[2]);
     if (d==Delta_of_PrimeNumbers[2][0] || d==Delta_of_PrimeNumbers[2][1]) Factors.push_back(2);
     d = normalized_signed_mod(Siebpos,PrimeNumbers[3]);
     if (d==Delta_of_PrimeNumbers[3][0] || d==Delta_of_PrimeNumbers[3][1]) Factors.push_back(3);
  }
  
  for (int nr=4; nr<Size_StaticFactorbase; nr+=4) 
    {
      register unsigned int pflags, pflags2;
      asm ( \
       "movq %[divisors](,%[i],4),%%mm4   # L2AMS  get y : x (as packed int) \n\t" \
       "movq 8+%[divisors](,%[i],4),%%mm5   # L2AMS  get y : x (as packed int) \n\t" \
       "cvtpi2ps %%mm4,%%xmm0 # convert packed int to float \n\t" \
       "cvtpi2ps %%mm5,%%xmm1 # convert packed int to float \n\t" \
       "movlhps %%xmm1,%%xmm0 \n\t" \
       "movaps %[_xmm7],%%xmm2 \n\t" \
       "movaps %%xmm0,%%xmm1 \n\t" \
       "divps %%xmm0,%%xmm2 \n\t" \
       "movaps %%xmm1,%%xmm3 \n\t" \
       "movhlps %%xmm2,%%xmm4 \n\t" \
       "cvttps2pi %%xmm2,%%mm0 \n\t" \
       "cvttps2pi %%xmm4,%%mm1 \n\t" \
       "movaps %[_xmm7],%%xmm2 \n\t" \
       "cvtpi2ps %%mm0,%%xmm0 \n\t" \
       "cvtpi2ps %%mm1,%%xmm4 \n\t" \
       "movlhps %%xmm4,%%xmm0 \n\t" \
       "mulps %%xmm1,%%xmm0 \n\t" \
       "subps %%xmm0,%%xmm2 \n\t" \
       "andps %[_xmm6],%%xmm1      # correct sign (part 1) \n\t" \
       "addps %%xmm1,%%xmm2        # correct sign (part 2) \n\t" \
       "movaps %%xmm2,%%xmm0 \n\t" \
       "cmpneqps %%xmm3,%%xmm0      #  assure modp=0 instead of modp=p \n\t" \
       "andps %%xmm0,%%xmm2   \n\t" \
       "movhlps %%xmm2,%%xmm4 \n\t" \
       "cvttps2pi %%xmm2,%%mm0 \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd %[Deltas](,%[i],8),%%mm1   # L2AM \n\t" \
       "pcmpeqd 8+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags]   # L6+ (VectorPath)  into bytemask \n\t" \
       "cvttps2pi %%xmm4,%%mm0 \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd 16+%[Deltas](,%[i],8),%%mm1 # L2AM \n\t" \
       "pcmpeqd 24+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags2]  # L6+ (VectorPath)  into bytemask" \
       : [pflags] "=r" (pflags), [pflags2] "=r" (pflags2)
       : [divisors] "o" (PrimeNumbers[0]), [Deltas] "o" (Delta_of_PrimeNumbers[0][0]),
         [_xmm7] "x" (XMM7), [_xmm6] "x" (XMM6), [i] "r" (nr)
       : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "xmm0", "xmm1", "xmm2", "xmm3", "xmm4");

      //cout << "Siebpos: " << Siebpos << ": " << PackedResult[0] << "," << PackedResult[1] << endl;
 
#if defined(DEBUG)
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr]);
         if ( (d==Delta_of_PrimeNumbers[nr][0]) != ((pflags&0x03)==0x03) ) { MARK; cerr << nr << "!!" << (pflags&0x03) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr][1]) != ((pflags&0x0c)==0x0c) ) { MARK; cerr << nr << "!!" << (pflags&0x0c) << endl; exit(7); }
       }
#endif
      if ( pflags & 0x0f ) Factors.push_back(nr);

#if defined(DEBUG)
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+1]);
         if ( (d==Delta_of_PrimeNumbers[nr+1][0]) != ((pflags&0x30)==0x30) ) { MARK; cerr << nr+1 << "!!" << (pflags&0x30) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+1][1]) != ((pflags&0xc0)==0xc0) ) { MARK; cerr << nr+1 << "!!" << (pflags&0xc0) << endl; exit(7); }
       }
#endif
      if ( pflags & 0xf0 ) Factors.push_back(nr+1);

#if defined(DEBUG)
       {
         // check pflags2 for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+2]);
         if ( (d==Delta_of_PrimeNumbers[nr+2][0]) != ((pflags2&0x03)==0x03) ) { MARK; cerr << nr+2 << "!!" << (pflags2&0x03) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+2][1]) != ((pflags2&0x0c)==0x0c) ) { MARK; cerr << nr+2 << "!!" << (pflags2&0x0c) << endl; exit(7); }
       }
#endif
      if ( pflags2 & 0x0f ) Factors.push_back(nr+2);

#if defined(DEBUG)
       {
         // check pflags2 for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+3]);
         if ( (d==Delta_of_PrimeNumbers[nr+3][0]) != ((pflags2&0x30)==0x30) ) { MARK; cerr << nr+3 << "!!" << (pflags2&0x30) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+3][1]) != ((pflags2&0xc0)==0xc0) ) { MARK; cerr << nr+3 << "!!" << (pflags2&0xc0) << endl; exit(7); }
       }
#endif
      if ( pflags2 & 0xf0 ) Factors.push_back(nr+3);
    }
  asm("emms                      # exit mmx");
}
// end of SSE optimized version


#elif defined(ASMATHLON) /* special code for AMD 3DNow! */

void GetStaticFactorHits(vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )

  // Athlon 3DNow! specific code

  #warning "new AMD 3DNow! specific code"
  typedef int v2sf __attribute__ ((mode(V2SF)));
  register v2sf MM7 asm("mm7"); // MM7 is now bound to %mm7
  register v2sf MM6 asm("mm6"); // MM6 is now bound to %mm6 (intention: signedness correction)

   // refer GCC manual, "Specifying registers for local variables";
   // but remember: that does not necessarily mean, that %mm7 cannot be
   // used for other variables as well! (It is only guaranteed, that if we access
   // MM7, this access is synonym to accessing %mm7; but accessing %mm7 is not necessarily
   // synonym to accessing MM7 (because if the compiler assumes, that MM7 is dead, %mm7 could
   // be used for other variables; and the compiler does *not* know, that we want
   // MM7 to be alive by accessing %mm7!).

  asm ( \
       "femms                      # L2AMS  fast enter mmx\n\t" \
       "movd %[Sieboffs],%[_mm7]   # -      MM7=(0,Sieboffs)\n\t" \
       "punpckldq %[_mm7],%[_mm7]  # L2AM   (0,Sieboffs) -> (Sieboffs,Sieboffs) \n\t" \
       "movq %[_mm7],%%mm0         # L2AM \n\t" \
       "pi2fd %[_mm7],%[_mm7]      # L4A    int to float \n\t" \
       "pxor %[_mm6],%[_mm6]       # L2AM   clear mm6 \n\t" \
       "pcmpgtd %%mm0,%%mm6        # L2AM   set bits if signed, clear bits if unsigned" \
       : [_mm7] "=y" (MM7), [_mm6] "=y" (MM6) : [Sieboffs] "g" (Siebpos) : "mm0");

  // PrimeNumbers[0] und PrimeNumbers[1] werden ausgelassen (da sie gesondert behandelt werden mssen)
  for (int nr=2; nr<Size_StaticFactorbase; nr+=2) 
    {
      register unsigned int pflags;
      asm ( \
       "movq %[divisors](,%[i],4),%%mm4   # L2AMS  get y : x (as packed int)\n\t" \
       "pi2fd %%mm4,%%mm0          # L4A    get y : x (as packed float)\n\t" \
       "movq %%mm4,%%mm5           # L2AMS \n\t" \
       "pfrcp %%mm0,%%mm1          # L3M    1/y : 1/x (approx.) \n\t" \
       "movq %%mm0,%%mm2           # L2AM   y : x -> mm0 \n\t" \
       "punpckhdq %%mm0,%%mm0      # L2AM   y : y \n\t" \
       "movq %%mm2,%%mm3           # L2AM   y : x -> mm3 \n\t" \
       "pfrcp %%mm0,%%mm0          # L3M    1/y : 1/y (approx.) \n\t" \
       "punpckldq %%mm0,%%mm1      # L2AM   1/y : 1/x (approx.) \n\t" \
       "pfrcpit1 %%mm1,%%mm2       # L4M    1/y : 1/x (refine) \n\t" \
       "pfrcpit2 %%mm1,%%mm2       # L4M    1/y : 1/x (final) \n\t" \
       "pand %%mm6,%%mm4           # L2AM   correct sign (part 1) \n\t" \
       "pfmul %[_mm7],%%mm2        # L4M    z/y : w/x \n\t" \
       "pf2id %%mm2,%%mm1          # L4A    packed float to int \n\t" \
       "pi2fd %%mm1,%%mm1          # L4A    and back (->floor) \n\t" \
       "pfmul %%mm3,%%mm1          # L4M    multiply \n\t" \
       "pfsubr %[_mm7],%%mm1       # L4A    subtract \n\t" \
       "pf2id %%mm1,%%mm0          # L4A    ->modulo \n\t" \
       "paddd %%mm4,%%mm0          # L2AM   correct sign (part 2) \n\t" \
       "movq  %%mm0,%%mm4          # L2AM \n\t" \
       "pcmpeqd %%mm5,%%mm0        # L2AM  assure modp=0 instead of modp=p \n\t" \
       "pandn %%mm4,%%mm0          # L2AM \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd %[Deltas](,%[i],8),%%mm1   # L2AM \n\t" \
       "pcmpeqd 8+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags]   # L6+ (VectorPath)  into bytemask" \
       : [pflags] "=r" (pflags)
       : [divisors] "o" (PrimeNumbers[0]), [Deltas] "o" (Delta_of_PrimeNumbers[0][0]),
         [_mm7] "y" (MM7), [_mm6] "y" (MM6), [i] "r" (nr)
       : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5");

      //cout << "Siebpos: " << Siebpos << ": " << PackedResult[0] << "," << PackedResult[1] << endl;
 
#ifdef DEBUG
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr]);
         if ( (d==Delta_of_PrimeNumbers[nr][0]) != ((pflags&0x03)==0x03) ) { MARK; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr][1]) != ((pflags&0x0c)==0x0c) ) { MARK; exit(7); }
       }
#endif
      if ( pflags & 0x0f ) Factors.push_back(nr);

#ifdef DEBUG
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+1]);
         if ( (d==Delta_of_PrimeNumbers[nr+1][0]) != ((pflags&0x30)==0x30) ) { MARK; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+1][1]) != ((pflags&0xc0)==0xc0) ) { MARK; exit(7); }
       }
#endif
      if ( pflags & 0xf0 ) Factors.push_back(nr+1);
    }
  asm("femms                      # L2AMS  fast exit mmx");
}
// end of AMD 3DNow! optimized version

#else /* generic version */

void GetStaticFactorHits(vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )
  for (int i=2; i<Size_StaticFactorbase; i++)
   {
     // shortcut:
     /* keine aufwendige Division durchfhren, wenn diese
        Primzahl gar nicht ausgesiebt wurde */ 

     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[i]);
     // same as: Siebpos%PrimeNumbers[i]; if (d<0) d+=PrimeNumbers[i];
     // Bemerkung: %-operation does *not* eliminate the sign!
     //            ugly: -1=1 (mod 2), but (-1)%2=-1, 1%2=1
     // -> so we have to eliminate the sign!!!
     // and don't try to use unsigned int d, this won't work!

#if 0
     // for debugging reasons:
     bool b1=(mpz_div_ui(x,y,PrimeNumbers[i])==0);
     bool b2=(d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]);
     if (b1!=b2) 
      {
        cerr << "error in shortcut: " << Siebpos << "; "
             << PrimeNumbers[i] << ": " << d << ", "
             << Delta_of_PrimeNumbers[i][0] << ", "
             << Delta_of_PrimeNumbers[i][1] << endl;
        exit(1);
      }
#endif

     if (d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]) Factors.push_back(i);
   }
}
#endif /* generic code */


// fr die Speicherung der gefundenen dynamischen Faktoren.
// Bemerkung: Um die dynamischen Faktoren bereits in der Siebphase
//            aufzusammeln, ist das Array global.
//             (Dafr wird die Anzahl der gefundenen dynamischen Faktoren
//              fr die aktuelle Relation dem Konstruktor bergeben.)
const short int MaxHits=64; // bei Bedarf erhhen...
static struct {int Faktor, Exponent;} Hits[MaxHits]; // Dynamische Special-Factoren der Relation aufsammeln
// wichtiger Hinweis: Der Exponent zum Faktor wird nicht beim Sieben, sondern erst bei der Refaktorisierung
// ermittelt. Falls also mit Primzahlpotenzen gesiebt werden sollte, braucht der Exponent nicht eingetragen
// zu werden. Als Faktor sollte in jedem Falle die Primzahl und nicht die Primzahlpotenz eingetragen werden!

CRelation::CRelation(const signed int Siebpos, short int HitCount /* =0 */)
: CProvideHilfsvariablen(), relevanter_Faktor(no_factor), Relation_sparse(new CTinyFBsizetypeVector), Relation_dense(NULL)
  /* Konstruktor: neue Relation einfgen.
     "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */
{
  // cout << "initialisiere Gleichungszeile" << endl;
  mpz_init(Delta);
  Polynom.get_values(Siebpos,Delta,y);
  // in y: reduziertes Quadrat (ist nun zu faktorisieren...) 
  mpz_invert(Delta,Delta,n); /* wir verwenden den Kehrwert von Delta,
				das erspart eine zustzliche Variable;
				gleichzeitig ist damit die zu findende
				quadratische Kongruenz auf Delta^2=1 (mod n)
				normiert. */
  
  if (mpz_sgn(y)<0) // Vorzeichenbehandlung 
    {
      Relation_sparse->append(0); 
      /* "nullte Primzahl" ist -1 wegen Bercksichtigung
         des Vorzeichens. Also die 0 fr Position von  -1 einfgen */
      mpz_neg(y,y); /* -1 war Teiler -> positiv machen */
    }
  
  // ersten Faktor der statischen Faktorbasis ebenfalls gesondert behandeln,
  // denn die generierten Polynome haben Probleme, die Primzahl 2 zu sieben...

  if (PrimeNumbers[1]==2)
   {
     // special treatment for 2, since it it very easy:
     const unsigned int z = mpz_scan1(y,0);
     if (z)
      {
        mpz_tdiv_q_2exp(y,y,z);
        mpz_mul_2exp(Delta,Delta,z>>1);
        if (z&1) Relation_sparse->append(1); // ungerade Potenz der Primzahl markieren
      }
   }
  else
   {
     // this is the common case (which works also for 2)  
     unsigned int z=0;
     while (mpz_div_ui(x,y,PrimeNumbers[1])==0) { mpz_set(y,x); z++; }

     if (z&1) Relation_sparse->append(1); // ungerade Potenz der Primzahl markieren
     if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
      {
        if (z<4) mpz_mul_ui(Delta,Delta,PrimeNumbers[1]);
        else
         {
	   mpz_set_ui(x,PrimeNumbers[1]); mpz_powm_ui(x,x,z/2,n);
	   mpz_mul(Delta,Delta,x);
         }
      }
   }

  {
    // find factors in static factorbase
    // (beginnend mit der 3. Primzahl der FB (s.o) )
    vector<unsigned int> FB_Primteiler;
    GetStaticFactorHits(FB_Primteiler,Siebpos);

    // Faktoren zusammenmultiplizieren und dann in einem Rutsch abdividieren
    mpz_set_ui(x,1);
    for (unsigned int i=0; i<FB_Primteiler.size(); i++) mpz_mul_ui(x,x,PrimeNumbers[FB_Primteiler[i]]);
    mpz_divexact(y,y,x);

    // new factor reduction using the principle of inclusion/exclusion
    // die Faktor-Potenzen nach dem Prinzip Inklusion/Exklusion abdividieren
    vector<unsigned int> WithOddExponent, TempTeiler;
//    cout << "------> " << y << " ----: " << x << " ";
    while (mpz_cmp_ui(x,1)>0) // Teiler aus FB vorhanden?
     {
//       cout << "+";
       mpz_gcd(x,x,y); // enthlt die Primzahlen der FB, die gerade-oft y teilen
       mpz_divexact(y,y,x); // diese auch abdividieren
       mpz_mul(Delta,Delta,x); // und unter das Quadrat ziehen
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // teilt weiter...
        else
          WithOddExponent.push_back(FB_Primteiler[i]); // teilt hier letztmalig
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();

       mpz_gcd(x,x,y); mpz_divexact(y,y,x); // ungerade-oft
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // teilt weiter...
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();
     }
    mpz_mod(Delta,Delta,n);
    std::sort(WithOddExponent.begin(),WithOddExponent.end());
    for (unsigned int i=0; i<WithOddExponent.size(); ++i)
     {
       Relation_sparse->append(WithOddExponent[i]);
//       cout << " " << WithOddExponent[i];
     }
//    cout << endl;
//    cout << "++++++> " << y << " ----: " << x << endl;

  }


  // reduce factors from dynamic factorbase (and maybe pushed squares...)
  // Faktoren zusammenmultiplizieren und dann in einem Rutsch abdividieren
  if (HitCount)
   {
     mpz_set_ui(x,Hits[0].Faktor); Hits[0].Exponent = 1;
     for (int i=1; i<HitCount; i++) { Hits[i].Exponent = 1; mpz_mul_ui(x,x,Hits[i].Faktor); }
#if defined(DEBUG)
     if (!mpz_divisible_p(y,x)) { MARK; cerr << "Wrong divisor in Hits!" << endl; exit(1); }
#endif
     mpz_divexact(y,y,x);
     mpz_gcd(x,x,y);
     if (mpz_cmp_ui(x,1)!=0) // are there higher exponents?
      {
        // find exponents of factors in dynamic factorbase (and maybe pushed squares...)
        for (int i=0; i<HitCount; ++i)
         {
           while (mpz_divisible_ui_p(y,Hits[i].Faktor))
     	    {
     	      mpz_divexact_ui(y,y,Hits[i].Faktor);
     	      Hits[i].Exponent++; // Exponent fr Faktor inkrementieren
     	    }
#ifdef DEBUG
           if (Hits[i].Exponent==2)
            {
              MARK;
              cout << "Square in Hits[]: " << Hits[i].Faktor;
              if (is_dynamic_factor(Hits[i].Faktor))
               cout << " (dynamic factor)" << endl;
              else
               cout << " (pushed square)" << endl;
            }
#endif
           if (Hits[i].Exponent>1)
     	    {
     	      mpz_set_ui(x, Hits[i].Faktor);
     	      mpz_powm_ui(x,x,Hits[i].Exponent/2, n);
     	      mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
     	      Hits[i].Exponent%=2;
     	    }
         }
      }
   }

  // Restfactor in static factorbase...
  const bool existiert_grosser_Restfaktor = (mpz_cmp_ui(y,biggest_Prime_in_Factorbase)>0);

  if (existiert_grosser_Restfaktor && collecting_phase_finished)
    { relevanter_Faktor=special_factor; return; } // vorzeitiges Abbrechen

  if (!existiert_grosser_Restfaktor)
    {
      if (mpz_cmp_ui(y,1)!=0) 
        {
          cout << "CRelation::constructor:" << endl;
          cout << "Fehler in Relation / BUG im Programm:" << endl;
          cout << y << " should be 1!" << endl;
          exit(1);
        }
      relevanter_Faktor=largest_factor_in_Relation();
      // mit der aufgenommenen Relation lt sich dieser Faktor eliminieren
    }

#if 0 || defined(SIEVING_MORE_LARGE_SQUARES)
  // dynamic factor sanity check
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
    {
      if (!is_dynamic_factor(Hits[i].Faktor))
	{
	  cerr << "Inconsistency in dynamic factorbase! " << Hits[i].Faktor << " not found as known dynamic factor!" << endl;
          // entweder tatschlich ein Fehler ODER
          // denkbarer (unwahrscheinlicher) Sonderfall:
          // ein gepushtes Primzahl-Quadrat auerhalb der dynamischen
          // Faktorbasis geht (sogar) mit ungerader Potenz ein...
          // ...dann liegt zwar kein Fehler vor, aber die Relation ist
          // trotzdem ungltig. Der Fall erscheint mir aber derart
          // unwahrscheinlich, dass ich dann trotzdem abbreche.
          // Hinzu kommt: Wenn ein Client einen solchen Faktor sieben sollte,
          // dann wird diese Bearbeitung hier gar nicht aufgerufen und
          // der Server wird im Fehlerfall die Relation nicht akzeptieren.
          // Dann gibt es zwar eine mgliche dynamische Relation weniger,
          // aber das sollte bei der Seltenheit dieses Falles nicht stren.
	  exit(1); // not necessary, because it should also be safe just to skip this relation
          relevanter_Faktor=dynamic_factor; return; // skip the relation
	}
    }
#else
#ifdef DEBUG
 #warning "dynamic factor sanity check disabled! (just want to mention it)"
#endif
#endif

  if (existiert_grosser_Restfaktor)
    {
      const bool is_possible_dynamic = (mpz_cmp_ui(y, SingleLargePrime_Threshold)<=0);
      if ( !is_possible_dynamic && CmpqsFactor::DLP_rejected(y) )
        {
          // We should first filter out invalid big numbers, especially the
          // primes! It doesn't harm (except for performance) if some composite
          // numbers are filtered out as primes.
          // Important is that "DLP_reject(y)" is not too expensive!
	  relevanter_Faktor=special_factor; // ignored factor (because we do not insert any relation)
	  return;
        }
        
      if (is_possible_dynamic && probab_prime(mpz_get_ui(y)) )
        {
          // es liegt ein neuer dynamic Factor vor
          //cout << "new dynamic-factor..." << endl;
          relevanter_Faktor=dynamic_factor;
          TDynamicFactorRelation relation;
          relation.factor=mpz_get_ui(y);
          relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);

          // the found factor could be a "non-sievable" SingleLargePrime
          // We should detect this and take appropriate action!
          if (!relation.sieveable() && is_dynamic_factor(relation.factor))
           {
             Hits[HitCount].Faktor=relation.factor; Hits[HitCount].Exponent=1; ++HitCount;
             relevanter_Faktor=largest_factor_in_Relation();
             goto done_with_grosser_Restfaktor;
           }

#if !defined(IS_SERVER) && !defined(USE_FAKEHEAP)
          compute_Deltas_for(relation, SieveOffset+PhysicalSieveSize);
#endif
      
#ifndef IS_CLIENT
          relation.fpos = save(DynamicRelations_to_file,relation.factor, HitCount);
#else
          // an das Ausgabefile
          relation.fpos=save(communication_stream,relation.factor, HitCount);
#endif
          DynamicFactorRelations.insert(relation);
          statistical_data::directly_sieved_DynamicFactors++;
#ifndef IS_CLIENT
          SpecialRelations::split_by_primefactor(relation.factor);
#endif
          return;
        }
        
      // es liegt ein Specialfactor vor, der nicht prim ist
      {
        relevanter_Faktor=special_factor;
        TSpecialFactorRelation relation;
        if (!relation.factor.DLP_get(y))
          {
            //cerr << "Hinweis: Special Factor " << y << " invalid." << endl;
            return; /* Treffer nicht bernehmen, da er ungltig ist. */
          }
#ifndef IS_CLIENT
        if (!SpecialRelations::insert(relation.factor,this,HitCount))
          {
            relevanter_Faktor=largest_factor_in_Relation();
            // wenn Factor nicht vermerkt wurde, heit das, dass er
            // neutralisiert werden konnte. -> normale Relation liegt vor.

            // /FIXME: nachfolgendes sollte vielleicht besser in SpecialRelations::insert() getan werden?
            // kombiniere Faktoren der dynamic Factorbase
            multi_combine_init();
            for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
             {
               //cout << "kombiniere dynamischen Factor " << Hits[i].Faktor << endl;
               TDynamicFactorRelation FaRelSearch;
               FaRelSearch.factor=Hits[i].Faktor;
               TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
               mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
               multi_combine_main(DynamicRelations_from_file,(*p).fpos);
               //cout << "fertig kombiniert von " << *pos << endl;
             }
            multi_combine_exit();
            return;
          }
#else
        // an das Ausgabefile
        relation.fpos=save(communication_stream, relation.factor, HitCount);
#endif
	statistical_data::directly_sieved_SpecialFactors++;
        return;
      }
    } // end of "if (existiert_grosser_Restfaktor)"
 done_with_grosser_Restfaktor:

#ifndef IS_CLIENT
  // kombiniere Faktoren der dynamic Factorbase
  multi_combine_init();
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
   {
     //cout << "kombiniere dynamischen Factor " << Hits[i].Faktor << endl;
     TDynamicFactorRelation FaRelSearch;
     FaRelSearch.factor=Hits[i].Faktor;
     TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
     mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
     multi_combine_main(DynamicRelations_from_file,(*p).fpos);
     //cout << "fertig kombiniert von " << *pos << endl;
   }
  multi_combine_exit();
#else
  // Ausgabe an File
  save(communication_stream,1, HitCount);
#endif
      
  if (HitCount<6) ++statistical_data::Erfolgsaufteilung[HitCount];
  else ++statistical_data::Erfolgsaufteilung[6];
  // bei Erfolgsaufteilung[0]: Relation vollstndig durch statische FB entdeckt
  // bei Erfolgsaufteilung[0<k<6]: Relation durch Eliminierung von (genau) k Faktoren aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[6]: Relation durch Eliminierung von mindestens 6 Faktoren aus dyn. FB entdeckt
  
  //cout << "GL: " << *this << endl;
  //cout << "relevanter Faktor: " << relevanter_Faktor << endl;
  //cout << "Initialisierung der Gleichungszeile abgeschlossen!" << endl;
}


ostream& operator<< (ostream& ostr, const CRelation& GL)
{
  if (GL.Relation_dense) ostr << "dense"; else ostr << "sparse";
  ostr << "1=";
  
  char *str;
  str = new char [mpz_sizeinbase(GL.Delta,10)+2]; mpz_get_str(str, 10, GL.Delta);
  ostr << str << "^2";
  delete [] str; 
  
  if (GL.Relation_dense)
    { 
      for (int i=GL.Relation_dense->first(); i>=0; i=GL.Relation_dense->next(i))
	ostr << "*p" << i;
    }
  else
    {
      const CTinyFBsizetypeVector& R = *GL.Relation_sparse;
      for (int pos=0; pos<R.size(); ++pos)
	ostr << "*p" << R[pos];
    }
  
  return ostr;
}


void CRelation::convert_Relation_to_dense()
{
  //cout << "Converting relation from sparse to dense representation." << endl;
  if (Relation_dense)
    { cerr << "Relation ist bereits dense!" << endl; return; }
  Relation_dense = new myBitString;
  for (int pos=Relation_sparse->size()-1; pos>=0; --pos)
    Relation_dense->set((*Relation_sparse)[pos]);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // Sparsedarstellung lschen
}

void CRelation::convert_Relation_to_sparse()
{
  //cout << "Converting relation from dense to sparse representation." << endl;
  if (Relation_sparse)
    { cerr << "Relation ist bereits sparse!" << endl; return; }
  Relation_sparse = new CTinyFBsizetypeVector(Relation_dense->count(1));
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->fast_append(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // Densedarstellung lschen
}


void CRelation::combine(const CRelation& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Kombiniere Gleichungszeilen..." << endl;
  //cout << "Gl.1: "; SanityCheck();
  //cout << "Gl.2: "; GL2.SanityCheck();
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen
      
      for (int i=GL2.Relation_dense->last(); i>=0; i=GL2.Relation_dense->prev(i))
	if (Relation_dense->test(i))
	  if (i>0)
	    {
	      mpz_mul_ui(x,x,PrimeNumbers[i]);
	      if (mpz_cmp(x,n)>=0)
		{
		  mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		  mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		}
	    } //else mpz_neg(Delta,Delta);
      
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test((*GL2.Relation_sparse)[pos]))
	      if ((*GL2.Relation_sparse)[pos]>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[(*GL2.Relation_sparse)[pos]]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	    Relation_dense->invert((*GL2.Relation_sparse)[pos]);
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(Size_StaticFactorbase);
	neueMenge.reset();
        const CTinyFBsizetypeVector& R = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto fertig;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto fertig;
	      }
	    while (R2[p2]==R[p1])
	      {
		if (R[p1]>0)
		  {
		    mpz_mul_ui(x,x,PrimeNumbers[R[p1]]);
		    if (mpz_cmp(x,n)>=0)
		      {
			mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
			mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		      }
		  } //else mpz_neg(Delta,Delta);
		++p1; ++p2;
		if (p2==R2.size()) goto fertig;
		if (p1==R.size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
        //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
      }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Relation; dieses also dazumultiplizieren!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  relevanter_Faktor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}

// Routinen, die ein mehrfaches Kombinieren beschleunigen:
// NEU: Multikombinationsroutinen sollten das Kombinieren einer
//      Relation mit vielen anderen Relationen beschleunigen, da
//      das Update der Faktorbasis-Exponenten der Relation
//      an das Ende aller Verknpfungen geschoben werden kann und
//      nicht mehr simultan durchgefhrt werden mu.

CRelation::TExponentArrayElement CRelation::ExponentArray[MaxSize_StaticFactorbase+32] = {0};
 // +32 because we want to allow fast routines to avoid costly range checks...

unsigned int CRelation::multi_combine_Counter = 0;
 
void CRelation::multi_combine_init()
{
  // Initialisierung des Exponentenvektors
  // Hinweis: Der Aufruf dieser Routine darf auch entfallen,
  //          sofern sichergestellt ist, dass der letzte
  //          multi_combine_...-Aufruf ein  multi_combine_exit() war.
  if (!multi_combine_Counter) return;
  multi_combine_Counter=0;
  for (int i=0; i<Size_StaticFactorbase; i++) ExponentArray[i]=0;
}

void CRelation::multi_combine_exit()
{
  if (!multi_combine_Counter) return; // gibt nichts zu tun...
#ifdef VERBOSE
  cout << "Multicombine_exit after " << multi_combine_Counter
       << " calls." << endl;
#endif

  if (multi_combine_Counter>std::numeric_limits<TExponentArrayElement>::max())
   {
     MARK;
     cerr << "Elements of CRelation::ExponentArray may have exceeded limits," << endl
          << "you have to increase the type of CRelation::TExponentArrayElement" << endl
          << "and recompile the program to assure correct computation." << endl
          << "(Data computed so far are still correct and can be reused afterwards.)" << endl;
     exit(8);
   }

  optisize(); // optimize Size of vector (sparse<->dense, etc.)

  // nun die Faktor-Exponenten in das Delta ziehen:
  mpz_set_ui(y,1); // Produkt hierin sammeln
  ExponentArray[0]=0; // Primzahl "-1" immer unbercksichtigt lassen

  /* 
     Idee zur Beschleunigung:
     Exponentenfeld absteigend nach Exponenten sortieren, anschlieend
     das Produkt analog zum "Horner-Schema" auswerten.
     Damit das Sortieren nicht zu teuer wird, machen wir eine Art Bucketsort,
     wobei das "Flag[i]" angibt, ob es zu i einen Exponenten gibt.
     Alle zu einem Exponenten e gehrenden Basen werden dann in "Basis[e]"
     multiplikativ zusammengefat.
  */

  bool* const Flag = new bool[Size_StaticFactorbase];
  for (int i=0; i<Size_StaticFactorbase; ++i) Flag[i]=false; // initialisieren
  mpz_t* const Basis = new mpz_t[Size_StaticFactorbase];

  for (unsigned int i=1; i<Size_StaticFactorbase; ++i)
   if (ExponentArray[i])
    {
      const unsigned int e=ExponentArray[i];
      ExponentArray[i]=0; // Exponenten wieder zurcksetzen

      if (e>=Size_StaticFactorbase)
       {
         cout << "Unexpected exponent " << e << "! Use normal exponentiation..." << endl;
         mpz_set_ui(y,PrimeNumbers[i]); mpz_powm_ui(y,y,e,n);
         mpz_mul(Delta,Delta,y);
         continue;
       }

      if (Flag[e])
       {
         // Basis erweitern
         mpz_mul_ui(Basis[e],Basis[e],PrimeNumbers[i]);
         if (mpz_cmp(Basis[e],n)>=0)
          { 
            //cout << "Basis erweitern, mod sinnvoll..." << endl;
            mpz_mod(Basis[e],Basis[e],n);
          }
       }
      else 
       { 
         // Basis initialisieren
         Flag[e]=true;
         mpz_init(Basis[e]); mpz_set_ui(Basis[e],PrimeNumbers[i]);
       }
    }

 /*
  Produkt der Basen analog zum Horner-Schema auswerten,
  Beispiel: a^7 * b^4 * c^3 * d 
            = a^(7-4) * (a*b)^(4-3) * (a*b*c)^(3-1) * (a*b*c*d)
 */     

  unsigned int e2 = Size_StaticFactorbase-1;
  Flag[0]=true; // sicherheitshalber fr nachfolgende Schleife...
  while (!Flag[e2]) --e2; // hchsten gesetzten Exponenten holen

  mpz_set_ui(y,1);
  if (e2>1)
   for (unsigned int e=e2-1; e>0; --e)
    if (Flag[e])
     {
       mpz_mul(Basis[e],Basis[e],Basis[e2]);
       if (mpz_cmp(Basis[e],n)>=0)
        {
          //cout << "Auswertung A, mod sinnvoll" << endl; 
          mpz_mod(Basis[e],Basis[e],n);
        }
       if (e2-e>1) mpz_powm_ui(Basis[e2],Basis[e2],e2-e,n);
       mpz_mul(y,y,Basis[e2]);
       if (mpz_cmp(y,n)>=0)
        {
          //cout << "Auswertung B, mod sinnvoll" << endl; 
          mpz_mod(y,y,n);
        }
       mpz_clear(Basis[e2]); // do not forget: mpz-Zahl wieder freigeben
       e2=e;
     }

  if (e2>1) mpz_powm_ui(Basis[e2],Basis[e2],e2,n);
  if (e2>0)
   {
      mpz_mul(y,y,Basis[e2]); mpz_mod(y,y,n);
      mpz_clear(Basis[e2]); // do not forget: mpz-Zahl wieder freigeben
   }

  delete [] Basis; delete [] Flag;
  mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
  multi_combine_Counter=0;
#ifdef VERBOSE
  cout << "Multi_combine_exit finished." << endl;
#endif
}

void CRelation::multi_combine_main(const CRelation& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Multi-Kombiniere Relationen..." << endl;
  ++multi_combine_Counter; // Anzahl Aufrufe zhlen

  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Exponentenfeld vermerken
  */

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen

      /* Zunchsten werden die Bits der beiden Relationen
         (temporr!) komponentenweise durch "and" verknpft.
         Dadurch werden die Primzahlen ermittelt, die verknpft Quadrate
         ergeben. Diese knnen dann im Exponentenfeld bercksichtigt werden.
         In gewisser Weise wird die Bitstrings komponentenweise wie
         mit einem "Halbaddierer" verknpft.
          Additionsergebnis (durch xor) berschreibt anschlieend den
          Bitstring, whrend der bertrag im Expoentenfeld addiert wird.
       */
      Relation_dense->test_and_add_carry(*GL2.Relation_dense,ExponentArray);
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	for (int pos=0; pos<R2.size(); ++pos)
	  {
	    if (Relation_dense->test_and_invert(R2[pos])) ++ExponentArray[R2[pos]];
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(Size_StaticFactorbase);
	neueMenge.reset();
	const CTinyFBsizetypeVector& R  = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto fertig;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto fertig;
	      }
	    while (R2[p2]==R[p1])
	      {
		++ExponentArray[R[p1]];
		++p1; ++p2;
		if (p2==R2.size()) goto fertig;
		if (p1==R.size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
      }
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Relation; dieses also dazumultiplizieren!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  //-------------------------------------------------------------------
  // WICHTIG: Nach dem letzten Aufruf von multi_combine_main
  // mu die Relation mit dem Exponentenfeld abgeglichen werden!!!
  // Hierzu ist ein abschlieendes multi_combine_exit aufzurufen!
  //-------------------------------------------------------------------

  relevanter_Faktor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}



CmpqsFactor CRelation::combine(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  CmpqsFactor f=combine(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}

CmpqsFactor CRelation::multi_combine_main(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  CmpqsFactor f=multi_combine_main(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}


int GetValue(istream &in)
{
  static int ahead_value = 0;
  if (ahead_value>0) { int w=ahead_value; ahead_value=0; return w; }

  char c;
  in >> c;
  if ( c>='A' && c<='Z' ) return c-('A'-1); // A=1,...,Z=26
  if ( c>='g' && c<='z' )
   {
     int h = (c-'g');
     ahead_value=(h%5)+1; // nchster "vorausgelesener" Wert
     return (h/5)+1; 
   }
  if (c=='0') return 0; // da keine fhrenden Nullen
  if (c=='$') return -1;
  if (c=='%') return -2;

  if (c==' ')
   {
     MARK;
     cerr << "GetValue: Read a spece? How to handle this?" << endl;
     exit(1);
   }

  // ansonsten normal gespeicherte Zahl:
  int w;
  in.unget();
  in >> w;
  return w;
}

void PutValue(ostream &out, const int w)
{
  static int prev_value = 0;
  if (w<=5 && w>=1 && prev_value>0) // dann kann kodiert werden
    {
      out << char( (prev_value-1)*5 + w-1 + 'g' );
      prev_value=0;
      return;
    }

  if (prev_value>0)
   {
     out << char(prev_value+('A'-1)); // Wert ausgeben
     prev_value=0;
   }

  if (w>26) 
   {
     out << w << " "; // Wert unkodiert ausgeben
     return;
   }
  if (w>0)
   {
     if (w<=4) { prev_value=w; return; } // nchstes Mal abwarten
     out << char(w+('A'-1)); // Wert als char kodiert ausgeben
     return;
   }
  switch (w)
   {
     case -2: out << "%"; break;
     case -1: out << "$"; break;
     case  0: out << "0"; break;
     default:
      MARK;
      cerr << "PutValue: don't know how to handle this..." << endl;
      cerr << "w=" << w << endl;
      exit(1);
   }
}

CmpqsFactor CRelation::combine(istream &in)
{
  //cout << "kombiniere Gleichungszeile vom File" << endl;
  CmpqsFactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  const ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x ist klassenweite Hilfsvariable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index))
	    if (index>0)
	      {
		mpz_mul_ui(x,x,PrimeNumbers[index]);
		if (mpz_cmp(x,n)>=0)
		  {
		    mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		    mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		  }
	      } //else mpz_neg(Delta,Delta);
	  Relation_dense->invert(index);
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(Size_StaticFactorbase);
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto fertig;
	    }
	  while (index==R[p1])
	    {
	      if (index>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[index]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==R.size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
      //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
    }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
 
  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // Breitensuche, um dauerndes seeken zu vermeiden
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unbekannter Korrekturfaktor " << FaRelSearch.factor << endl;
	      exit(1); // sicherheitshalber fr die partiell korrekte Variante...
	      ret=0; break; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (combine(DynamicRelations_from_file,(*p).fpos).Typ()==CmpqsFactor::empty) { ret=0; break; } // auf Hintergrundspeicher kombinieren und bei Fehler ret=0 setzen
	  //cout << "fertig kombiniert von " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevanter_Faktor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

CmpqsFactor CRelation::multi_combine_main(istream &in)
{
  //cout << "Multi-Kombiniere Gleichungszeilen vom File..." << endl;
  ++multi_combine_Counter; // Anzahl Aufrufe zhlen

  CmpqsFactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  const ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x ist klassenweite Hilfsvariable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test_and_invert(index)) ++ExponentArray[index];
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(Size_StaticFactorbase);
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto fertig;
	    }
	  while (index==R[p1])
	    {
	      ++ExponentArray[index];
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==R.size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
    }

  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */

      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // Breitensuche, um dauerndes seeken zu vermeiden
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unbekannter Korrekturfaktor " << FaRelSearch.factor << endl;
              exit(1); // fr die partiell korrekte Variante...
	      ret=0; break; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (multi_combine_main(DynamicRelations_from_file,(*p).fpos).Typ()==CmpqsFactor::empty) { ret=0; break; } // auf Hintergrundspeicher kombinieren; im Fehlerfall ret=0 setzen
	  //cout << "fertig kombiniert von " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevanter_Faktor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

streampos CRelation::save(ostream &out, const CmpqsFactor factor, const short int HitCount /* = 0 */) const
{
  //cout << "Speichern einer Gleichungszeile" << endl;
#ifndef IS_CLIENT
  streampos pos;
  // out.seekp(0,ios::end);  // sollte jetzt schon am Ende sein
  pos = out.tellp(); // fileposition ans Ende setzen und auslesen
#endif
  char *str;

  str = new char [mpz_sizeinbase(Delta,mpzbase_f)+2]; mpz_get_str(str,mpzbase_f, Delta);
  out << "G " << setprecision(20) << factor << " " << str << " ";
  delete [] str;

  // ab hier wird hexadezimal gearbeitet:
  const ios::fmtflags oldFlags = out.flags(); // alte Formatflags retten
  out << hex; // Hexadezimalausgabe
  out.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /* Bemerkung:
     Statt der Werte werden die Abstnde gespeichert, das spart Platz */
  if (Relation_dense)
    { 
      int prev = 0;
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
          PutValue(out,i-prev);
	  prev=i;
	} 
    }
  else
    {
      int prev = 0;
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
 	  int h = (*Relation_sparse)[i];
          PutValue(out,h-prev);
	  prev=h;
	}
    }

  int Anzahl_echte_Faktoren = 0;
  for (int i=0; i<HitCount; i++) if (Hits[i].Exponent>0) Anzahl_echte_Faktoren++;
  if (Anzahl_echte_Faktoren>0)
    {
      PutValue(out,-2);
      //cout << "Korrekturfaktoren fr Relation mit Faktor " << setprecision(20) << factor << endl;
      for (int i=0; i<HitCount; i++)
	{ 
	  //cout << Hits[i].Faktor << "^" << Hits[i].Exponent << endl; 
	  if (Hits[i].Exponent>0) PutValue(out,Hits[i].Faktor);
	}
    }
  PutValue(out,-1); out << endl;
  out.flags(oldFlags); // alte Formatflags wiederherstellen
#ifdef IS_CLIENT
  return 0;
#else
  return pos;
#endif
}

void CRelation::SanityCheck() const
{
  mpz_set_ui(y,1); 
  if (Relation_sparse)
    {
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  if ((*Relation_sparse)[i]>0) 
	    mpz_mul_ui(y,y,PrimeNumbers[(*Relation_sparse)[i]]);
	  else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  else
    {
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(y,y,PrimeNumbers[i]); else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  
  mpz_invert(x,Delta,n); mpz_powm_ui(x,x,2,n);
  
  cout << "checking for congruency: ";
  cout << x << " = " << y << " (mod N)" << endl;
  if (mpz_cmp(x,y)!=0) 
    {
      cout << "Error! Values are NOT congruent!" << endl;
      exit(1);
    }
  
}



#ifndef IS_CLIENT
const bool CRelation::ComputeQuadraticCongruence() const
{
  /* Die ermittelte Relation erfllt die Kongruenz (A=1) mod n.
     Wegen A-1=(A+1)*(A-1)=0 mod n knnen wir daher nun durch ggT(A+1,n)
     einen Teiler von n ermitteln und haben gute Chancen, dass es sich dabei
     auch um einen echten (also nichttrivialen) Teiler handelt... */

  StatusReport(true);
  cout << endl << "Found a suitable relation..." << endl;

  if (!empty())
   {
     MARK;
     cerr << "Error! Relation contains unresolved members!" << endl;
   }

  cout << "[" << Delta << "] = 1 (mod N)" << endl;
   
  mpz_powm_ui(x,Delta,2,n);
  cout << "-> " << x << " = 1 (mod N)" << endl;
  
  if (mpz_cmp_ui(x,1)!=0) 
    {
      cout << "Error! Squares are NOT congruent!" << endl;
      exit(1);
    }
  
  mpz_add_ui(x,Delta,1);
  mpz_gcd(x,x,n); mpz_divexact(y,n,x); /* beachte: n und nicht kN
				          ist die eigentlich zu
				          faktorisierende Zahl! */
  
  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
      cout << "factorization was trivial..." << endl << endl;
    }
  else
    {
      if (mpz_cmp(x,y)>0) mpz_swap(x,y); // we want x<y

      cout << endl << "factor found!" << endl;
      cout << x << " * ";
      cout << y << endl;
      cout << "= ";
      cout << n << endl;
      
      if (mpz_probab_prime_p(x,probab_prime_checks))
	{
	  mpz_divexact(n,n,x);
	  cout << x << " is a factor." << endl;
	  Factorization_to_file << MAL() << x << flush;
	}
      else
      	if (Potenztest(x)) mpz_divexact(n,n,x); // ggf. Potenzen streichen...

      if (mpz_probab_prime_p(y,probab_prime_checks))
	{
	  mpz_divexact(n,n,y);
	  cout << y << " is a factor." << endl;
	  Factorization_to_file << MAL() << y << flush;
	}
      else
	if (Potenztest(y)) mpz_divexact(n,n,y); // ggf. Potenzen streichen...
      

      if (mpz_cmp_ui(n,1)==0) // factorization complete?
	{ 
	  cout << "factorization successfully completed." << endl;
          Factorization_to_file << endl;
          return true; 
	}
    }
  cout << "continue factorization..." << endl << endl;
  return false;
}
#endif
  

void StaticRelations::Load()
{
  // Static Relations einlesen
  //cout << "Reading static relations from file..." << endl;
  StaticRelations_from_file.seekg(0,ios::beg);
  while (StaticRelations_from_file.peek()=='G') 
   {
     CRelation* GL = new CRelation();
     GL->combine(StaticRelations_from_file);
     GL->optisize(); // Relation optimieren (dense<->sparse, etc.)
     GLS[GL->relevanter_Faktor]=GL; ++Fuellung_GLS; // Relation ins Gleichungssystem einfgen
     StaticRelations_from_file.ignore(1,'\n');
   }
  StaticRelations_from_file.clear();
  //cout << "Load: " << Fuellung_GLS << " static relations read." << endl;
}

void StaticRelations::insert(CRelation *GL, const bool do_multi_combine_init /* =true */)
{ /* hier wird eine neu gefundene Relation in das bisherige Gleichungssystem
     eingefgt. Falls sie redundant ist bzw. zu einer Faktorisierung fhrt,
     so wird dies automatisch erkannt. */
#ifdef IS_CLIENT
  if (GL->relevanter_Faktor >= 0) Fuellung_GLS++; // beim Client mangels Gleichungssystem: jede gefundene Relation
  delete GL;
  return;
#else  
  if (GL->relevanter_Faktor==CRelation::dynamic_factor || GL->relevanter_Faktor==CRelation::special_factor)
    { delete GL; return; } /* diese Relationen sind auf Hintergrundspeicher ausgelagert
			      und brauchen deshalb nicht im Hauptspeicher zu verbleiben */
  
  if (do_multi_combine_init) GL->multi_combine_init();
  while (!GL->empty())
    {
      if (GL->relevanter_Faktor>=Size_StaticFactorbase)
       {
         MARK;
         cerr << "value too big!" << endl;
         exit(1);
       }
      CRelation *pos = GLS[GL->relevanter_Faktor];
      if (pos!=NULL)
	{ /* Relation mit diesem Eintrag existiert schon -> neue Kombination mglich! */
#if 1 /* Minimierung der Relationen ein-/ausschalten */
          if (
	       (GL->Relation_sparse) && (pos->Relation_sparse) // beide in Sparse-Darstellung
               && GL->SizeOfRelation()<=pos->SizeOfRelation()
	       && GL->second_largest_factor_in_Relation()<=pos->second_largest_factor_in_Relation()
             )
	    { /* weniger Primzahlen in der Relation bedeuten schnelleres Abarbeiten.
	  	 Da beide Relationen quivalent sind, knnen sie problemlos so
		 getauscht werden, dass die platzsparende im Gleichungssystem verbleibt.
		 Andererseits bestimmt das Maximum der zweitgrten Faktoren zweier
		 Relationen die nchste Verknpfung. Also ist auch dies zu minimieren.
                 Beim Zielkonflikt wren beide Kriterien also abzuwgen. Dies geschieht
		 hier nicht. Nur wenn beide Kriterien zugleich die Lage nicht
                 verschlechtern, werden die Relationen getauscht.

                 Man beachte, dass der Tausch das Ergebnis der aktuellen Verknpfung
                 nicht verndert, sondern nur zuknftige Verknpfungen
	         beschleunigen kann bzw. den Speicherbedarf vermindert.
              */
	      GL->multi_combine_exit(); 
	      GLS[GL->relevanter_Faktor]=GL;
	      std::swap(pos,GL); // Relationen tauschen
              // GL->multi_combine_init(); <- darf entfallen
	    }
#endif
          // grter Faktor mu eliminiert werden:
 	  GL->multi_combine_main(*pos); 
	}
      else
	{
	  GL->multi_combine_exit();
	  GLS[GL->relevanter_Faktor]=GL; // Relation ins Gleichungssystem eintragen
	  Fuellung_GLS++; // eine Relation mehr im Gleichungssystem
          GL->save(StaticRelations_to_file);
          StatusReport();
	  return;
	}
    }

  GL->multi_combine_exit();

  // the relation is redundant:
  // -> we cannot insert it into the static factorbase,
  // -> it can yield a factor instead!
  // -> compute quadratic congruency!
  const bool complete = GL->ComputeQuadraticCongruence();
  delete GL;
  if (complete)
   {
     // Game over, we have won!
     // leave the program...
     ExitManager::StopFactorization(); // Programm erfolgreich beenden...
     exit(0); // this line should not be reached!
   }
#endif
}


void SpecialRelations::Load()
{
  // Special Relations einlesen
  // zustzlich Speicherkompaktifizierung der special relations
  // auf dem File durchfhren, d.h. ungltige (=als redundant
  // markierte) Relationen streichen.
  cout << "Reading special relations and reorganizing file..." << endl;
  unsigned int u_hits = 0, g_hits = 0;
  streampos rpos=0, wpos=0;
  // Special Relations einlesen
  SpecialRelations_from_file.seekg(0,ios::beg);
  SpecialRelations_to_file.seekp(0,ios::beg);

  while (SpecialRelations_from_file.peek()=='G')
   {
     TSpecialFactorRelation relation;
     relation.fpos = SpecialRelations_from_file.tellg();
     string s; SpecialRelations_from_file >> s; // "G " einlesen
     SpecialRelations_from_file >> relation.factor; // special relation factor einlesen
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
     ++g_hits;
    }
  wpos=rpos=SpecialRelations_from_file.tellg(); //Schreibposition=Leseposition

  while (true) 
    {
      SpecialRelations_from_file.seekg(rpos);
      while (SpecialRelations_from_file.peek()=='U')
       {
         SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
         ++u_hits;
       }

      int cached_g = 0;
      const int max_cached_g = 1000;
      string sG[max_cached_g];
      while (SpecialRelations_from_file.peek()=='G')
       {
         while (SpecialRelations_from_file.peek()!='\n')
          {
            char buffer[2048];
            SpecialRelations_from_file.get(buffer, sizeof(buffer), '\n');
            sG[cached_g]+=buffer;
          }
         SpecialRelations_from_file.ignore(10,'\n');
         if (++cached_g>=max_cached_g) break;
         while (SpecialRelations_from_file.peek()=='U')
          {
            SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
            ++u_hits;
          }
       }
      rpos=SpecialRelations_from_file.tellg();
      if (cached_g==0) break;
      SpecialRelations_to_file.seekp(wpos);

      for (int i=0; i<cached_g; ++i)
       {
         TSpecialFactorRelation relation;
         relation.fpos = SpecialRelations_to_file.tellp();
         {
           istringstream is(sG[i]);
           string s; is >> s;
           is >> relation.factor;
         }
         SpecialRelations_to_file << sG[i] << "\n";
         SpecialFactorRelations.insert(relation);
         relation.factor.swap();
         SpecialFactorRelations.insert(relation);
         relation.factor.swap();
       };
      SpecialRelations_to_file.flush();
      wpos=SpecialRelations_to_file.tellp();
      g_hits+=cached_g;
      cout << g_hits << " special relations written, "
           << u_hits << " removed." << "\r" << flush;
      //cout << (long long int)static_cast<streamoff>(rpos) << " , " << (long long int)static_cast<streamoff>(wpos) << endl;
    }
   cout << endl;
   
  // mangels truncate-Funktion den Rest mit U's auffllen
  {
    streampos pos, ende;
    pos = wpos;
    SpecialRelations_to_file.seekp(0, ios::end);
    ende = SpecialRelations_to_file.tellp();
    SpecialRelations_to_file.seekp(wpos);
    //cout << (long long int)static_cast<streamoff>(pos) << " , " << (long long int)static_cast<streamoff>(ende) << endl;

    char buffer[4096];
    for (unsigned int i=0; i<sizeof(buffer); ++i) buffer[i]='U';
    while (static_cast<streamoff>(SpecialRelations_to_file.tellp()) < static_cast<streamoff>(ende))
      {
        SpecialRelations_to_file.write(buffer,sizeof(buffer));
        //cout << (long long int)static_cast<streamoff>(SpecialRelations_to_file.tellp()) << " , " << (long long int)static_cast<streamoff>(ende) << endl;
      }
    SpecialRelations_to_file << flush;
    SpecialRelations_to_file.seekp(pos);
  }
  SpecialRelations_from_file.clear();
  cout << "Recover: " << SpecialFactorRelations.size()
       << " special relations read, " << endl
       << u_hits << " invalid DLP-relations eliminated." << endl;
}



const bool SieveControl::Hit_after_DirtyCorrection(const signed int Siebpos, TSieveElement Value)
{
  // An dieser Stelle im Sieb wurde ein Treffer gefunden;
  // aber wegen der Dirty-Siebfaktoren, die fr alle Siebelemente als Treffer gezhlt werden,
  // knnte es tatschlich gar kein Treffer sein.
  // Wir mssen daher den Dirty-Siebwert um die nicht-teilenden Dirty-Faktorbasiselemente korrigieren.
  // Dieser Korrekturwert wird hier bestimmt (allerdings mit shortcut).

  // Wenn diese Routine aufgerufen wird, ist mglicherweise ein Siebtreffer gelandet worden;
  // Da es sich um ein Regelungssystem handelt, ist dies (zwar kein perfekter, aber)
  // ein sinnvoller Indikator, um nun wieder etwas weniger Dirtyness zu verlangen,
  // wenn doch kein Siebtreffer vorlag...

  for (int i=_sieben_ab_Primzahlindex-1; i>=PrimzahlStartIndex; --i)
   {
     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[i]);
     // same as: Siebpos%PrimeNumbers[i]; if (d<0) d+=PrimeNumbers[i];
     // Bemerkung: %-operation does *not* eliminate the sign!
     //            ugly: -1=1 (mod 2), but (-1)%2=-1, 1%2=1
     // -> so we have to eliminate the sign!!!
     const bool Treffer = (d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]);
     if (Treffer)
      {
        // Value verkleinert sich
        Value+=PrimzahlSiebWichtung_dirty(i); // dirty-Wichtung wieder wegnehmen
	Value-=PrimzahlSiebWichtung(i); // und durch die echte Wichtung ersetzen
      }
     else
      {
        // hat nicht getroffen:
        // Value vergert sich
        Value+=PrimzahlSiebWichtung_dirty(i); // dirty-Wichtung wegnehmen, da kein Treffer vorlag

        // shortcut "return false" on value>=0 doesn't work anymore, since it's not monotone increasing
        //  2004-03-12: maybe we should revert to the old behaviour?
         // the more dirty the sieving is -> the more values to analyse
         // the more expensive divisions we need to take and we cannot shortcut! :-(
      }
   }

  if (Value>=0)
   {
     SieveControl::RequestLessDirtyness(); // "gegenregeln"
     return false; // Threshold berschritten, kein Siebtreffer
   }
  return true; // Threshold nicht berschritten -> Siebtreffer
}


TSieveElement SieveControl::log_Threshold = 0; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert

void SieveControl::compute_SieveThreshold(void)
{
  // Hier wird der Threshold (berschlagsmig) ermittelt
  
  // Voraussetzungen zum Aufruf der Initialisierung:
  // 1. Logische Siebintervallgre LogicalSieveSize (fr Intervall [-LogicalSieveSize,LogicalSieveSize])
  // 2. kN (die mit Vorfaktor versehene zu faktorisierende Zahl)
  // 3. die statische Faktorbasis
  // mssen bereits bekannt und berechnet sein.
 
  // logarithmischen Schwellwert kalkulieren:
  mpz_t x; mpz_init(x);
  mpz_div_ui(x,kN,2); mpz_sqrt(x,kN); mpz_mul_ui(x,x,LogicalSieveSize);
  double d = log(fabs(mpz_get_d(x))) - (Factor_Threshold*log(biggest_Prime_in_Factorbase));
  mpz_clear(x);
  if (PrimeNumbers[1]==2) d-=log(2); // Falls 2 in der Faktorbasis, dann teilt 2 alle Siebeintrge
  d*=log_Multiplikator;
  cout << "Threshold (initial sieve entries) for detecting relations set to " << d << endl;
  log_Threshold=static_cast<TSieveElement>(floor(d));
  if (log_Threshold!=floor(d)) 
   {
     cerr << "With the compiled-in types and constants this" << endl;
     cerr << "factorization cannot be done, sorry." << endl; 
     cerr << "TSieveElement (compile-time type) needs to be increased to fit the threshold!" << endl;
     cerr << "Please modify log_Multiplikator and/or TSieveElement" << endl;
     cerr << "and compile the program again." << endl;
     exit(1);
   }

#if TSIEBELEMENTSIZE > 0
  // sanity check...
  if (sizeof(TSieveElement)!=TSIEBELEMENTSIZE)
   {
     cerr << "Die einkompilierten Konstanten/Typen fr " << endl
          << "TSieveElement (Size=" << sizeof(TSieveElement) << ")"
          << " und TSIEBELEMENTSIZE (=" << TSIEBELEMENTSIZE << ")" << endl
          << "stimmen nicht berein. Bitte korrigieren!" << endl;
     exit(1);
   }
#endif
}


#ifndef IS_SERVER

#ifdef ASM386
#define SIEBASM386

unsigned int clobbered_int; // dummy

#if TSIEBELEMENTSIZE == 1
#define asm_sieb_init() \
  asm volatile ( \
       "movb %%al,%%ah\n\t" \
       "pushw %%ax\n\t" \
       "shll $16,%%eax\n\t" \
       "popw %%ax\n\t" \
       "shrl $2,%%ecx\n\t" \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "cc", "memory");

#ifdef ASMATHLON
// ATHLON-optimized code
#ifdef DEBUG
#warning "using ATHLON version for macro: asm_search_sieb"
#endif
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $7,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "jns 0b \n\t" \
       "js 4f # Treffer gefunden \n\t" \
       "1: # Treffer suchen (big steps) loop \n\t" \
       "prefetch -256(%[sieb],%[pos]) # tuning cache!! \n\t" \
       "subl $64,%[pos] \n\t" \
       "movq  (%[sieb],%[pos]),%%mm0 \n\t" \
       "movq 8(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 16(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 24(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 32(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 40(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 48(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 56(%[sieb],%[pos]),%%mm1 \n\t" \
       "por %%mm1,%%mm0 \n\t" \
       "pmovmskb %%mm0,%%eax # highest bits of packed bytes to %%ah \n\t" \
       "testl %%eax,%%eax \n\t" \
       "jz 1b \n\t" \
       "femms # clear MMX state \n\t" \
       "2: # Treffer suchen (little steps) init \n\t" \
       "addl $64,%[pos] \n\t" \
       "3: # Treffer suchen (little steps) loop \n\t" \
       "subl $4,%[pos] \n\t" \
       "testl $0x80808080,(%[sieb],%[pos]) \n\t" \
       "jz 3b \n\t" \
       "movl (%[sieb],%[pos]),%%eax \n\t" \
       "addl $3,%[pos] \n\t" \
       "testl $0x80000000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x00800000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x00008000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "4: # Treffer gefunden" \
       : [pos] "=q" (offset) : "[pos]" (offset), [sieb] "r" (SieveArray) : "cc", "eax", "mm0", "mm1");
#else
// optimized i386 code
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $3,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "jns 0b \n\t" \
       "js 2f # Treffer gefunden \n\t" \
       "1: # Treffer suchen \n\t" \
       "subl $4,%[pos] \n\t" \
       "testl $0x80808080,(%[sieb],%[pos]) \n\t" \
       "jz 1b \n\t" \
       "addl $3,%[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#endif
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmpl %[disp1],%[disp0]\n\t" \
       "jle 3f\n\t" \
       "xchgl %[disp1],%[disp0]\n" \
       "3:\n\t" \
       "testl %[disp0],%[disp0]\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subb %[val],(%[feld],%[disp1])\n\t" \
       "subb %[val],(%[feld],%[disp0])\n\t" \
       "subl %[step],%[disp1]\n\t" \
       "subl %[step],%[disp0]\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "testl %[disp1],%[disp1]\n\t" \
       "jl 2f\n\t" \
       "subb %[val],(%[feld],%[disp1])\n" \
       "2:\n" \
       : [disp0] "=q" (clobbered_int), [disp1] "=q" (clobbered_int) : [feld] "r" (SieveArray), [step] "g" (P), [val] "r" (lp), "[disp0]" (d0), "[disp1]" (d1) : "cc");

#elif TSIEBELEMENTSIZE == 2
#define asm_sieb_init() \
  asm volatile ( \
       "pushw %%ax\n\t" \
       "shll $16,%%eax\n\t" \
       "popw %%ax\n\t" \
       "shrl $1,%%ecx\n\t" \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "cc", "memory");
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $1,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testw $0x8000,(%[sieb],%[pos],2) \n\t" \
       "js 2f # Treffer gefunden \n\t" \
       "1: # Treffer suchen \n\t" \
       "subl $2,%[pos] \n\t" \
       "testl $0x80008000,(%[sieb],%[pos],2) \n\t" \
       "jz 1b \n\t" \
       "incl %[pos] \n\t" \
       "testw $0x8000,(%[sieb],%[pos],2) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subw %4,(%2,%1,2)\n\t" \
       "subw %4,(%2,%0,2)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subw %4,(%2,%1,2)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (SieveArray), "g" (P), "q" (lp), "0" (d0), "1" (d1) : "cc");
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (SieveArray), "g" (P), "q" (lp) : "cc");
#elif TSIEBELEMENTSIZE == 4
#define asm_sieb_init() \
  asm volatile ( \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "memory");
#define asm_search_sieb(offset) \
  asm volatile ( \
       "1: # Treffer suchen \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x80008000,(%[sieb],%[pos],4) \n\t" \
       "jz 1b \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subl %4,(%2,%1,4)\n\t" \
       "subl %4,(%2,%0,4)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subl %4,(%2,%1,4)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (SieveArray), "g" (P), "q" (lp), "0" (d0), "1" (d1) : "cc");
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (SieveArray), "g" (P), "q" (lp) : "cc");
#else
#undef SIEBASM386
#warning "Kein handoptimierter Asssemblerkode fr Aussieben mit TSieveElement!"
#endif
#endif


void initialize_Sieve(void)
{
  //cout << "Initializing sieve from " << SieveOffset
  //     << " to " << SieveOffset+PhysicalSieveSize << endl;

 SieveControl::RequestMoreDirtyness();

#ifdef SIEBASM386
  asm_sieb_init(); // siehe das entsprechende Assembler-Makro
#else
  register TSieveElement w = SieveControl::GetLogThreshold();
  /* log_Threshold bezeichnet den Schwellwert, der zum Erkennen der Relation
     unterschritten werden mu. Es ist besser, den Threshold bereits bei der
     Initialisierung zu bercksichtigen (d.h. abzuziehen), da beim Auslesen
     des Siebs dann nur auf negative Werte getestet werden braucht. */
  
  // Initialisierung des Siebs mit dem berschlagsmigen Schwellwert
  register int i=PhysicalSieveSize; do SieveArray[--i]=w; while (i);
#endif

  // nachfolgender Block dient (falls aktiviert) nur noch Validierungszwecken:
  // enable this block to compare the polynomial value and its logval against they actually used threshold
#if 0
  {
    mpz_t x,y;
    mpz_init(x); mpz_init(y);
    Polynom.get_values(SieveOffset,x,y);
    TSieveElement w_test = static_cast<TSieveElement>(log_Multiplikator*log(fabs(mpz_get_d(y))));
    TSieveElement FB_Threshold
      = static_cast<TSieveElement>(log_Multiplikator*Factor_Threshold*log(biggest_Prime_in_Factorbase));
    w_test-=FB_Threshold;
    if (PrimeNumbers[1]==2) w_test-=static_cast<TSieveElement>(log_Multiplikator*log(2));
 #if TSIEBELEMENTSIZE == 1
    // chars are not printed as numbers, so we cast them:
    cout << "value " << y << " -> logval: " << static_cast<signed int>(w_test)
         << " <-> threshold: " << static_cast<signed int>(SieveControl::GetLogThreshold()) << endl;
 #else
    cout << "value " << y << " -> logval: " << w_test << " <-> threshold: " << SieveControl::GetLogThreshold() << endl;
 #endif
    mpz_clear(x); mpz_clear(y);
  }
#endif
}


// helper class to encapsulate private data
class CSieveStaticFactors
{
 private:
  static int Local_Delta_of_PrimeNumbers[MaxSize_StaticFactorbase][2];
  static void init_LocalDeltas();
  static void do_sieving_StaticFactors();
  friend int main(const int argc, const char* const argv[]);
};
int CSieveStaticFactors::Local_Delta_of_PrimeNumbers[MaxSize_StaticFactorbase][2];


void CSieveStaticFactors::init_LocalDeltas()
{
  int nr=Size_StaticFactorbase-1;
  // groe Primzahlen (maximal 1 Siebtreffer pro Delta) 
  for (;; --nr) 
    {
      register int Primzahl=PrimeNumbers[nr];

      if (Primzahl<PhysicalSieveSize) break;

      // d0/d1-Ermittlung
      register int d0,d1;
      {
        register int h=normalized_signed_mod(SieveOffset,Primzahl); // damned signed %-Operation...
	d0=Delta_of_PrimeNumbers[nr][0]-h;
	d1=Delta_of_PrimeNumbers[nr][1]-h;
	if (d1<0)
          d1+=Primzahl;
	if (d0<0)
          d0+=Primzahl;
      }
     Local_Delta_of_PrimeNumbers[nr][0]=d0;
     Local_Delta_of_PrimeNumbers[nr][1]=d1;
    }
}

void CSieveStaticFactors::do_sieving_StaticFactors()
{
  //cout << "Aussieben mit Primzahlen der statischen FB: " << SieveOffset<< endl;
  
  /* wir sieben erst ab einer greren Primzahl aus der Faktorbasis,
     da wir sowieso unsauber und schnell sieben...
     ( je kleiner die Primzahl, um langsamer ist das Sieben mit dieser
     Zahl; und bei der gewhlten Implementierung ohne Bercksichtigung
     von Primzahlpotenzen ist auch der zustzlich gemachte Fehler
     vernachlssigbar und wird zudem ber Factor_Threshold kompensiert.
     Aber Vorsicht: Shortcut-Konstante nicht zu hoch ansetzen, sonst fallen
     zu viele Relationen raus!
     Achtung: als Startwert mindestens 2 whlen, damit -1 und 2 ignoriert
     werden knnen. Dann gibt es auch 2 Deltawerte. )
     -> relevant fr Schleife mit den kleinen Primzahlen
  */

  // standard code, but uses also various assembler optimizations

  int nr=Size_StaticFactorbase-1;
  // groe Primzahlen (maximal 1 Siebtreffer pro Delta) 
  for (;; --nr) 
    {
      register int Primzahl=PrimeNumbers[nr];

      if (Primzahl<PhysicalSieveSize) break;

      // d0/d1-Ermittlung
      register int d0,d1;
      
      d0=Local_Delta_of_PrimeNumbers[nr][0];
      d1=Local_Delta_of_PrimeNumbers[nr][1];

      register TSieveElement log_Primzahl = SieveControl::PrimzahlSiebWichtung(nr);
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      SieveArray[d0<PhysicalSieveSize ? d0 : PhysicalSieveSize]-=log_Primzahl;
      SieveArray[d1<PhysicalSieveSize ? d1 : PhysicalSieveSize]-=log_Primzahl;

      d0-=PhysicalSieveSize; if (d0<0) d0+=Primzahl;
      d1-=PhysicalSieveSize; if (d1<0) d1+=Primzahl;
      Local_Delta_of_PrimeNumbers[nr][0]=d0;
      Local_Delta_of_PrimeNumbers[nr][1]=d1;
    }

  // kleine Primzahlen (mehrere Siebtreffer pro Delta)
  const int NonDirtyIndex = SieveControl::sieben_ab_Primzahlindex();
  for (; nr>=NonDirtyIndex; --nr) 
    {
      register int Primzahl = PrimeNumbers[nr];
      //cout << "Quadratisches Sieb mit " << Primzahl << endl;
      
      // d0/d1-Ermittlung
      register int d0,d1;
      {
	register int h=normalized_signed_mod(SieveOffset+PhysicalSieveSize,Primzahl); // damned signed %-Operation...
	d0=Delta_of_PrimeNumbers[nr][0]-h;
	d1=Delta_of_PrimeNumbers[nr][1]-h;
	if (d1>=0) d1-=Primzahl;
	if (d0>=0) d0-=Primzahl;
      }
      
      register TSieveElement log_Primzahl = SieveControl::PrimzahlSiebWichtung(nr);
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      d0+=PhysicalSieveSize;
      d1+=PhysicalSieveSize;

      //if (d0==d1) cout << Primzahl << " teilt Vorfaktor!" << endl;
      
#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, Primzahl);
#else 
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so dass d0<d1
      while (d0>=0)
	{
	  SieveArray[d0]-=log_Primzahl; d0-=Primzahl;
	  SieveArray[d1]-=log_Primzahl; d1-=Primzahl;
	}
      if (d1>=0) SieveArray[d1]-=log_Primzahl;
#endif
    }

  // hhere Primzahlpotenzen
  for (nr=0; nr<NumberOf_more_PrimePowers; nr++)
    {
      register int PPotenz = PrimePowers[nr];

      // d0/d1-Ermittlung
      register int d0,d1;
      {
	register int h=normalized_signed_mod(SieveOffset+PhysicalSieveSize,PPotenz); // damned signed %-Operation...
	d0=Delta_of_PrimePowers[nr][0]-h;
	d1=Delta_of_PrimePowers[nr][1]-h;
	if (d1>=0) d1-=PPotenz;
	if (d0>=0) d0-=PPotenz;
      }
      
      register TSieveElement log_Primzahl = log_Primzahl_of_PrimePowers[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      // beachte: da mit den Primzahlen selbst auch gesiebt wird, darf nicht log(Primzahlpotenz), sondern
      // nur log(Primzahl) abgezogen werden!
      // Analogie: Wenn man schon durch p geteilt hat, mu man nocheinmal durch p teilen,
      //           um insgesamt durch p^2 zu teilen!

      d0+=PhysicalSieveSize;
      d1+=PhysicalSieveSize;

#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, PPotenz);
#else
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so dass d0<d1
      while (d0>=0)
	{
	  SieveArray[d0]-=log_Primzahl; d0-=PPotenz;
	  SieveArray[d1]-=log_Primzahl; d1-=PPotenz;
	}
      if (d1>=0) SieveArray[d1]-=log_Primzahl;      
#endif
    }

}


void do_sieving_Squares()
{
  // Quadrate von statischen (und ggf. dynamischen Faktoren) sieben.
  // Wichtig: Daher nur den einfachen Logarithmus des Faktors subtrahieren!
  // Bemerkung: Zur Effizienzsteigerung wird davon ausgegangen, dass jedes Quadrat nur einmal
  //            innerhalb des gesamten Siebintervalls trifft!

  //cout << "Quadrate sieben Start" << endl;
  if (Sieve_Delta_HeapOfSquares.empty()) return; // es gibt nichts zu sieben...
  TSieve_Delta sieb_delta = Sieve_Delta_HeapOfSquares.top();
  register int delta = sieb_delta.delta-SieveOffset;
  
  while (delta<PhysicalSieveSize)
    {
      Sieve_Delta_HeapOfSquares.pop(); // Element aus der Priorittswarteschlange nehmen
      SieveArray[delta]-=static_cast<TSieveElement>(ceil(log_Multiplikator*log(sieb_delta.factor))); // Teiler im Sieb vermerken
      if (Sieve_Delta_HeapOfSquares.empty()) break;
      sieb_delta = Sieve_Delta_HeapOfSquares.top();
      delta = sieb_delta.delta-SieveOffset;
    }
  //cout << "Quadrate sieben Ende" << endl;
}


void do_sieving_DynamicFactors()
{
  if (Sieve_Delta_Heap.empty()) return; // es gibt nichts zu sieben...
  short int HitCount=0;
  TSieve_Delta sieb_delta = Sieve_Delta_Heap.top();
  register int delta = sieb_delta.delta-SieveOffset;
  goto Schleifentest;
  
 Schleifenanfang_weiterer_Faktor:
  ++HitCount;
  //if (HitCount>=MaxHits) { cerr << "MaxHits erhhen!"; exit(1); } 
 Schleifenanfang_normal:
  Sieve_Delta_Heap.pop(); // Element aus der Priorittswarteschlange nehmen
  SieveArray[delta]-=log_DynamicLargePrime(sieb_delta.factor); // Teiler im Sieb vermerken
  
  // Da dynamische Faktoren selten auftreten, diese Faktoren schon whrend
  // des Siebens vermerken...
  // ...das erhht die Effizienz bei der Refaktorisierung...    
  Hits[HitCount].Faktor=sieb_delta.factor;

#ifdef USE_FAKEHEAP
  if (Sieve_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben?
#else
  sieb_delta.delta += sieb_delta.factor; // nchste Hit-Position
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
   {
     Sieve_Delta_Heap.push(sieb_delta); // wieder in Priorittswarteschlange eintragen
   }
  else
    if (Sieve_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben?
#endif
  
  sieb_delta = Sieve_Delta_Heap.top(); // nchsten Eintrag aus Priorittswarteschlange holen
  register int delta2; delta2 = sieb_delta.delta-SieveOffset;
  if (delta==delta2) goto Schleifenanfang_weiterer_Faktor;  
  if (SieveArray[delta]<0) // Relation gefunden?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      StaticRelations::insert(new CRelation(delta+SieveOffset, HitCount+1)); /* dann einfgen */
    }
  delta=delta2;
  
  HitCount=0; // und auf ein neues
 Schleifentest:
  if (delta<PhysicalSieveSize) goto Schleifenanfang_normal; // noch im aktuellen (physikalischen) Intervall?
  return; // fertig mit diesem Intervall
  
 Schlange_leer:
  if (SieveArray[delta]<0) // Relation gefunden?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      StaticRelations::insert(new CRelation(delta+SieveOffset,HitCount+1)); // dann einfgen
    }
  return; 
}


#if 0
#include <map>
void Werteverteilung_im_Sieb(void)
{
  // fr statistische Zwecke kann es ganz sinnvoll sein, wenn man
  // sich mal anschauen kann, wie die Werte um den Nullwert herum verteilt sind...
  typedef map<TSieveElement,unsigned int> TStats;
  static TStats S;
  for (int i=0; i<PhysicalSieveSize; ++i) S[SieveArray[i]]++;

  static int z = 1000;
  if (--z==0)
   {
     z=1000;

     for (TStats::iterator pos=S.begin(); pos!=S.end(); ++pos)
      cout << static_cast<signed long int>(pos->first) << " : " << pos->second << endl;

     cout << "Eingabe fr weiter" << endl;
     string s; cin >> s;
   }
}
#endif


void do_scanning_Sieve(void)
{
  //cout << "Auslesen des (gesiebten) Siebintervalls" << endl;

  // da nur ein Bruchteil der Siebelemente Treffer sind, sollte man das Suchen beschleunigen!
  // Constraint: the 64 positions below the sieve (SieveArray[-64]..SieveArray[-1]) are already set to -1.
  register int d=PhysicalSieveSize;
  do
   {
     // todo: hier bietet sich in assembler doch frmlich an...
#ifdef SIEBASM386
     asm_search_sieb(d); // nach Siebtreffer suchen...
     //if (d>=0 && SieveArray[d]<0) cout << "Siebtreffer fr " << d << endl;
     //else cout << "??? fr " << d << endl; // d==-1 is okay...
#else
     while (SieveArray[--d]>=0);; // nach Siebtreffer suchen...
#endif
     if (d>=0) // gefundener Treffer *nicht* Dummytreffer?
      {
        if (SieveControl::Hit_after_DirtyCorrection(SieveOffset+d,SieveArray[d]))
          StaticRelations::insert(new CRelation(SieveOffset+d)); // dann einfgen
      }
   } while (d>0); // solange noch regulre Treffer mglich: weitersuchen
// Werteverteilung_im_Sieb_ausgeben();
}

#endif /* #ifndef IS_SERVER */



/* ---------- Vorfaktorbestimmung ------------ */

// Um verschiedene Vorfaktorbestimmungsmethoden (mit jeweiliger Neu-Compilierung)
// testen zu knnen, ist die Prozedur " void determine_best_kN() "
// auf das File "Vorfaktorbestimmung.cc" ausgelagert.

#include "Vorfaktorbestimmung.cc"

/* ------------------------------------------- */

#include "mpqsPolynom.cc" // we still need to include the sources here
                          // (the module has been "outsourced", but it still includes
                          // some dependencies!)

#ifndef IS_SERVER
void recompute_all_Deltas()
{
  /* Beim Wechsel auf ein neues Polynom werden alle bisherigen Deltawerte der Primzahlen
     ungltig. Daher ist es erforderlich, diese neu zu berechnen */

  //cout << "Berechne die Deltas neu" << endl;

  // zunchst alle (mglicherweise) noch bestehenden dynamischen Faktoren aus der Warteschlange nehmen:
#ifdef USE_FAKEHEAP
//  cout << "clearing Fakeheap with " << Sieve_Delta_Heap.size() << " elements..." << endl;
  Sieve_Delta_Heap.clear();
#else
  while (!Sieve_Delta_Heap.empty()) Sieve_Delta_Heap.pop(); // fuer priority_queue, die kein clear hat
#endif

  // ebenso die Quadrate:
  //Sieve_Delta_HeapOfSquares.clear();
  while (!Sieve_Delta_HeapOfSquares.empty()) Sieve_Delta_HeapOfSquares.pop(); // fuer priority_queue, die kein clear hat
 
  
  // als erstes die Deltas aus der statischen Faktorbasis ermitteln
  unsigned int inv_A2_modp, PolyB;
  mpz_t inv_A2, P, wuq, x, y;
  mpz_init(inv_A2); mpz_init(P); mpz_init(wuq); mpz_init(x); mpz_init(y);


#if 1 /* Test!!!! 2004-03-09 */
  for (int i=FB_maxQuadrate+1; i<Size_StaticFactorbase && PrimeNumbers[i]<25000; ++i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
	  // Fr alle/einige "groe" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) parallel mit den dynamischen Faktoren sieben.
          const unsigned int P = PrimeNumbers[i]*PrimeNumbers[i];
          const unsigned int wuq = SQRT_kN_of_PrimeSquares[i];
          inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),P),P);
          PolyB=mpz_remainder_ui(Polynom.get_B(),P);
      
          // Delta1
          signed int d = wuq>=PolyB ? wuq-PolyB : wuq-PolyB+P; // wg. berlauf
          d=mulmod(d,inv_A2_modp,P);
          d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }

	  // Delta2
          unsigned int h = wuq+PolyB; d=P-mulmod(h,inv_A2_modp,P); d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif



#ifdef SIEVING_LARGE_SQUARES
  for (int i=Size_StaticFactorbase-1; i>FB_maxQuadrate && PrimeNumbers[i]>25000; --i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
	  // Fr alle "groen" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) zusammen mit den dynamischen Faktoren sieben.
	  mpz_set_ui(P,PrimeNumbers[i]); mpz_mul(P,P,P); // Quadrat der Primzahl holen
	  mpz_invert(inv_A2,Polynom.get_A2(),P);

	  // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
	  // Wurzeln von kN modulo Primzahl^2 ermitteln
	  const int Wu = SQRT_kN_of_PrimeNumbers[i];
	  mpz_set_ui(y,2*Wu);
	  if (mpz_invert(y,y,P)==0)
	    {
	      cerr << "PQ-Wurzel fr " << PrimeNumbers[i] << ": Inverses existiert nicht!" << endl;
	      exit(1);
	    }
	  
	  mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
	  mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG	  
	  // ggf. Prfung, ob Wurzel korrekt ist:
          mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
          if (mpz_cmp_ui(x,0)!=0)
           {
             cerr << "PQ-Wurzel nicht korrekt!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[i] << endl;
             cerr << "kN= " << kN << endl;
             cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[i] << endl;
             cerr << "SQRT(kN) mod p^2=" << wuq << endl;
             cerr << "aber falsch: " << x << endl;
             exit(1); 
           }
#endif
	  // nun die Deltas:

	  // Delta1
	  mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }
	  
	  // Delta2
	  mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif

  
  // Faktorbasis: erste Siebtreffer berechnen; Primzahl 2 berspringen, da Polynom damit Probleme hat...
  for (int i=Size_StaticFactorbase-1; PrimeNumbers[i]>2; --i)
    {
      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimeNumbers[i]),PrimeNumbers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimeNumbers[i]);
      
      register int d;
      d=SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][0]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);
      
      d=PrimeNumbers[i]-SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][1]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);

#ifdef DEBUG
      //ggf. testen, ob Werte in Ordnung sind...
      Polynom.get_values(Delta_of_PrimeNumbers[i][0],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0) 
	{ 
	  cerr << "Deltawert0 " << Delta_of_PrimeNumbers[i][0] << " liefert Teiler fr "
	       << PrimeNumbers[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_of_PrimeNumbers[i][0]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][0]<0) Delta_of_PrimeNumbers[i][0]+=PrimeNumbers[i];
	}
      Polynom.get_values(Delta_of_PrimeNumbers[i][1],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0)
	{ 
	  cerr << "Deltawert1 " << Delta_of_PrimeNumbers[i][1] << " liefert Teiler fr "
	       << PrimeNumbers[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_of_PrimeNumbers[i][1]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][1]<0) Delta_of_PrimeNumbers[i][1]+=PrimeNumbers[i];
	}
#endif
    }

#ifdef SIEVING_MORE_LARGE_SQUARES
  // und weitere Quadrate auerhalb der statischen Faktorbasis pushen:
  {
    unsigned int Primzahl = biggest_Prime_in_Factorbase;
    const unsigned int Obergrenze = 1000000; // or maybe 4*biggest_Prime_in_Factorbase;
    cout << "pushing more prime squares up to " << Obergrenze << endl; 
    while (Primzahl<Obergrenze)
     {
       do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; 
           while( Primzahl%3==0 || Primzahl%5==0 || Primzahl%7==0
                  || Primzahl%11==0 || Primzahl%13==0 || Primzahl%17==0
                  || !probab_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(P,Primzahl);
	}
       while (mpz_legendre(kN,P)<0); /* Primzahl gltig fr Primzahlbasis? */

       mpz_mul(P,P,P); // Quadrat der Primzahl holen
       mpz_invert(inv_A2,Polynom.get_A2(),P);

       // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
       // Wurzeln von kN modulo Primzahl^2 ermitteln
       const int Wu = SQRT_kN_mod_PrimeNumber(Primzahl);
       mpz_set_ui(y,2*Wu);
       if (mpz_invert(y,y,P)==0)
        {
          cerr << "PQ-Wurzel fr " << Primzahl << ": Inverses existiert nicht!" << endl;
          exit(1);
	}
	  
       mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
       mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG
       // ggf. Prfung, ob Wurzel korrekt ist:
       mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
       if (mpz_cmp_ui(x,0)!=0)
        {
          cerr << "PQ-Wurzel nicht korrekt!" << endl;
          cerr << "Primzahl=" << Primzahl << endl;
          cerr << "kN= " << kN << endl;
          cerr << "SQRT(kN) mod p=" << Wu << endl;
          cerr << "SQRT(kN) mod p^2=" << wuq << endl;
          cerr << "aber falsch: " << x << endl;
          exit(1); 
        }
#endif
       // nun die Deltas:

       // Delta1
       mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
          // Bemerkung: Die Primzahl wird im Sieb_Quadrate_Heap *und* 
          // zustzlich auch im Sieve_Delta_Heap abgelegt, wenn es
          // sich *nicht* um einen (bereits bekannten) dynamischen Faktor
          // handelt (denn dieser wird/wurde ja bereits dort vermerkt und
          // eine Doppelzhlung sollte vermieden werden).
          // Damit ist sichergestellt, dass
          //  - der Wert in den integer-Bereich pat
          //  - der Wert doppelt (d.h. als Quadrat) im Sieb vermerkt wird.
          // Der Effekt eines zu hufigen Siebens im Falle, dass
          // die Primzahl als dynamischer Faktor whrend des Siebens entdeckt
          // und dann nachtrglich eingefgt werden knnte, wird vernachlssigt.
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // und ablegen
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta1)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta1)" << endl; };
#endif
        }

       // Delta2
       mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // liegt der Wert im Intervall?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // und ablegen
          // siehe auch Bemerkung fr Delta1
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // und ablegen
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta2)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta2)" << endl; };
#endif
        }

     }
  }
#endif

#if defined(VERBOSE) || defined(DEBUG)
  if (!Sieve_Delta_HeapOfSquares.empty())
   cout << Sieve_Delta_HeapOfSquares.size() << " squares have been pushed." << endl;
#endif
  
  for (int i=0; i<NumberOf_more_PrimePowers; ++i)
    {
      // Fr diese Zahlen auch die Deltas bzgl. der jeweiligen Potenz ermitteln
      // (fr das Sieben zusammen mit den statischen Faktoren)

      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimePowers[i]),PrimePowers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimePowers[i]);

      register int d;
      d=SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][0]=mulmod(d,inv_A2_modp,PrimePowers[i]);
      
      d=PrimePowers[i]-SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][1]=mulmod(d,inv_A2_modp,PrimePowers[i]);
    }
  
  mpz_clear(inv_A2); mpz_clear(P); mpz_clear(wuq); mpz_clear(x); mpz_clear(y);
  
  // dann die dynamische Faktorbasis behandeln:
  TDynamicFactorRelations::iterator pos;
  for (pos=DynamicFactorRelations.begin();
       pos!=DynamicFactorRelations.end() && pos->sieveable(); ++pos) compute_Deltas_for(*pos, -LogicalSieveSize);
  
 // cout << "Deltaberechnung beendet." << endl;
}
#endif



#ifndef IS_CLIENT

#include "parse_term.cc"

// this is a predicate stating "false"
const bool without_multi_combine_init = false; // true ->multi-init, false ->no multi_init on StaticRelations::insert

bool SpecialRelations::insert(const CmpqsFactor &DLP, const string &GL_String)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren
  // und ins Gleichungssystem eintragen.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat! (in LAZY)" << endl;
     istringstream is(GL_String);
     CRelation* GL = new CRelation();
     GL->multi_combine_init();
     CmpqsFactor DLP_2 = GL->multi_combine_main(is);
     if (DLP!=DLP_2) { cerr << "DLP-Fehler irgendwo in LAZY!" << endl; exit(1); }
     statistical_data::Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     StaticRelations::insert(GL,without_multi_combine_init); // LAZY -> einfgen geschieht hier...
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  const TSpecialFactorRelations::iterator pos = SpecialFactorRelations.find(relation);
  if (pos==SpecialFactorRelations.end())
   {
     //cout << "Speichere DLP-Relation (LAZY)" << endl;
     //relation.fpos=GL->save(SpecialRelations_to_file,relation.factor);
     relation.fpos=SpecialRelations_to_file.tellp();
     SpecialRelations_to_file << GL_String;

     SpecialFactorRelations.insert(relation);
     relation.factor.swap();
     SpecialFactorRelations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
        SpecialRelations::CycleSearch();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit durch bekannten Special Factor! (LAZY)" << endl;
     istringstream is(GL_String);
     CRelation* GL = new CRelation();
     GL->multi_combine_init();
     CmpqsFactor DLP_2 = GL->multi_combine_main(is);
     if (DLP!=DLP_2) { cerr << "DLP-Fehler irgendwo in LAZY!" << endl; exit(1); }
     statistical_data::Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.in_mpz_ablegen(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->multi_combine_main(SpecialRelations_from_file,(*pos).fpos);
     //GL->multi_combine_exit() kann entfallen, wenn zugehriges init bei StaticRelations::insert entfllt
     StaticRelations::insert(GL,without_multi_combine_init); // LAZY -> einfgen geschieht hier...
     return false;
   }
}
#endif


#ifdef IS_STANDALONE
void process_ecm()
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // Random-Seed setzen (Zeitfunktion drfte hierzu zufllig genug sein)
  
  unsigned int dezimalstellen = mpz_sizeinbase(n,10);
  tune_parameters(dezimalstellen); // Parameter abhngig von der Gre der Zahl!

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // endl-berlesen
  while (Anzahl_Kurven<elcu_Kurven)
  {

    Anzahl_Kurven++;
    cout << "elliptic curve #" << Anzahl_Kurven << "/" << elcu_Kurven << endl;

    int ecm_sigma = rand(); // zufllige Kurve whlen
    elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
    elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); /* Faktorisierung mit elliptischer Kurve versuchen */

    // Anzahl bearbeiteter Kurven ins Recovery-File schreiben...
    Recovery_to_file.seekp(fpos); 
    Recovery_to_file << Anzahl_Kurven << endl << flush;

    // Teiler gefunden?
    if (dezimalstellen==mpz_sizeinbase(n,10)) continue; // kein neuer Faktor
    
    // Stelligkeit von n hat sich gendert -> neuer Faktor

    if (mpz_probab_prime_p(n,probab_prime_checks))
      {
        Factorization_to_file << MAL() << n;
        cout << "remaining factor: " << n << endl;
        cout << "remaining factor is most probably prime!" << endl;
        mpz_set_ui(n,1); // n ist faktorisiert
      }

    if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
      { 
        cout << "factorization successfully completed." << endl;
        Factorization_to_file << endl;
        ExitManager::StopFactorization(); // Programm erfolgreich beenden...
      }

    // fr weitere ecm-Runden Parameter neu tunen...
    dezimalstellen=mpz_sizeinbase(n,10);
    tune_parameters(dezimalstellen);

    // nun mu noch das recovery-File angepat werden, da sich
    // die zu faktorisierende Zahl verkleinert hat!
    Recovery_buffer.close(); // File schlieen
    Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc); // wieder ffnen (leeres File erzeugen!)
    Recovery_to_file << n << endl; // zu faktorisierende Zahl ablegen
    fpos = Recovery_from_file.tellg(); // Position merken
    Recovery_to_file << Anzahl_Kurven << endl << flush; // Anzahl bisher bearbeiteter Kurven vermerken

  }
  cout << "ECM-Phase (standalone) completed." << endl;
}
#endif



#ifdef IS_SERVER

void bearbeite_daten_stream_ecm(const connection_waiter& my_connection_waiter)
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // Random-Seed setzen (Zeitfunktion drfte hierzu zufllig genug sein)

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // endl-berlesen

  while (Anzahl_Kurven<elcu_Kurven)
  {
    unix_io_stream daten(my_connection_waiter); // auf Verbindung warten und client als stream verbinden

    daten >> s;

    if (s == "ECM?")
     {
       int ecm_sigma = rand();
       daten << ecm_sigma << " "
             << elcu_Phase1 << " "
             << elcu_Phase2 << " "
             << n << endl; // sigma und n bertragen

       Anzahl_Kurven++;
       cout_network << "elliptic curve #" << Anzahl_Kurven << "/" << elcu_Kurven
            << " with sigma=" << ecm_sigma
            << " sent to " << peer_info(daten) << endl;

       // Anzahl vergebener Kurven ins Recovery-File schreiben...
       Recovery_to_file.seekp(fpos); 
       Recovery_to_file << Anzahl_Kurven << endl << flush;

       daten << flush;
       continue;
     }

    if (s == "Faktor(ECM)!") // Faktoren annehmen...
      {
        cout_network << "Factor received from " << peer_info(daten) << endl;
        int sigma;
        mpz_t x;
        mpz_init(x);
        daten >> sigma >> x; // Kurveninformation (sigma) und Faktor holen
	while (isspace(daten.peek())) daten.get();
        string ecm_s = "ecm";
	if (daten.peek()!=EOF) daten >> ecm_s;
        mpz_gcd(x,x,n); // ggt bilden (Faktor knnte bereits entdeckt sein!)
        if (mpz_cmp_ui(x,1)!=0)
         {
           // theoretisch knnte Faktor mehrfach teilen...
	   const unsigned int exponent = mpz_remove(n,n,x);
           if (mpz_probab_prime_p(x,probab_prime_checks))
             {
               cout << x << " is factor." << endl;
               if (mpz_sizeinbase(x,10)<28)
                 Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << "]" << flush;
               else
                 Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << ",sigma=" << sigma << "]" << flush;
             }
           else
             {
               cout << x << " is a composite factor." << endl;
	       if (mpz_probab_prime_p(n,probab_prime_checks))
		{
		  if (exponent>1) mpz_pow_ui(x,x,exponent);
                  mpz_swap(n,x);
		  cout << x << " is factor. (factorswap)" << endl;
		  Factorization_to_file << MAL() << x << " [" << ecm_s << "/factorswap,sigma=" << sigma << "]" << flush;
                }
	       else
                {
                  if (mpz_sizeinbase(x,10)<28)
                   Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << "] [composite]" << flush;
                  else
                   Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << ",sigma=" << sigma << "] [composite]" << flush;
                }
             }
         }
        mpz_clear(x);
        daten << flush;

        if (mpz_probab_prime_p(n,probab_prime_checks))
          {
            Factorization_to_file << MAL() << n;
            cout << "remaining factor: " << n << endl;
            cout << "remaining factor is most probably prime!" << endl;
            mpz_set_ui(n,1); // n ist faktorisiert
          }

        if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
          { 
            cout << "factorization successfully completed." << endl;
            Factorization_to_file << endl;
            ExitManager::StopFactorization(); // Programm erfolgreich beenden...
          }
        tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning

        // nun mu noch das recovery-File angepat werden, da sich
        // die zu faktorisierende Zahl verkleinert hat!
        Recovery_buffer.close(); // File schlieen
        Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc); // wieder ffnen (leeres File erzeugen!)
        Recovery_to_file << n << endl; // zu faktorisierende Zahl ablegen
        fpos = Recovery_from_file.tellg(); // Position merken
        Recovery_to_file << Anzahl_Kurven << endl << flush; // Anzahl bisher bearbeiteter Kurven vermerken
	continue;
      }

    if (s == "Hunger!")
     {
       // requesting MPQS parameters is not adequate while processing ecm!
       cerr << "Inadequate request of MPQS parameters from " << peer_info(daten) << endl;
       daten << -1 << " " << "please wait until ecm is processed!" << endl;
       continue;
     }

    cerr << "Invalid request \"" << s << "\""
         << " from " << peer_info(daten) << endl;
    daten << flush;
  }
  cout << "distributed ECM-Phase (server) completed." << endl;
}


class Cprocess_clients
{
 private:
  static CMutex process;
  static void bearbeite_daten_stream(unix_io_stream &daten, const string client);
 public:
  static void* THREAD_bearbeite_daten_stream(void* arg);
};
CMutex Cprocess_clients::process;

void Cprocess_clients::bearbeite_daten_stream(unix_io_stream &daten, const string client)
{
  string s;
  daten >> s;
  
  if (s == "Hunger!")
    {
      cout_network << "Sending number to factor to " << client << endl;
      daten << n << " " << Size_StaticFactorbase << " " 
            << Factor_Threshold << " " << LogicalSieveSize << endl;
      return;
    }

  if (s == "ECM?")
   {
     daten << "-1" << endl; // in dieser Phase keine elliptischen Kurven (-> sigma<0)
   }

  if (s == "Faktor(ECM)!") // Faktoren werden noch akzeptiert...
    {
      cout_network << "Factor received from " << client << endl;
      int sigma;
      mpz_t x;
      mpz_init(x);
      daten >> sigma >> x; // Kurveninformation (sigma) und Faktor holen
      while (isspace(daten.peek())) daten.get();
      string ecm_s = "ecm";
      if (daten.peek()!=EOF) daten >> ecm_s;
      mpz_gcd(x,x,n); // ggt bilden (Faktor knnte bereits entdeckt sein!)
      if (mpz_cmp_ui(x,1)==0) // teilt "Faktor"?
       { mpz_clear(x); return; } // "Faktor" teilt nicht (mehr)

      if (mpz_cmp_ui(x,1)!=0)
       {
         // theoretisch knnte Faktor mehrfach teilen...
         const unsigned int exponent = mpz_remove(n,n,x);
         if (mpz_probab_prime_p(x,probab_prime_checks))
           {
             cout << x << " is factor." << endl;
             if (mpz_sizeinbase(x,10)<28)
               Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << "]" << flush;
             else
               Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << ",sigma=" << sigma << "]" << flush;
           }
         else
           {
             cout << x << " is a composite factor." << endl;
             if (mpz_sizeinbase(x,10)<28)
               Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << "] [composite]" << flush;
             else
               Factorization_to_file << MAL(x,exponent) << " [" << ecm_s << ",sigma=" << sigma << "] [composite]" << flush;
           }
       }
      mpz_clear(x);

      if (mpz_probab_prime_p(n,probab_prime_checks))
        {
          Factorization_to_file << MAL() << n;
          cout << "remaining factor: " << n << endl;
          cout << "remaining factor is most probably prime!" << endl;
          mpz_set_ui(n,1); // n ist faktorisiert
        }

      if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
        { 
          cout << "factorization successfully completed." << endl;
          Factorization_to_file << endl;
          ExitManager::StopFactorization(); // Programm erfolgreich beenden...
        }

      /* Falls der Programmflu diese Stelle erreicht, so kann das jetzt
         nur noch durch Fall 3 geschehen:
         Fall 1: Ein defektierender Client hat einen falschen Faktor
                 als ECM-Faktor gemeldet -> schadet sonst nicht weiter,
                 da er nicht abgespalten konnte.
                 (dieser Fall wird nunmehr schon oben abgefangen)
         Fall 2: Ein langsamer Client hat im ECM-Mode als "Nachzgler" einen
                 bereits entdeckten Faktor aufgrund seiner veralteten Daten
                 "wiederentdeckt" -> schadet sonst nicht weiter, da der
                 Faktor nicht abgespalten werden konnte.
                 (dieser Fall wird nunmehr schon oben abgefangen)
         !!!!!!
         Fall 3: Ein Client hat in seiner ECM-Phase einen neuen Faktor
         !!!!!!  entdeckt, der nicht zu einer kompletten Faktorisierung
                 gefhrt hat. -> Da sich der Server aber bereits in der
                 Siebphase befindet (sehr wahrscheinlich sogar noch in der
                 Anfangsphase), sind extreme Laufzeitverluste zu befrchten,
                 da das Sieb nicht auf die nun verkleinerte Zahl umgestellt
                 werden kann.
                 -> Richtig schlimm wird es aber erst, wenn das Programm
                 abgebrochen wird, da die Recovery-Daten nun inkonsistent
                 sind. (Abgespaltener Faktor kann nur in den Recovery-Daten
                 gespeichert werden, wenn die Siebdaten gelscht werden!
                 Umgekehrt wre es unklug, den gefundenen Faktor einfach
                 zu ignorieren.)
                 Bemerkung: Fall 3 lt sich beim verteilten Faktorisieren
                 nicht ausschlieen, da der Server keine Annahmen ber den
                 Status der Clients treffen kann. Es bringt also nichts,
                 die Siebphase solange verzgern zu wollen, bis alle Clients
                 den ECM-Mode verlassen haben. (Abgesehen davon wre das
                 dazu quivalent, alle ECM-Faktoren in der Siebphase einfach
                 zu ignorieren.)

           <- diese Signatur dient nur dem schnellen Auffinden von "Programm-Baustellen"
          Bercksichtigung insbesondere des dritten Falls ist noch nicht
          implementiert. Am besten wre es wohl, die Recovery-Daten anzupassen
          und das Programm mit Neustart-Aufforderung zu beenden.
          Andererseits wird durch den "RECOVERY"-Marker in "factorizations.txt"
          sowieso auf mgliche Inkonsistenzen aufmerksam gemacht...

          BUG entdeckt (2000-02-09):
          Programmflu hat diese Stelle erreicht, Faktor wurde von n
          abdividiert, aber: dieser Client siebt ja spter mit! Und - das
          ist das fatale - er berechnet aufgrund des kleineren n einen
          anderen Vorfaktor! -> klar, dass das zu inkonsistenten Daten fhrt...
          Da mu ich noch mal genau berlegen, wie man das am sinnvollsten
          beheben kann... (Man sieht mal wieder: Parallelitt fhrt zu
          sehr komplexen Fallunterscheidungen!)
          --> als Workaround Faktorisierung unvollstndig beenden.
      */

       {
          // Als Workaround wird jetzt erstmal die komplette Faktorisierung
          // mit dem Hinweis beendet, dass der Restfaktor noch zusammengesetzt
          // ist...

          Factorization_to_file << MAL() << n << " [composite]" << endl;
          cout << "remaining factor: " << n << endl
               << "remaining factor is composite!" << endl;
          cout << "factorization needs to be stopped to avoid inconsistent files." << endl
               << "please restart factorization with remaining number." << endl;
          cout << "Faktorisierung mu abgebrochen werden, da ansonsten" << endl
               << "Inkonsistenzen bei den Files auftreten werden." << endl
               << "Bitte starten Sie das Programm mit der verbliebenen Zahl erneut." << endl;

          Factorization_to_file << endl;
          ExitManager::StopFactorization(); // Programm erfolgreich beenden...
        }
                 
      return;
    }

  if (s=="DynamicFactors?_ab_index")
    {
      // this feature is new!
      // Online-clients should use this one to fetch dynamic factors.
      cout_network << "new-style dynamic factor request from " << client << endl;

      process.unlock();
      int i;
      daten >> i;
      const int start = i;

      cout_network << "Sending dynamic factors starting with index " << i
                   << " to " << client << endl;
      for (int j=0; j<20; ++j)
       { 
         const int i_start = i;
         while (i<DynamicFactorRelations.monitoredSize())
          {
            // binary transmission of data to increase throughput
            register unsigned int h=DynamicFactorRelations[i];
            char c[4];
            c[0]=h&0xff; c[1]=(h>>8)&0xff; c[2]=(h>>16)&0xff; c[3]=(h>>24)&0xff;
            //cout << "HUU: " << h << endl;
            daten.write(c,4); if (daten.fail()) break;
            //daten << DynamicFactorRelations[i] << " ";
            ++i;
          }
         //cerr << " ---- " << client << " ----  j = " << j << " -------" << endl;
	 pthread_testcancel();
         if (i>i_start)
          { 
            daten.flush();
            j=0;
          }
         if (daten.fail()) break;
         sleep(2);
         pthread_testcancel();
       }
      //daten << "-1" << endl;
      if (daten.fail()) cout_network << "dynamic factors thread: transmission failure!" << endl;
      else
       {
         daten.put(0); daten.put(0); daten.put(0); daten.put(0); daten.flush();
       }
      cout_network << "Dynamic factors from " << start << " to " << i-1
                   << " sent to " << client << "." << endl;
      if (!daten.fail())
       {
         // ggf. berschssige Daten auslesen
         for(int i=0; i<4; ++i)
          if (daten.peek()!=EOF)
           {
             int h=daten.get();
	     cerr << "unexpected value (" << h << ") instead of EOF on stream!" << endl;
           }
       }
      cout_network << "Connection to " << client << " closed." << endl;
      process.lock(); // it will be unlocked again by the calling function
      return;
    }
  
  if (s=="DynamicFactors?_ab_fpos")
    {
      cout_network << "old-style dynamic factor request from " << client << endl;
      // this feature is deprecated!
      // Online-clients should use the above one to fetch dynamic factors.
      // However, for offline clients and downwards compatibility with
      // client-versions up to 2.92 this feature is still provided...

      //streampos startfpos; // this is meant
      long long int startfpos; // and this will compile...
      daten >> startfpos;

      ostringstream temp_stringstream; // temporary to avoid long process locks
      cout_network << "Generiere dynamische Faktoren ab Fileposition " << startfpos << endl;
      const ostream::pos_type saved_fpos = DynamicRelations_to_file.tellp();
      DynamicRelations_to_file.seekp(0,ios::end);
#ifdef STL_STREAM_workaround
      daten << static_cast<streamoff>(DynamicRelations_to_file.tellp()) << " ";
#else
      daten << DynamicRelations_to_file.tellp() << " ";
#endif
      DynamicRelations_to_file.seekp(saved_fpos);
      
      int prev=0; // wir bertragen die Abstnde
      for (TDynamicFactorRelations::iterator pos=DynamicFactorRelations.begin();
	   pos != DynamicFactorRelations.end() && pos->sieveable(); ++pos)
#ifdef STL_STREAM_workaround
	if (static_cast<streamoff>((*pos).fpos)>=startfpos)
#else
        if ((*pos).fpos>=startfpos)
#endif
	  {
	    temp_stringstream << (*pos).factor-prev << " ";
            prev=(*pos).factor;
	  }

      cout_network << "Sending dynamic factors starting with fileposition " << startfpos << endl;
      process.unlock(); // allow other threads to sneak in
      if (!temp_stringstream.str().empty()) daten << temp_stringstream.str();
      daten << "-1" << endl;
      process.lock(); // it will be unlocked again by the calling function
      return;
    }
  
  if (s=="Polynom?")
    {
      int intervall;
      daten >> intervall;
      cout_network << "Sending new MPQS interval (" << intervall << ")"
                   << " to " << client << endl;
      Polynom.save(daten); // Intervallanfang
      Polynom.compute_next_polynomial(intervall);
      Polynom.save(daten); // Intervallende
      
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.save(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.flush(); // sicherheitshalber...
      Recovery_to_file.seekp(fpos);
      return;
    }

  // LAZY-Variante (ersetzt die nunmehr obsolete BUSY-Variante,
  // die bis einschlielich Version 2.92 als Option noch implementiert war):
  // Wenn die empfangenen Relationen Korrekturfaktoren enthalten, werden
  // diese erst dann ausgewertet, wenn eine statische Relation entsteht.
  // Diese Variante ist insbesondere fr das Client-Server-Modell
  // implementiert, da das relativ einfach ist und der Server alle
  // Relationen schnell entgegennehmen soll, was in der bisherigen
  // BUSY-Variante bei greren Gleichungssystemen nicht mehr gewhrleistet
  // werden konnte. Auerdem drfte die LAZY-Variante Plattenplatz sparen.
  // Nachteil: Der Check auf defektierende Relationen wird nun
  // unverhltnismig aufwendig. (Per Induktion gilt zwar: Wenn jede Relation
  // getestet wird, dann reicht es, alle Faktoren (einschlielich der 
  // Korrekturfaktoren) und den quadrierten Restfaktor zusammenzumultiplizieren
  // und das Produkt modulo n auf 1 zu testen. Die Korrektheit der zu den
  // Korrekturfaktoren gehrigen Relationen ergibt sich dann durch einen
  // Existenzcheck der zugehrigen Relation (und der Induktionsannahme, dass
  // diese korrekt ist). Dieser Test ist also einfach zu implementieren, aber
  // whrend in der Busy-Variante der Test nicht allzu sehr ins Gewicht fllt,
  // nimmt er hier einen groen Teil der Verarbeitungszeit ein, da die
  // Relation nun nicht mehr als Zeichenkette verarbeitet werden kann, sondern
  // tatschlich ausgewertet werden mu. Gerade aber die Auswertung sollte
  // durch die LAZY-Methode auf das notwendige Ma beschrnkt werden.
  //
  // Vorgehensweise der LAZY-Variante:
  //  1. Relation als String empfangen
  //  2. Relation analysieren (factor.Typ ermitteln)
  //  3. Falls statische Relation -> wie gehabt verfahren (BUSY)
  //  4. Analyse-Routinen wie gehabt ausfhren
  //  5. String in das zugehrige File ablegen.
  if (s == "NewRelations!")
    {
      // verify, that relations belong to our factorization
      mpz_t received_kN;
      mpz_init(received_kN);
      daten >> received_kN;
      if (mpz_cmp(received_kN,kN)!=0) daten << "VOID number!!" << endl;
      else daten << "proceed" << endl;
      mpz_clear(received_kN);

      daten >> s;
      if (s!="Relation!")
       {
         cerr << "expected block of relations, but got nonsense!" << endl;
         return;
       }

      bool Relationsblock_okay = true;
      int anz_empfangene_Relationen = 0;
      
      statistical_data::AllClientStats_Mutex.lock();
      statistical_data::CClientStats &ClientStats
       = statistical_data::AllClientStats[client.substr(0,client.find_first_of(' '))];
      statistical_data::AllClientStats_Mutex.unlock();
      // generate (or access) and reference a statistical "worksheet" for
      // the specified client (whicht is given by a string)

      // additional remark:
      // concurrent access is sequentialised by a MUTEX inside CClientStats!

      ClientStats.PutTimeStamp(); // put time stamp on for this connection

      do // Block von Relationen empfangen
        {
          process.unlock();
          // allow other threads to sneak in, because this
          // loop can take really a long time if we would do it atomicly!

          //cout << "Empfange bertragene Relation." << endl;

          ostringstream os; // hierin wird die Relation eingelesen

          //cout << "scanne Relation vom File" << endl;
          daten >> s; // Startzeichen
          if (s != "G")
           {
	     MARK;
             cerr << "error: wrong position while reading relation" << endl;
             exit(1);
           }

          CmpqsFactor factor;
          daten >> factor; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
          os << "G " << setprecision(20) << factor; // << " ";
          while (daten.peek()!='\n')
           {
             char line[1024];
             daten.get(line,sizeof(line),'\n'); os << line;
             if (daten.fail())
              {
                cerr << "stream in failstate for " << client << ", aborting thread." << endl;
                process.lock(); return;
              } 
           }
          { char c; daten.get(c); os << c; }
          //cout << os.str();
      
          //daten.ignore(1,'\n');
          daten >> s;

          while (s=="Relationsblock_Sync")
           {
             // client wishes to sync with server
             cout << "syncing with " << client << ", " << anz_empfangene_Relationen << " received so far." << endl;
             daten << "synced. " << flush;
             //cout << "synced with " << client << endl;
             daten >> s;
	     if (daten.fail()) { process.lock(); return; }
             ClientStats.PutTimeStamp(); // put time stamp on for this connection
             process.lock(); StatusReport(); process.unlock();
           }

          anz_empfangene_Relationen++;
          process.lock(); // still each step should be processed atomicly...
          
          ClientStats.increment(factor.Typ());
          if (factor.Typ()==CmpqsFactor::static_prime)
	   {
             // Static Factor
	     //cout << "Habe statische Relation empfangen!" << endl;
             istringstream is(os.str()); // eingelesenen String aus Eingabe zur Verfgung stellen
             CRelation* GL = new CRelation();
             GL->multi_combine_init();
             factor = GL->multi_combine_main(is);
	     ++statistical_data::Erfolgsaufteilung[0];
             //cout << "Fge Relation ein." << endl;
	     StaticRelations::insert(GL,without_multi_combine_init);
             continue;
	   }
          if (factor.Typ()==CmpqsFactor::single_large_prime)
	   {
	    // Dynamic Factor
	    TDynamicFactorRelation relation;
	    relation.factor = factor.int_value();
	    const TDynamicFactorRelations::iterator pos = DynamicFactorRelations.find(relation); // dynamischen Faktor suchen
	    if (pos == DynamicFactorRelations.end())
	      { // noch nicht bekannt
	        //relation.fpos = GL->save(DynamicRelations_to_file,relation.factor);
                relation.fpos=DynamicRelations_to_file.tellp();
                DynamicRelations_to_file << os.str(); // schnelles, weil direktes speichern
	        relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
	        DynamicFactorRelations.insert(relation);
	        SpecialRelations::split_by_primefactor(relation.factor);
	      } 
	    else
	      { // schon bekannt
	        //cout << "Kann dynamische Relation zu statischer kombinieren!" << endl;
                istringstream is(os.str()); // eingelesenen String aus Eingabe zur Verfgung stellen
                CRelation* GL = new CRelation();
                GL->multi_combine_init();
                factor = GL->multi_combine_main(is);
                ++statistical_data::Erfolgsaufteilung[1];
                {
                  mpz_t x; mpz_init(x);
                  factor.in_mpz_ablegen(x); // factor in mpz-Variable x ablegen
	          mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
                  mpz_clear(x);
                }
	        GL->multi_combine_main(DynamicRelations_from_file, (*pos).fpos);
                //GL->multi_combine_exit();
	        StaticRelations::insert(GL,without_multi_combine_init);
	      }
	    continue;
	   }
          if (factor.Typ()==CmpqsFactor::double_large_prime)
           {
             // Special-Factor

             // Specialfactor mit Dynamicfactoren testen:
             /* Anmerkung:
   	         Beim verteilten Sieben kann es passieren, dass ein (notwendigerweise zusammengesetzter)
   	         Special-Factor von einem Client entdeckt wird, der nicht mit dem aktuellen Satz der
   	         dynamischen Faktoren siebt. - In diesem Fall, der mit steigender Anzahl der Clients immer
   	         wahrscheinlicher wird, knnte es sinnvoll sein, Specialfaktoren auf ihre Teilbarkeit
   	         mit den vorliegenden dynamischen Faktoren des Servers zu testen und ggf. zu splitten.
   	         Allerdings bringt ein Verzicht auf diesen Test keine weiteren Nachteile mit sich.
   	         Wir haben daher auf diese - nur beim verteilten Sieben mgliche - Prozedur bisher verzichtet.

                 Nachtrag 2004-05-02:
                 Ist jetzt wieder aktuell geworden, da ich eine Option eingebaut habe,
                 nur bis zu einer Obergrenze der dynamischen Faktoren zu sieben. Der Rest der dynamischen
                 Faktoren verhlt sich also passiv und knnte durchaus im Special-Factor enthalten sein,
                 wenn der Client ihn nicht bereits entfernt und die Relation als Single-Large-Prime-Relation
                 klassifiziert hat. Und das kann durchaus passieren, wenn der Client nicht im Besitz der vollen
                 Menge der SingleLargePrimes ist!
                 Wir machen es jetzt auf die unelegante Tour mit einer vollen Breitseite:
                  -> zuerst wird die Relation als Special-Relation aufgenommen
                  -> Dann werden ihre beiden Teilfaktoren auf mgliche SingleLargePrimes berprft,
                     und (sollte das der Fall sein):
                      -> SpecialRelations::split_by_primefactor() aufgerufen.
              */
   	     SpecialRelations::insert(factor,os.str());
             if (is_dynamic_factor(factor.LP1())) SpecialRelations::split_by_primefactor(factor.LP1());
             if (is_dynamic_factor(factor.LP2())) SpecialRelations::split_by_primefactor(factor.LP2());
             continue;
           }

         cerr << "Error: Trash or a relation of unknown type has been received!" << endl;
         Relationsblock_okay=false;
        } while (s=="Relation!");
      // Relationsblock ist eingelesen
      cout_network << "Block of " << anz_empfangene_Relationen 
                   << " relations received from " << client << "." << endl;
      if (!Relationsblock_okay) daten << "ignoriert!" << flush;
      else
       if (s=="Relationsblock_Ende") daten << "empfangen." << flush;
       else
        {
          cerr << "Relationsblock-Endekennung nicht erkannt!" << endl;
          daten << "Relationsblock_korrekt_beendet???" << flush;
        }
      StatusReport();
      return;
    }

  cerr << "Invalid request \"" << s << "\""
       << " from " << client << endl;
}


void* Cprocess_clients::THREAD_bearbeite_daten_stream(void* arg)
{
  ExitManager::register_cancel();
  const int communication_socket = *static_cast<int*>(arg);
  delete static_cast<int*>(arg);
  unix_io_stream communication_stream(communication_socket);

  process.lock();
  //cout << "processing thread..." << endl;
  bearbeite_daten_stream(communication_stream,peer_info(communication_stream));
  //cout << "... thread processed!" << endl;
  process.unlock();
  communication_stream.flush();

  ExitManager::unregister_cancel();
  return 0;
}

#endif /* IS_SERVER */

#ifdef IS_CLIENT
#ifdef USE_NETWORK

class CClientPolynomFetcher
{
private:
  static string buffer;
  static void * THREAD_fetch_polynom(void *);
public:
  static void fetch(mpz_t Obergrenze_D);
};
string CClientPolynomFetcher::buffer;

void * CClientPolynomFetcher::THREAD_fetch_polynom(void *)
{
  static unsigned int PolyInterval = 1000;
  static time_t prev_time = 0;
  if (PolyInterval>500 && time(NULL)>prev_time+300) PolyInterval=100*(PolyInterval/200);
  else if (time(NULL)<prev_time+120) PolyInterval+=100*(PolyInterval/200);
  prev_time=time(NULL);

  unix_io_stream communication_stream2(communication_name, server_port);
  cout_network << "Fetching mpqs polynomial interval (" << PolyInterval << ")..." << endl;
  communication_stream2 << "Polynom? " << PolyInterval << endl;
  buffer.clear();
  while (!communication_stream2.eof())
   {
     char c;
     communication_stream2.get(c);
     buffer+=c;
   }
  cout_network << "background thread: polynomial fetched." << endl;
  return NULL;
}

void CClientPolynomFetcher::fetch(mpz_t Obergrenze_D)
{
  static pthread_t thread_fetch_polynom;
  int retcode;

  static bool first_time_first_love = true;
  if (first_time_first_love)
   {
     first_time_first_love=false;
     retcode = pthread_create(&thread_fetch_polynom, NULL, THREAD_fetch_polynom, NULL);
     if (retcode != 0)
      {
        cerr << "CClientPolynomFetcher: pthread_create failed!" << endl;
        exit(1);
      }
   }

  // Thread joinen
  cout_network << "Polynomfetch join..." << endl;
  retcode = pthread_join(thread_fetch_polynom, NULL);
  if (retcode != 0)
   {
     cerr << "CClientPolynomFetcher: Joining thread failed!" << endl;
     exit(1);
   }

  cout << "Setting mpqs polynomial interval..." << endl;

  // Polynomdaten vom internen Klassen-Buffer einlesen
  istringstream is(buffer);
  Polynom.load(is);
  cout << "Old Obergrenze_D = " << Obergrenze_D << endl;
  is >> Obergrenze_D;
  cout << "New Obergrenze_D = " << Obergrenze_D << endl;

  // und schonmal das nchste Intervall vielfdelnd im Hintergrund holen...
  retcode = pthread_create(&thread_fetch_polynom, NULL, THREAD_fetch_polynom, NULL);
  if (retcode != 0)
   {
     cerr << "CClientPolynomFetcher::pthread_create failed!" << endl;
     exit(1);
   }
}


class CClientDynamicFactorFetcher
{
  static CMutex Mutex;
  static queue<int> buffer;
private:
  static void * THREAD_fetch_DynamicFactors(void *);
public:
  static void fetch();
};
CMutex CClientDynamicFactorFetcher::Mutex;
queue<int> CClientDynamicFactorFetcher::buffer;

void * CClientDynamicFactorFetcher::THREAD_fetch_DynamicFactors(void *)
{
 int sleep_secs = 100;
 while(true)
  {
   {
     unix_io_stream communication_stream2(communication_name, server_port);
     cout << "Fetching dynamic factors from server..." << endl;
     static unsigned int dynfac_pos = 0;
     communication_stream2 << "DynamicFactors?_ab_index " << dynfac_pos << endl;

     int factor = 0;
     int counter = 0;
     while (true) // dynamische Relationen holen
      {
        //communication_stream2 >> factor;
        // cout << factor << "\r" << flush;
        char c[4];
        if (communication_stream2.peek()==EOF)
         {
           // magical mystery tour...
           // it seems that EOF is set sometimes although the stream
           // is still fully readable...
           // for that reason we clear the stream just before the
           // real read...
           if ( Cpoll(communication_stream2).readable_chars_within(128,2000) < 4)
             cerr << "possibly failing..." << endl;
           communication_stream2.clear();
         }
        communication_stream2.read(c,4);
        factor=static_cast<unsigned int>(static_cast<unsigned char>(c[0]));
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[1]))<<8;
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[2]))<<16;
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[3]))<<24;
        if (factor<=0 || communication_stream2.fail())
         {
           MARK;
           break;
         }
        Mutex.lock(); buffer.push(factor); Mutex.unlock();
        ++counter;
      }
     dynfac_pos+=counter;
     cout_network << "background thread: " << counter << " dynamic factors fetched." << endl;
     if (counter>200) sleep_secs-=sleep_secs/4;
     else if (counter<20 && sleep_secs<10000) sleep_secs+=(sleep_secs+3)/4;
   }
   cout_network << "next dynamic factor fetch request in " << sleep_secs << " seconds." << endl;
   sleep(sleep_secs); // ein wenig warten...
  }
 return NULL;
}


void CClientDynamicFactorFetcher::fetch()
{
  static pthread_t thread_fetch_DynamicFactors;

  int retcode;
  static bool first_time_first_love = true;

  if (first_time_first_love)
   {
     first_time_first_love=false;
     retcode = pthread_create(&thread_fetch_DynamicFactors, NULL, THREAD_fetch_DynamicFactors, NULL);
     if (retcode != 0)
      {
        cerr << "CClientDynamicFactorFetcher: pthread_create fehlgeschlagen!" << endl;
        exit(1);
      }
   }

  Mutex.lock();
  if (buffer.empty())
   {
     Mutex.unlock();
     return;
   }

#ifdef VERBOSE
  cout << "Inserting " << buffer.size() << " dynamic factors..." << endl;
#endif

  // Dynamische Faktoren vom internen Klassen-Buffer einlesen
  TDynamicFactorRelation relation;
  relation.fpos=0; relation.factor=0;
  while (!buffer.empty()) // dynamische Relationen holen
   {
     relation.factor=buffer.front(); buffer.pop();
     // cout << relation.factor << "\r" << flush;
     relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
     DynamicFactorRelations.insert(relation);
   }
  Mutex.unlock();
}


class CClientRelation_Versand
{
private:
  static volatile bool Sende_Relationen; // consumer/producer trigger
  static int  Sende_Nr;
  static string file_name;
  static ifstream ConsumerFile;
  static CMutex Consumer;

  static void init(void);
  static const string create_unique_tempfilename(const string servername);
  static void cleanup_files(void);
  static void * THREAD_uebertrage_Relationen(void * arg);
  static void go(void);

  friend int main(const int argc, const char* const argv[]);
};
volatile bool CClientRelation_Versand::Sende_Relationen = true; // consumer/producer trigger
int  CClientRelation_Versand::Sende_Nr = 0;
string CClientRelation_Versand::file_name;
ifstream CClientRelation_Versand::ConsumerFile;
CMutex CClientRelation_Versand::Consumer;


const string CClientRelation_Versand::create_unique_tempfilename(const string servername)
{
  ostringstream ostr;
  pid_t pid = getpid();
  struct utsname buf;
  uname(&buf);
  ostr << "/tmp/qsieve-" << buf.nodename
       << "-" << servername << "_" << pid;
  return ostr.str();
}

void CClientRelation_Versand::init(void)
{
  // first check whether the host is reachable
  // and get the hostname

  // refer "man getaddrinfo" and "man socket" 
  struct addrinfo hints; // our wishes are placed here
  memset(&hints,0,sizeof(hints)); // must be zeroed for default options
  hints.ai_family=PF_INET; // we want IPv4 as protocol
  hints.ai_socktype=SOCK_STREAM; // and we need a stream, not datagram!
  struct addrinfo *addrinfo_res = NULL; // here the result will be delivered
  const int retval = getaddrinfo(communication_name.c_str(),NULL,&hints,&addrinfo_res);
  if ( retval || addrinfo_res==NULL ) // any error?
   {
     cerr << "can't reach " << "\"" <<  communication_name << "\"" << endl;
     cerr << "Error given by getaddrinfo: " << endl;
     cerr << gai_strerror(retval) << endl;
     exit(1);
   }
  if (addrinfo_res->ai_socktype!=SOCK_STREAM) // we got a "stream"-protocol?
   {
     cerr << "provided protocol doesn't support SOCK_STREAM" << endl;
     exit(1);
   }

  freeaddrinfo(addrinfo_res); // free ressources allocated by getaddrinfo

  // create an unique tempfilename based on the hostname
  file_name = create_unique_tempfilename(communication_name);
}

void CClientRelation_Versand::cleanup_files(void)
{
  string name;
  name = file_name+"A";
  unlink(name.c_str());
  name = file_name+"B";
  unlink(name.c_str());
}

void * CClientRelation_Versand::THREAD_uebertrage_Relationen(void *)
{
  // important: this thread shall not run concurrently to itself!
  static int retval; // &retval is returned to thread who is joining
  string s;
  char line[1024];

  retval = -1; // default: return an error
  Consumer.lock();

  unix_io_stream communication_stream2(communication_name, server_port);
  
  communication_stream2 << "NewRelations! " << kN << endl;
  // verify, that relations belong to our factorization
  communication_stream2 >> s;
  if (s!="proceed")
   {
     cerr << "Oops! Server will not accept my relations: " << endl;
     cerr << s << endl;
     goto fertig;
   }

  if (!ConsumerFile.is_open())
   {
     cerr << "ConsumerFile is not open!" << endl;
     goto fertig;
   }

transmit:
  while (ConsumerFile.peek()!=EOF)
   {
     int count=1000+1; // maximal count-1 Relationen am Block bertragen
     const int max_Uebertragungsdauer = 8; // "Maximale Sendezeit" fr Teilblock in Sekunden (damit auch andere drankommen knnen...)
     const time_t Uebertragungsstart = time(NULL);
     while (ConsumerFile.peek()!=EOF && --count && time(NULL)-Uebertragungsstart<=max_Uebertragungsdauer)
      {
        communication_stream2 << "Relation! ";
	while (ConsumerFile.peek()!='\n')
	 {
	   ConsumerFile.get(line,sizeof(line),'\n');
	   communication_stream2 << line;
	 }
	char c; ConsumerFile.get(c);
	communication_stream2 << endl;
      }
     
     if (communication_stream2.fail())
      {
	MARK;
        cerr << "stream state indicates that transmission had failed!" << endl;
        goto fertig;
      }

     if (ConsumerFile.peek()!=EOF)
      {
        // synchronize communication with server
        communication_stream2 << "Relationsblock_Sync " << flush;
        communication_stream2 >> s;
        if (s!="synced.")
         {
           cerr << "sync with server failed: " << endl; cerr << s << endl;
           goto fertig;
         }
      }
   }
  ConsumerFile.clear(); ConsumerFile.close();

  // synchronize communication with server
  communication_stream2 << "Relationsblock_Sync " << flush;
  communication_stream2 >> s;
  if (s!="synced.")
   {
     cerr << "sync with server failed: " << endl; cerr << s << endl;
     goto fertig;
   }

  Consumer.unlock();
  sleep(5);
  Consumer.lock();
  if (ConsumerFile.is_open()) goto transmit;

  communication_stream2 << "Relationsblock_Ende " << flush;
  communication_stream2 >> s;
  if (s!="empfangen.")
    cerr << "Transmission of relations not acknowledged!" << endl;
  if (s=="ignoriert!")
   {
     cerr << "Oops! - Transmitted relations were rejected by the server!" << endl;
     goto fertig;
   }
  sleep(60); // damit bei schnellen Rechnern nicht permanent neue Verbindungen geffnet werden
  retval=0;
fertig:
  Sende_Relationen=true; // neuen Thread fr bertragung anstoen...
  Consumer.unlock();
  return &retval;
}

void CClientRelation_Versand::go(void)
{
  if (!communication_stream)
   {
     MARK;
     cerr << "communication stream is in error state! aborting..." << endl;
     exit(1);
   }

#ifdef STL_STREAM_workaround
  if ( (communication_stream.tellp()==std::streampos(0)) || (communication_stream.tellp()==std::streampos(-1)) )
#else
  if (communication_stream.tellp()<1)
#endif
   {
#ifdef VERBOSE
     MARK;
     cout << "no produced data..." << endl;
#endif
     return;
   }

  static pthread_t thread_Relationen_versenden;
  static const char* const  SendeSuffix[2] = { "A", "B" };

  if (Sende_Relationen)
   {
     // ready to transmit relations to server:
     // - set flag to trigger transmission: -> Sende_Relationen=false
     // - create thread for consuming and transmitting the data
     // - the consumer-thread will reset the flag after he has finished his job
     // - if the thread indicates, he had done his work, join him and create a new one
     
     Sende_Relationen = false;
     
     // close file (producer)
     communication_stream.close();

     // open file (consumer)
     const string ConsumerFileName = file_name + SendeSuffix[Sende_Nr];
     ConsumerFile.open(ConsumerFileName.c_str(), ios::in);
     if (!ConsumerFile)
      {
        MARK; cerr << "opening consumer file \"" << ConsumerFileName << "\" failed!" << endl;
        exit(1);
      }

     int retcode;
     void* thread_retval = NULL;

     static bool first_time_first_love = true;

     if (first_time_first_love)
      { // vor dem ersten Thread nicht joinen, ...
        first_time_first_love=false;
        retcode = pthread_create(&thread_Relationen_versenden, NULL,
                    CClientRelation_Versand::THREAD_uebertrage_Relationen, NULL);
      }
     else
      { // ... danach schon...
        retcode = pthread_join(thread_Relationen_versenden, &thread_retval);
        if (retcode != 0)
         {
           cerr << "Joining Consumer-Thread failed!" << endl;
           exit(1);
         }
        else
         {
           if (thread_retval && *(static_cast<int*>(thread_retval)) ) // was a value returned?
            {
              // this means, an error occurred...
              MARK;
              cerr << "Consumer-Thread returned an error value: " << *(static_cast<int*>(thread_retval)) << endl;
              exit(1);
            }
         }
        retcode = pthread_create(&thread_Relationen_versenden, NULL,
                    CClientRelation_Versand::THREAD_uebertrage_Relationen, NULL);
      }

     if (retcode != 0)
      {
        cerr << "Thread zum bertragen der Relationen fehlgeschlagen!" << endl;
        exit(1);
      }

     Sende_Nr = 1 - Sende_Nr;
     // open new file (producer)
     const string name = file_name + SendeSuffix[Sende_Nr];
     communication_stream.open(name.c_str() ,ios::out|ios::trunc);
   }
  else
   {
     // consumer thread is transmitting or waiting for data or just exiting
     if (Consumer.trylock())
      {
        // remember, that it is unlikely, but possible, that Sende_Relationen==true now,
        // (we could get interrupted after checking its value!)
 
        if (Sende_Relationen || ConsumerFile.is_open()) { Consumer.unlock(); return; }

        // well, now we know that he's waiting :-)
        // and that means that we can give him relations to consume...

        // close file (producer)
        communication_stream.close();

        // open file (consumer)
        const string ConsumerFileName = file_name + SendeSuffix[Sende_Nr];
        ConsumerFile.open(ConsumerFileName.c_str(), ios::in);
        if (!ConsumerFile)
         {
           MARK; cerr << "opening consumer file \"" << ConsumerFileName << "\" failed!" << endl;
           exit(1);
         }

        Sende_Nr = 1 - Sende_Nr;
        // open new file (producer)
        const string name = file_name + SendeSuffix[Sende_Nr];
        communication_stream.open(name.c_str() ,ios::out|ios::trunc);

        Consumer.unlock();
      }
   }
}

#endif
#endif


#ifdef USE_NETWORK
void kill_all_other_threads()
{
  //cout << "Killing explicitely all other threads..." << endl;
  //pthread_kill_other_threads_np();
}
#endif

void cleanup_files()
{
  cout << "cleaning files..." << endl;
  StaticRelations::cleanup_files();
  SpecialRelations::cleanup_files();

  // close streams
  DynamicRelations_buffer.close();
  Recovery_buffer.close();

  // delete files
  remove(DynamicRelationsFile.c_str());
  remove(RecoveryFile.c_str());
}

void cleanup_memory()
{
  cout << "cleanup allocated memory" << endl;
  StaticRelations::cleanup_memory();
  cout << "cleanup mpz numbers" << endl;
  mpz_clear(r); mpz_clear(kN); mpz_clear(n);
  cout << "mpz numbers cleared." << endl;
}


#ifdef IS_SERVER

void catch_sigint(int )
{
  exit(1);
}

void signalhandler(int signr)
{
  cout << "Signal " << signr << " received" << endl;
  exit(1);
}

#endif


#ifdef USE_NCURSES


class Cncursed
{
private:
  static Cncursed* myself;
  static void suicide() { delete myself; }

  typedef streambuf* Pstreambuf;
  typedef Cwin* Pwin;
  Pstreambuf cout_titel_buf, cout_network_buf, cout_status_buf,
             cerr_buf, clog_buf, cout_buf;
  Pwin window1, window2, window3;
public:
 Cncursed()
 {
   if (myself)
    {
      MARK;
      cerr << "You may not call me twice, I'm unique!" << endl;
      exit(93);
    }
   myself=this; // I am myself, really!
   init_my_ncurses();
   window1 = new Cwin(8,0,0,0);
   window2 = new Cwin(8,0,8,0);
   window3 = new Cwin(0,0,16,0);
   cout_titel_buf=cout_titel.rdbuf(window1->rdbuf());
   cout_network_buf=cout_network.rdbuf(window1->rdbuf());
   cout_status_buf=cout_status.rdbuf(window2->rdbuf());
   cerr_buf=cerr.rdbuf(window2->rdbuf());
   clog_buf=clog.rdbuf(window2->rdbuf());
   cout_buf=cout.rdbuf(window3->rdbuf());
   atexit(suicide);
 }
 ~Cncursed()
 {
   cout_titel.rdbuf(cout_titel_buf);
   cout_network.rdbuf(cout_network_buf);
   cout_status.rdbuf(cout_status_buf);
   cerr.rdbuf(cerr_buf);
   clog.rdbuf(clog_buf);
   cout.rdbuf(cout_buf);
   delete window1; delete window2; delete window3;

   exit_my_ncurses();

   // print statistical information before exiting,
   // since screen should be cleared now!
   display_StatusLegend();
   StatusReport(true);
 }
};
Cncursed* Cncursed::myself = NULL;
#endif

int main(const int argc, const char* const argv[])
{

#ifdef USE_NETWORK
 atexit(kill_all_other_threads);
#endif

#ifdef USE_NCURSES
  new Cncursed(); // trigger activation of ncursed streams
#endif

  cout.setf(ios::fixed); // Dezimaldarstellung, nicht Scientific!

 {
   string name = "Faktorisierungsprogramm ";
   name+=VERSION;
   string author = "written 2004-07 by Thorsten Reinecke";
   string email = "(thre@thorstenreinecke.de)";
   const int L = 5+MAX(name.length(),MAX(author.length(),email.length()));
   const string sep(L,'-');
   int f;
   f=(L-name.length())>>1; name=string(f,' ')+name+string(f+1,' '); name.resize(L);
   f=(L-author.length())>>1; author=string(f,' ')+author+string(f+1,' '); author.resize(L);
   f=(L-email.length())>>1; email=string(f,' ')+email+string(f+1,' '); email.resize(L);
   cout_titel << " -" << sep << "-" << endl;
   cout_titel << "- " << name << " -" << endl;
   cout_titel << "- " << author << " -" << endl;
   cout_titel << "- " << email << " -" << endl;
   cout_titel << " -" << sep << "-" << endl;
 }
#ifdef __GNUG__
  {
    const string s=__VERSION__;
    cout << "(" << __DATE__ << ", " << __TIME__ << " using ";
    if (s[0]<='9') cout << "gcc ";
    cout << __VERSION__ << ")" << endl;
  }
#endif
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc>2)
    {
#ifdef IS_CLIENT
       cerr << "servername expected!" << endl;
#else
      cerr << "number for factorization expected!" << endl;
#endif
      exit(1);
    }
  
  const bool recover = (argc==1); // ohne Argument: Recovery-Modus

  {
   // resolve ConfigFile
   if (getenv("QSIEVE_CFG"))
    {
      ConfigFile=getenv("QSIEVE_CFG");
      cout << "Setting ConfigFile to \"" << ConfigFile << "\"" << endl;
    }
   // and convert it to real/absolute path
#ifdef PATH_MAX
   char resolved_path[PATH_MAX];
#else
   char resolved_path[4096];
#endif /* PATH_MAX */
   if (realpath(ConfigFile.c_str(),resolved_path))
    ConfigFile=resolved_path;
   else
    {
      cerr << "Cannot find ConfigFile " << "\"" << ConfigFile << "\"" << endl;
      // server & standalone versions need a configuration file,
      // clients do not need them
#ifndef IS_CLIENT
      exit(1);
#endif
    }
  }

  mpz_init(n); /* Variable fr zu faktorisierende Zahl bereitstellen */
  mpz_init(kN);
  mpz_init(r);
  ExitManager::register_exithandler(cleanup_memory); // on sucessful exit free allocated data

#ifndef IS_CLIENT
  Read_ConfigFile();

#ifdef IS_SERVER
  // establish the network service at this point,
  // to allow clients to connect
  const connection_waiter my_connection_waiter(server_port);
#endif

  if (!recover) // Nicht im Recovery-Modus -> Argumente auswerten
    { 
      // bergebene Zahl holen (und ggf. ausrechnen)
      char* const neuer_str = strdup(argv[1]);
      char* str = neuer_str; 
      if (!parse_term::get_number(n, str))
	{
	  cout << "Wrong input at: '" << str << "'" << endl;
	  exit(1);
	}
      else
	if (str[0]!='\0')
	  {    
	    cout << "Syntax error in input term. (parenthesis?)" << endl;
	    exit(1);
	  }
	else
	  if (mpz_cmp_ui(n,0)<=0)
	    {
	      cout << "Input must be positive natural number!" << endl;
	      exit(1);
	    }
      free(neuer_str); // kein delete [] wegen stdup/malloc
    }

  if (recover)
    {
      // Recovery: continue a factorization, which was previously aborted
      // (try to) open the necessary file streams
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::in|ios::out);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }
      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::ate);

      // offenbar wird trotz ios::ate nicht (immer) ans Ende positioniert
      // deshalb wird die Testabfrage modifiziert:
      if (Recovery_from_file) Recovery_from_file.seekg(0,ios::end);
#ifdef STL_STREAM_workaround
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()==std::streampos(0)) || (Recovery_from_file.tellg()==std::streampos(-1)) )
// tellg()==0 indicates empty file -> we cannot recover
// tellg()==-1 indicates failure -> we cannot recover
// remark:
//  in gcc 3.4 (cvs-experimental-2003-10-17 we cannot compare with "<" !!)
//  do we really need a workaround to check this condition? 
#else
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()<1) )
#endif /* STL_STREAM_workaround */
	{
	  cerr << "Recovery not possible!" << endl;
	  exit(1);
	}
      
      Recovery_from_file.seekg(0,ios::beg);
      Recovery_to_file.seekp(0,ios::beg);
      Recovery_from_file >> n; // Zahl holen
      Recovery_from_file.ignore(1,'\n');

      cout << "Continue factorizaction for " << endl;
      cout << n << endl;
      // auerdem Recovery-Modus im Factorization-File vermerken, denn es
      // 1. knnte gelscht sein
      // 2. knnte der Recovery-Modus in seltenen Fllen Faktoren unterdrcken
      //    (wenn vorher das Schreiben unterbrochen wurde)
      // 3. knnte der Recovery-Modus zu doppelten Faktor-Ausgaben fhren
      //    (falls das quadratische Sieb einen zusammengesetzten Faktor
      //    gefunden hat und dann noch unterbrochen wird)
      // 4. knnte das Multiplikationssymbol inkorrekt hufig auftreten
      //    (wenn vorher kein Faktor gefunden wurde)
      Factorization_to_file << " [RECOVERY] "; // ohne flush, falls vor Finden weiterer Faktoren erneut abgebrochen wird... 
    }
  else
    {
      // Neustart: Starten einer neuen Faktorisierung
      // Streams ffnen
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::out|ios::trunc);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out|ios::trunc);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out|ios::trunc))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }

      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc);
      
      cout_status << "Starting factorization for" << endl;
      cout_status << n << endl;
      
      Factorization_to_file << endl << "Factorization of " << argv[1] << ":" << endl;
      Factorization_to_file << n << " = " << endl;
      
      easy_factor(); // "einfache" Faktoren rausstreichen
      
      cout_status << "Starting factorization with ECM and MPQS for" << endl;
      cout_status << n << endl;
      Recovery_to_file << n << endl;
    }
  ExitManager::register_exithandler(cleanup_files); // on sucessful exit delete the files
#ifdef IS_SERVER
  signal(SIGINT, signalhandler);
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning

#ifdef NOTIFY_PARENT
  // notify parent that now the distributed phase starts
  // and the server will process client connections
  kill(getppid(),SIGUSR1); // send SIGUSR1 to the parent process
#endif

  if (!SkipECM)
   {
     cout << "Waiting for net-clients (ECM)..." << endl;
     bearbeite_daten_stream_ecm(my_connection_waiter); // erstmal ECM verteilt starten...
   }
#endif /* IS_SERVER */

#ifdef IS_STANDALONE
  // Auch die Standalone-Version sollte elliptische Kurven benutzen.
  if (!SkipECM) process_ecm(); // erstmal ECM starten...
#endif /* IS_STANDALONE */

#ifdef USE_DFT
  // an dieser Stelle sind in allen Programmvarianten die
  // Faktorisierungsalgorithmen, die discrete fast fourier transform
  // verwenden knnten, abgeschlossen.
  // Das DFT-Objekt kann also wieder freigegeben werden...
  polynomial::clear_dft_tempmemory(); // Lschen des DFT-Objektes triggern...
#endif /* USE_DFT */

  // nun wird alles fr das Quadratische Sieb vorbereitet...

  determine_best_kN(); /* optimaler Wert, fr den gesiebt werden kann */

  /* Streitfall:
     Soll man nach Ermittlung des bestem Multipliers die Siebparameter
     noch tunen?
     Ja   -> denn die Zahl kN hat ja die entsprechende Gre...
     Nein -> denn fr die Wahl des Multipliers wurden diese Gren nicht
             bercksichtigt und entscheidend sollte die Gre der
             zu faktorisierenden Zahl sein. Nur so ist eine Vergleichbarkeit
             der Multiplier unter gleichen Voraussetzungen zu gewhrleisten.
    Wie auch immer:
     Eine "optimale" Ermittlung der Siebgre und der Faktorbasis kann nur
     mittelbar von der Gre der Zahl abhngen. Die Laufzeit der Siebphase
     hngt davon ab, wie viele Relationen zu ermitteln sind *und* wie
     schnell die Relationen ermittelt (=gesiebt) werden knnen!
     Letzterer Wert ist funktional von der Gre der Zahl und der Qualitt
     der Faktorbasis abhngig. Diese Qualitt ist zwar durch Vorgabe eines
     Multipliers (nach Annahme) maximiert worden; aber das heit noch nicht,
     dass dadurch die funktionale Abhngigkeit beseitigt wre.
     Mit anderen Worten: Wenn wir die Bestimmung der Faktorbasisgre und
     der Siebgre *nur* von der Gre der Zahl abhngig machen, so gehen
     wir idealisierend von einer einheitlichen Qualitt der Faktorbasis aus.
     Da wir diese einheitliche Qualitt der Faktorbasis nicht definiert haben,
     besitzen wir auch keine Entscheidungsgrundlage darber, ob fr unsere
     Zahl (n bzw. kN) eine unterdurchschnittliche, durchschnittliche oder
     berdurchschnittliche Faktorbasisqualitt besteht. Aus dieser Sichtweise
     erscheint der obige Streitfall nicht entscheidbar zu sein.
  */

#if 0   
  tune_parameters(mpz_sizeinbase(kN,10)); // Parametertuning fr kN
#else
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning fr n
#endif
  /* Hinweis:
      Beim Tuning der Parameter ist in der verteilten Version auf Konsistenz
      zu achten, d.h. ein Client sollte keine Parameter tunen!
      Unterschiedliche cfg-Files auf verschiedenen Rechnern wrden sonst
      zu Problemen fhren!
  */

#else /* IS_CLIENT */
  if (recover)
    {
      cerr << "Fehlender Parameter" << endl;
      exit(1);
    }
  communication_name = argv[1];
#ifdef USE_NETWORK
  CClientRelation_Versand::init();

  { // zunchst elliptische Kurven abarbeiten (net-client):
    int ecm_sigma = -1, runde=0;
    while (true)
     {
       unix_io_stream communication_stream(communication_name, server_port);
 waitloop:
       communication_stream << "ECM?" << endl;
       communication_stream >> ecm_sigma;
       if (ecm_sigma==-3)
        {
          // -3 is special token for idle waiting loop
          string s;
          communication_stream >> s;
          cout << s << endl;
          goto waitloop;
        }

       if (ecm_sigma<0) break; // sigma<0 -> keine Kurven mehr!
       communication_stream >> elcu_Phase1 >> elcu_Phase2;
       communication_stream >> n; // Zahl holen
       cout << "elliptic curve " << ++runde << endl;
       elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
       elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); /* Faktorisierung mit elliptischer Kurve versuchen */
     }
  }
  
  {
    unix_io_stream communication_stream(communication_name, server_port);
    communication_stream << "Hunger!" << endl;
    communication_stream >> n;
    communication_stream >> Size_StaticFactorbase;
    communication_stream >> Factor_Threshold;
    communication_stream >> LogicalSieveSize;
    cout << "n=" << n << endl;
    cout << "Gre der statischen Faktorbasis: " << Size_StaticFactorbase << endl;
    cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
    cout << "Siebintervall pro Polynom: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
  }
  {
    string name;
    name = CClientRelation_Versand::file_name+"A";
    communication_stream.open(name.c_str(), ios::out|ios::trunc);
  }
  atexit(CClientRelation_Versand::cleanup_files); // Vorsicht: Bei Absturz sind dann noch nicht bertragene Relationen futsch!
#else /* no USE_NETWORK */
  // existiert das file?
  ifstream communication_stream2(communication_name.c_str());
  if (!communication_stream2)
    {
      cerr << "Kann File " << communication_name << " nicht oeffnen" << endl;
      exit (1);
    }
  communication_name+=".out";
  
  communication_stream2 >> n; 
  communication_stream2 >> Size_StaticFactorbase;
  communication_stream2 >> Factor_Threshold;
  communication_stream2 >> LogicalSieveSize;
  communication_stream2.ignore(1,'\n');
  cout << "n=" << n << endl;
  cout << "Gre der statischen Faktorbasis: " << Size_StaticFactorbase << endl;
  cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
  cout << "Siebintervall pro Polynom: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
  communication_stream.open(communication_name.c_str(), ios::out|ios::trunc);
#endif /* USE_NETWORK */
  determine_best_kN(); /* optimaler Wert, fr den gesiebt werden kann */
#endif /* not IS_CLIENT */
  
  if (mpz_cmp_ui(r,PhysicalSieveSize)<0)
    {
      cerr << "Sieb ist berdimensioniert (ggf. verkleinern)!" << endl;
      exit(1);
    }

  // Fr Double-Large-Primes einen Schwellwert setzen,
  // dieser ist das Quadrat des maximalen Specialfactors...
  mpz_init(CmpqsFactor::DLP_Threshold);
  mpz_set_ui(CmpqsFactor::DLP_Threshold,SingleLargePrime_Threshold);
  mpz_mul(CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold);
  
  compute_StaticFactorbase();
  cout << "The first 20 members of the static factorbase are" << endl;
  for (int i=0; i<20; i++) cout << PrimeNumbers[i] << " "; cout << endl;
  cout << "Biggest prime in static factorbase: " << biggest_Prime_in_Factorbase << endl;
  cout << "#squares (of primes) in FB: " << FB_maxQuadrate << endl;
  cout << "#powers (of primes) in FB: " << NumberOf_more_PrimePowers << endl;

  SieveControl::compute_SieveThreshold(); // Schwellwert zur Erkennung der Relationen fr Siebvorgang bestimmen
  for (int i=0; i<64; ++i) SieveArray_[i]=-1; // initialize array underflow trigger values

  Polynom.compute_first_polynomial(kN,LogicalSieveSize); // erstes Polynom ermitteln

#if defined(IS_CLIENT) && !defined(USE_NETWORK)
 // file-client: write kN to file
 communication_stream << kN << endl;
#endif
  
#ifndef IS_CLIENT
  if (recover)
    {
      // Static Relations einlesen
      cout << "Reading static relations from file..." << endl;
      StaticRelations::Load();
      cout << "Recover: " << StaticRelations::Count() << " static relations read." << endl;

      // Dynamic Relations einlesen
      cout << "Reading dynamic relations from file..." << endl;
      DynamicRelations_from_file.seekg(0,ios::beg);
      while (DynamicRelations_from_file.peek()=='G') 
	{
	  TDynamicFactorRelation relation;
	  relation.fpos = DynamicRelations_from_file.tellg();
	  string s; DynamicRelations_from_file >> s;
	  DynamicRelations_from_file >> relation.factor;
	  relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
	  DynamicRelations_from_file.ignore(1000000,'\n');
	  DynamicFactorRelations.insert(relation);
	}
      DynamicRelations_from_file.clear();
      DynamicRelations_to_file.seekp(0, ios::end); // sicherheitshalber...
      cout << "Recover: " << DynamicFactorRelations.size()
	   << " dynamic relations read." << endl;

      SpecialRelations::Load();
      SpecialRelations::CycleSearch();
      
      
#ifdef IS_SERVER
      // beim verteilten Sieben knnten den Clients einige potentielle dynamische Faktoren
      // durch die Lappen gegangen sein. Dies geschieht, falls ein Client einen Special-Factor
      // findet, der nicht mit dem vollen Satz der aktuellen (und daher nur dem Server bekannten)
      // dynamischen Faktoren gesiebt wurde.
      // Anstelle eines Tests zur Laufzeit werden diese im Normalfall eher seltenen Flle
      // nun hier beim Neu-Hochfahren des Servers bercksichtigt:
      // WICHTIG: erst mssen alle dynamischen Faktoren eingelesen sein! Ansonsten knnen
      //          doppelt auftretende dynamische Faktoren nicht zu einer Relation reduziert werden!
      // folgende Anweisungen wrden das bewerkstelligen, aber es lohnt sich kaum und dauert (zu) lange:
#if 1 /* 0: wir lassen es, 1: wir machen es (trotzdem) */
      cout << "Try to discover previously hidden special factors." << endl;
      for (TDynamicFactorRelations::iterator pos = DynamicFactorRelations.begin();
	   pos!=DynamicFactorRelations.end(); ++pos)
	SpecialRelations::split_by_primefactor((*pos).factor);
      cout << "discovered dynamic factors: " << statistical_data::Special_to_dynamic_Factor_hit << endl;
#endif
#endif
      
    }
#else // is client:
#ifndef USE_NETWORK 
  TDynamicFactorRelation relation;
  relation.fpos = 0; relation.factor=0;
  int diff; // wir lesen die Abstnde zwischen den Faktoren ein
  communication_stream2 >> diff; // filepos ueberlesen
  while (true)
    {
      communication_stream2 >> diff;
      relation.factor+=diff;
      // cout << "*" << relation.factor << "*" << flush;
      if (diff<=0) break;
      relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
      DynamicFactorRelations.insert(relation);
    }
  //cout << endl;
#endif
#endif  
  
  
  
#ifndef IS_CLIENT
  if (recover)
    {
      streampos fpos = Recovery_from_file.tellg();
      Polynom.load_if_available(Recovery_from_file); // aktuelles Polynom (zum Fortsetzen) ermitteln
      Recovery_to_file.seekp(fpos);
    }
#endif

display_StatusLegend();

#ifdef IS_SERVER
  cout << "Waiting for clients (MPQS)..." << endl;
#if 0
  signal(SIGINT,  signalhandler);
  signal(SIGTSTP, signalhandler);
  signal(SIGQUIT, signalhandler);
  signal(SIGTTIN, signalhandler);
  signal(SIGTTOU, signalhandler);
  signal(SIGHUP,  signalhandler);
  signal(SIGCHLD, signalhandler);
#endif
  while (true)
   {    
     int* pdata = new int;
     *pdata = my_connection_waiter.get_new_client_socket_descriptor();
     pthread_attr_t detached_thread;
     pthread_t thread_process_client_data;

     pthread_attr_init(&detached_thread);
     pthread_attr_setdetachstate(&detached_thread, PTHREAD_CREATE_DETACHED);

     //cout << "creating thread..." << endl;
     const int retcode = pthread_create(&thread_process_client_data,&detached_thread,
                          Cprocess_clients::THREAD_bearbeite_daten_stream, pdata);
     if (retcode != 0)
      {
        cerr << "pthread_create failed!" << endl;
	delete pdata;
        exit(1);
      }
      
     // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
     /* Durch das Flag "collecting_phase_finished" ist es mglich, das Aufsammeln weiterer dynamischer
        oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	Sammeln der eigentlichen Relationen macht, dass dynamische oder special-Faktoren hinderlich
	wirken.
	Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	des Programms wesentlich flexibler beeinflussen.
      */
   }

// Server
#else
// not Server


#ifdef IS_CLIENT
  mpz_t Obergrenze_D;
  mpz_init(Obergrenze_D);
    
#ifdef USE_NETWORK

 ClientMainLoop: // Hauptschleife: Polynomintervall holen und diese durchsieben

  CClientPolynomFetcher::fetch(Obergrenze_D);

#else // kein Netzwerk
  Polynom.load(communication_stream2);
  communication_stream2 >> Obergrenze_D;
  if (communication_stream.fail())
   {
     cerr << "No more data from file??" << endl;
     exit(1);
   }
#endif // if Netzwerk else keinNetzwerk
#endif // Client
  
  /* Quadratischen Reste mit Primzahlen aus der Primzahlbasis aussieben */
  do /* solange Relationen aussieben, bis mehr Relationen gefunden sind
	als Primzahlen in der Primzahlbasis sind (oder eine Faktorisierung
	gefunden ist). */
    {
#ifdef USE_NETWORK
      if (StaticRelations::Count() > 4*Size_StaticFactorbase)
       {
          cerr << "#static relations > 4*Size_StaticFactorbase: this is weird." << endl;
          cerr << "I will sleep 30 seconds for transmission to end, and then I will abort." << endl;
          for (int i=0; i<15; ++i)
           {
             CClientRelation_Versand::go();
             sleep(2);
           }
          exit(1);
       }
      CClientDynamicFactorFetcher::fetch();
#endif

#ifndef IS_CLIENT
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.save(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
#endif

      SieveOffset=-LogicalSieveSize;
      recompute_all_Deltas();
#ifdef USE_FAKEHEAP
      Sieve_Delta_Heap.sort();
#endif
      CSieveStaticFactors::init_LocalDeltas();
      while (SieveOffset<LogicalSieveSize)
	{
	  initialize_Sieve();
	  CSieveStaticFactors::do_sieving_StaticFactors(); // mit Primzahlen der statischen Faktorbasis
	  do_sieving_Squares(); // (grere) Quadrate von statischen Faktoren sieben
	  do_sieving_DynamicFactors(); // ermglicht weitere Relationen...
	  do_scanning_Sieve();
	  SieveOffset+=PhysicalSieveSize;
	}
      StatusReport(); // Ausgabe von Informationen

#if defined(IS_CLIENT) && defined(USE_NETWORK)
      CClientRelation_Versand::go();
#endif  // Client && Network

      Polynom.compute_next_polynomial();
      
      // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
      /* Durch das Flag "collecting_phase_finished" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, dass dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
    }
#ifndef IS_CLIENT
  while (true);
#else
  while (Polynom < Obergrenze_D);
#ifdef USE_NETWORK
  goto ClientMainLoop; // Hauptschleife: neues Polynomintervall holen und fortsetzen
#endif
  mpz_clear (Obergrenze_D);
#endif
  
#endif /* not server */
  
  cout << endl << "Session ended successfully." << endl;
  return 0;
}
